/*****************************************************************************
 *
 * camera.h -- camera API header
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#ifndef OPENASTRO_CAMERA_H
#define OPENASTRO_CAMERA_H

#include <openastro/openastro.h>
#include <openastro/camera/controls.h>
#include <openastro/camera/formats.h>
#include <openastro/camera/features.h>

enum oaCameraInterfaceType {
  OA_CAM_IF_V4L2			= 1,
  OA_CAM_IF_PWC				= 2,
  OA_CAM_IF_ZWASI			= 3,
  OA_CAM_IF_QHY				= 4,
  OA_CAM_IF_IIDC			= 5,
  OA_CAM_IF_UVC				= 6,
  OA_CAM_IF_SX				= 7,
  OA_CAM_IF_ATIK_SERIAL			= 8,
  OA_CAM_IF_COUNT			= 9
};

extern oaInterface	oaCameraInterfaces[ OA_CAM_IF_COUNT + 1 ];

// This may need looking at...

#define	OA_MAX_RESOLUTIONS		31
#define	OA_MAX_FRAMERATES		31

typedef struct FRAMESIZE {
  unsigned int		x;
  unsigned int		y;
} FRAMESIZE;

typedef struct FRAMERATE {
  int			numerator;
  int			denominator;
} FRAMERATE;

typedef struct START_PARMS {
  FRAMESIZE	size;
  FRAMERATE	rate;
} START_PARMS;

// ...to here

typedef struct OA_CONTROL_VALUE {
  unsigned char	valueType;
  union {
    int32_t	int32;
    int64_t	int64;
    int8_t	boolean;
    int8_t	menu;
    int8_t	discrete;
    int64_t	readonly;
    const char* string;
  };
} oaControlValue;

struct oaCamera;
struct oaCameraDevice;

typedef struct oaCameraFuncs {
  struct oaCamera* ( *initCamera )( struct oaCameraDevice* );
  int              ( *closeCamera )( struct oaCamera* );
  int              ( *resetCamera )( struct oaCamera* );
  int              ( *startCapture )( struct oaCamera* );
  int              ( *startCaptureExtended )( struct oaCamera*,
                       START_PARMS* );
  int              ( *stopCapture )( struct oaCamera* );

  int              ( *hasLoadableFirmware )( struct oaCamera* );
  int              ( *isFirmwareLoaded )( struct oaCamera* );
  int              ( *loadFirmware )( struct oaCamera* );

  const char*      ( *getName )( struct oaCamera* );
  const char*      ( *getMenuString )( struct oaCamera*, int, int );

  int              ( *getControlRange )( struct oaCamera*, int, int64_t*,
                       int64_t*, int64_t*, int64_t* );
  int              ( *readControl )( struct oaCamera*, int, oaControlValue* );
  int              ( *setControl )( struct oaCamera*, int, oaControlValue* );
  int              ( *testControl )( struct oaCamera*, int, oaControlValue* );
  int              ( *setControlMulti )( struct oaCamera* );
  int              ( *getAutoWBManualSetting )( struct oaCamera* );
  int              ( *hasAuto )( struct oaCamera*, int );
  int              ( *isAuto )( struct oaCamera*, int );

  int              ( *startReadFrame )( struct oaCamera*, unsigned int );
  int              ( *readFrame )( struct oaCamera*, void** );	
  int              ( *finishReadFrame )( struct oaCamera* );

  const FRAMESIZE* ( *enumerateFrameSizes )( struct oaCamera* );
  const FRAMERATE* ( *enumerateFrameRates )( struct oaCamera*, int, int );
  int	           ( *hasFixedFrameRates )( struct oaCamera*, int, int );

  int              ( *getFramePixelFormat )( struct oaCamera*, int );
  int              ( *testROISize )( struct oaCamera*, unsigned int,
                       unsigned int, unsigned int*, unsigned int* );
} oaCameraFuncs;

#define OA_MAX_NAME_LEN			40
#define OA_MAX_DEVICES			32

#if HAVE_LIBDC1394
#include <openastro/camera/IIDCstate.h>
#endif
#include <openastro/camera/PWCstate.h>
#include <openastro/camera/QHYstate.h>
#if HAVE_LIBV4L2
#include <openastro/camera/V4L2state.h>
#endif
#include <openastro/camera/ZWASIstate.h>
#if HAVE_LIBUVC
#include <openastro/camera/UVCstate.h>
#endif
#include <openastro/camera/SXstate.h>
#if HAVE_LIBUDEV || HAVE_LIBFTDI
#include <openastro/camera/AtikSerialstate.h>
#endif

typedef struct oaCamera {
  enum oaCameraInterfaceType	interface;
  char				deviceName[OA_MAX_NAME_LEN+1];
  oaCameraFuncs			funcs;
  uint8_t			controls[OA_CAM_CTRL_LAST_P1];
  int64_t               	min[ OA_CAM_CTRL_LAST_P1 ];
  int64_t               	max[ OA_CAM_CTRL_LAST_P1 ];
  int64_t               	step[ OA_CAM_CTRL_LAST_P1 ];
  int64_t               	def[ OA_CAM_CTRL_LAST_P1 ];
  oaCameraFeatures		features;
  union {
#if HAVE_LIBDC1394
    IIDC_STATE           	_iidc;
#endif
    PWC_STATE           	_pwc;
    QHY_STATE           	_qhy;
#if HAVE_LIBV4L2
    V4L2_STATE          	_v4l2;
#endif
    ZWASI_STATE         	_zwasi;
#if HAVE_LIBUVC
    UVC_STATE           	_uvc;
#endif
    SX_STATE            	_sx;
#if HAVE_LIBUDEV || HAVE_LIBFTDI
    AtikSerial_STATE    	_atikSerial;
#endif
  };
} oaCamera;

typedef struct oaCameraDevice {
  enum oaCameraInterfaceType	interface;
  char				deviceName[OA_MAX_NAME_LEN+1];
  oaCamera*			( *initCamera )( struct oaCameraDevice* );
  int           		( *hasLoadableFirmware )( struct oaCameraDevice* );
  int           		( *isFirmwareLoaded )( struct oaCameraDevice* );
  int           		( *loadFirmware )( struct oaCameraDevice* );
  unsigned long			_devIndex;
  unsigned long			_devType;
  unsigned short		_vendorId;
  unsigned short		_productId;
  unsigned long			_misc;
  uint64_t			_guid;
  uint8_t			_unit;
  unsigned int			_colour;
  char				sysPath[ PATH_MAX+1 ];
  unsigned int			hasFirmware;
  unsigned int			firmwareLoaded;
} oaCameraDevice;

extern int		oaGetCameras ( oaCameraDevice** );
extern unsigned		oaGetCameraAPIVersion ( void );
extern const char*	oaGetCameraAPIVersionStr ( void );
extern void		oaSetCameraDebugLevel ( int );
extern int		oaGetAutoForControl ( int );
extern int		oaGetControlForAuto ( int );
extern int		oaIsAuto ( int );
extern void		oaSetRootPath ( const char* );

// These may no longer be required

#define OA_FRAMESIZES_DISCRETE		1
#define OA_FRAMESIZES_CONTINUOUS	2
#define OA_FRAMESIZES_STEPWISE		3

#define OA_FRAMERATES_DISCRETE		1
#define OA_FRAMERATES_CONTINUOUS	2
#define OA_FRAMERATES_STEPWISE		3

#define	OA_BIN_MODE_NONE		1
#define	OA_BIN_MODE_2x2			2
#define	OA_BIN_MODE_3x3			3
#define	OA_BIN_MODE_4x4			4

#endif	/* OPENASTRO_CAMERA_H */
