/*****************************************************************************
 *
 * IIDCcontrol.c -- control functions for IIDC cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBDC1394

#include <openastro/camera.h>

#include "IIDCoacam.h"

#define	cameraState		camera->_iidc


int
oaIIDCCameraSetControl ( oaCamera* camera, int control, oaControlValue* val )
{
  int found, iidcControl;
  unsigned int i;

  for ( i = 0, found = -1; found < 0 && i < numIIDCControls; i++ ) {
    if ( dc1394Controls[i].oaControl == control ||
        dc1394Controls[i].oaAutoControl == control ) {
      iidcControl = dc1394Controls[i].iidcControl;
      found = 1;
    }
  }

  if ( OA_CAM_CTRL_EXPOSURE == control || OA_CAM_CTRL_EXPOSURE_ABSOLUTE ==
      control ) {
    iidcControl = DC1394_FEATURE_SHUTTER;
  }

  if ( oaIsAuto ( control )) {
    uint32_t val_u32;
    if ( OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
      // FIX ME -- actually perhaps this should be DISCRETE
      // This should be INT32, though actually non-negative
      if ( OA_CTRL_TYPE_BOOLEAN != val->valueType ) {
        fprintf ( stderr, "%s: invalid control type %d where bool expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      val_u32 = val->int32;
      if ( val_u32 != OA_EXPOSURE_AUTO && val_u32 != OA_EXPOSURE_MANUAL ) {
        fprintf ( stderr, "%s: control value out of range\n", __FUNCTION__ );
        return -OA_ERR_OUT_OF_RANGE;
      }
      // need to switch this around for the ternary operator below to work
      val_u32 = ( OA_EXPOSURE_AUTO == val_u32 ) ? 1 : 0;
    } else {
      // anything here should be a boolean value
      if ( OA_CTRL_TYPE_BOOLEAN != val->valueType ) {
        fprintf ( stderr, "%s: invalid control type %d where bool expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      val_u32 = val->boolean;
    }
    if ( dc1394_feature_set_mode ( cameraState.iidcHandle, iidcControl,
        val_u32 ? DC1394_FEATURE_MODE_AUTO : DC1394_FEATURE_MODE_MANUAL ) !=
        DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_mode failed for control %d\n",
          __FUNCTION__, iidcControl );
      return -1;
    }
    return OA_ERR_NONE;
  }

  if ( found >= 0 || OA_CAM_CTRL_EXPOSURE == control ) {
    uint32_t val_u32;
    int64_t val_s64;
    // The value for dc1394_feature_set_value() is uint32_t, which we have
    // to map to an int64_t, so we need to throw an error if we don't have
    // an INT64 value here, or if it is negative
    if ( OA_CTRL_TYPE_INT64 != val->valueType ) {
      fprintf ( stderr, "%s: invalid control type %d where int64 expected "
          "for OA_CAM_CTRL_EXPOSURE\n", __FUNCTION__, val->valueType );
      return -OA_ERR_INVALID_CONTROL_TYPE;
    }
    val_s64 = val->int64;
    if ( val < 0 ) {
      fprintf ( stderr, "%s: invalid control value %d for exposure\n",
          __FUNCTION__, val->valueType );
      return -OA_ERR_OUT_OF_RANGE;
    }
    val_u32 = val_s64 & 0xffffffff;
    if ( dc1394_feature_set_value ( cameraState.iidcHandle, iidcControl,
        val_u32 ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_value failed for "
          "control %d\n", __FUNCTION__, control );
      return -1;
    }
    return OA_ERR_NONE;
  }

  // need to handle absolute exposure time separately
  if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == control ) {
    float decval;
    int64_t val_s64;
    // We should also have a 64-bit value here and it can't be negative.
    // Probably shouldn't be non-positive, really.
    if ( OA_CTRL_TYPE_INT64 != val->valueType ) {
      fprintf ( stderr, "%s: invalid control type %d where int64 expected "
          "for OA_CAM_CTRL_EXPOSURE_ABSOLUTE\n", __FUNCTION__,
          val->valueType );
      return -OA_ERR_INVALID_CONTROL_TYPE;
    }
    val_s64 = val->int64;
    decval = val_s64 / 1000000.0;
    if ( dc1394_feature_set_absolute_value ( cameraState.iidcHandle,
        iidcControl, decval ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_set_absolute_value %d failed\n",
          __FUNCTION__, control );
      return -1;
    }
    return OA_ERR_NONE;
  }

  fprintf ( stderr, "Unrecognised control %d in %s\n", control, __FUNCTION__ );
  return -OA_ERR_INVALID_CONTROL;
}


int
oaIIDCCameraReadControl ( oaCamera* camera, int control, oaControlValue* val )
{
  int found, iidcControl;
  unsigned int i;

  for ( i = 0, found = 0; !found && i < numIIDCControls; i++ ) {
    if ( dc1394Controls[i].oaControl == control ||
        dc1394Controls[i].oaAutoControl == control ) {
      iidcControl = dc1394Controls[i].iidcControl;
      found = 1;
    }
  }

  if ( found ) {
    if ( iidcControl != DC1394_FEATURE_SHUTTER || OA_CAM_CTRL_EXPOSURE ==
        control || OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
      if ( !oaIsAuto ( control )) {
        uint32_t val_u32;
        val->valueType = OA_CTRL_TYPE_INT64;
        if ( dc1394_feature_get_value ( cameraState.iidcHandle, iidcControl,
            &val_u32 ) != DC1394_SUCCESS ) {
          fprintf ( stderr, "%s: dc1394_feature_get_value failed for "
              "control %d\n", __FUNCTION__, control );
          return -1;
        }
        val->int64 = val_u32;
      } else {
        dc1394feature_mode_t mode;

        if ( dc1394_feature_get_mode ( cameraState.iidcHandle, iidcControl,
            &mode ) != DC1394_SUCCESS ) {
          fprintf ( stderr, "%s: dc1394_feature_get failed for feature %d\n",
              __FUNCTION__, iidcControl );
          return -1;
        } else {
          if ( OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
            // FIX ME -- should this be DISCRETE?
            val->valueType = OA_CTRL_TYPE_INT32;
            val->int32 = ( mode == DC1394_FEATURE_MODE_AUTO ) ?
                OA_EXPOSURE_AUTO : OA_EXPOSURE_MANUAL;
          } else {
            val->valueType = OA_CTRL_TYPE_BOOLEAN;
            val->boolean = ( mode == DC1394_FEATURE_MODE_AUTO ) ? 1 : 0;
          }
        }
      }
    }
    return OA_ERR_NONE;
  }

  if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == control ) {
    // need to handle absolute exposure time separately
    float decval;
    val->valueType = OA_CTRL_TYPE_INT64;
    if ( dc1394_feature_get_absolute_value ( cameraState.iidcHandle,
        DC1394_FEATURE_SHUTTER, &decval ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_feature_get_absolute_value %d failed\n",
          __FUNCTION__, iidcControl );
      return -1;
    } else {
      val->int64 = decval * 1000000.0;
    }
    return OA_ERR_NONE;
  }

  fprintf ( stderr, "Unrecognised control %d in %s\n", control, __FUNCTION__ );
  return -OA_ERR_INVALID_CONTROL;
}


int
oaIIDCCameraTestControl ( oaCamera* camera, int control, oaControlValue* val )
{
  uint32_t	val_u32;
  int64_t	val_s64;

  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  if ( camera->controls [ control ] != val->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
    case OA_CAM_CTRL_CONTRAST:
    case OA_CAM_CTRL_SATURATION:
    case OA_CAM_CTRL_HUE:
    case OA_CAM_CTRL_SHARPNESS:
    case OA_CAM_CTRL_GAMMA:
    case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_GAIN:
      // This should be unsigned 32-bit
      val_s64 = val->int64;
      if ( val_s64 < 0 ) {
        return -OA_ERR_OUT_OF_RANGE;
      }
      val_u32 = val_s64 & 0xffffffff;
      if ( val_u32 >= camera->min[ control ] &&
          val_u32 <= camera->max[ control ] &&
          ( 0 == ( val_u32 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val_s64 = val->int64;
      if ( val_s64 <= 0 ) {
        return -OA_ERR_OUT_OF_RANGE;
      }
      return OA_ERR_NONE;

    case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
    case OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_HUE_AUTO:
      // These just need to be boolean and we've already checked that
      return OA_ERR_NONE;
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      if ( val->int32 != OA_EXPOSURE_AUTO && val->int32 !=
          OA_EXPOSURE_MANUAL ) {
        return -OA_ERR_OUT_OF_RANGE;
      }
      return OA_ERR_NONE;
      break;     

    case OA_CAM_CTRL_BIT_DEPTH:
    case OA_CAM_CTRL_BINNING:
      return -OA_ERR_INVALID_CONTROL;
      break;

    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
oaIIDCCameraSetROI ( oaCamera* camera, int x, int y )
{
  fprintf ( stderr, "attempt to set ROI of %dx%d in %s\n", x, y,
      __FUNCTION__ );
  return 0;
}


int
oaIIDCCameraStart ( oaCamera* camera, START_PARMS* parms )
{
  // configure the camera for the requested resolution and frame rate
  // with the current colour mode and start capturing frames

  dc1394video_modes_t		videoModes;
  dc1394format7modeset_t	modeList;
  dc1394color_coding_t		codec;
  dc1394framerates_t		framerates;
  unsigned int			i, j, thisOne, requiredFrameRate;
  /*
  unsigned int			minSize, maxSize, actualBytes, w, h;
  uint64_t			totalBytes;
   */
  int				ret, matched;

  if ( cameraState.started ) {
    dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_OFF );
    dc1394_capture_stop ( cameraState.iidcHandle );
    cameraState.started = 0;
  }

  if ( cameraState.haveFormat7 ) {

    // Now we have to find a Format7 mode with the correct colour coding
    // and frame size

    if ( dc1394_format7_get_modeset ( cameraState.iidcHandle, &modeList ) !=
        DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_format7_get_modeset return error\n",
          __FUNCTION__ );
      return -1;
    }

    matched = 0;
    for ( i = 0; !matched && i < DC1394_VIDEO_MODE_FORMAT7_NUM; i++ ) {
      if ( modeList.mode[i].present ) {
        if ( modeList.mode[i].size_x == parms->size.x &&
            modeList.mode[i].size_y == parms->size.y ) {
          for ( j = 0; j < modeList.mode[i].color_codings.num; j++ ) {
            if ( modeList.mode[i].color_codings.codings[j] ==
                cameraState.currentCodec ) {
              thisOne = i + DC1394_VIDEO_MODE_FORMAT7_MIN;
              matched = 1;
            }
          }
        }
      }
    }

    if ( !matched ) {
      fprintf ( stderr, "%s: matching Format7 mode not found\n", __FUNCTION__ );
      return -1;
    }

    if ( dc1394_video_set_mode ( cameraState.iidcHandle, thisOne ) !=
        DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: unable to set Format7 mode %d\n", __FUNCTION__,
          thisOne );
      return -1;
    }

    if ( dc1394_format7_set_roi ( cameraState.iidcHandle, thisOne,
        cameraState.currentCodec, DC1394_USE_MAX_AVAIL, 0, 0,
        parms->size.x, parms->size.y ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: unable to set mode %d, codec %d, roi %dx%d\n",
          __FUNCTION__, thisOne, cameraState.currentCodec, parms->size.x,
          parms->size.y );
      return -1;
    }

    /*
    dc1394_format7_get_packet_parameters ( cameraState.iidcHandle,
        thisOne, &minSize, &maxSize );
    fprintf ( stderr, "%s: minSize = %d, maxSize = %d\n", __FUNCTION__,
        minSize, maxSize );

    dc1394_format7_get_packet_size ( cameraState.iidcHandle, thisOne,
        &actualBytes );
    fprintf ( stderr, "%s: actual bytes = %d\n", __FUNCTION__, actualBytes );

    dc1394_format7_get_total_bytes ( cameraState.iidcHandle, thisOne,
        &totalBytes );
    fprintf ( stderr, "%s: total bytes = %lu\n", __FUNCTION__, totalBytes );

    dc1394_format7_get_image_size ( cameraState.iidcHandle, thisOne, &w, &h );
    fprintf ( stderr, "%s: image size = %dx%d\n", __FUNCTION__, w, h );
     */

  } else {

    requiredFrameRate = 0;
    for ( i = 0; i < numIIDCFrameRates; i++ ) {
      if ( dc1394FrameRates[i].numerator == parms->rate.numerator &&
          dc1394FrameRates[i].denominator == parms->rate.denominator ) {
        requiredFrameRate = dc1394FrameRates[i].iidcFrameRate;
      }
    }

    if ( !requiredFrameRate ) {
      fprintf ( stderr, "%s: no frame rate matching %d/%d found\n",
          __FUNCTION__, parms->rate.numerator, parms->rate.denominator );
      return -1;
    }

    if ( dc1394_video_get_supported_modes ( cameraState.iidcHandle,
        &videoModes ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_video_get_supported_modes failed",
          __FUNCTION__ );
      return -1;
    }

    thisOne = 0;
    for ( i = 0, matched = 0; !matched && i < videoModes.num; i++ ) {
      uint32_t w, h;

      if ( videoModes.modes[i] >= DC1394_VIDEO_MODE_FORMAT7_MIN &&
          videoModes.modes[i] <= DC1394_VIDEO_MODE_FORMAT7_MAX ) {
        continue;
      }

      if ( dc1394_get_image_size_from_video_mode ( cameraState.iidcHandle,
          videoModes.modes[i], &w, &h ) != DC1394_SUCCESS ) {
        fprintf ( stderr, "%s: dc1394_get_image_size_from_video_mode failed",
          __FUNCTION__ );
      } else {
        if ( w == parms->size.x && h == parms->size.y ) {
          if ( dc1394_get_color_coding_from_video_mode ( cameraState.iidcHandle,
              videoModes.modes[i], &codec ) != DC1394_SUCCESS ) {
            fprintf ( stderr, "%s: dc1394_get_color_coding_from_video_mode "
                "failed", __FUNCTION__ );
          } else {
            if ( codec == cameraState.currentCodec ) {
              if ( dc1394_video_get_supported_framerates (
                  cameraState.iidcHandle, videoModes.modes[i], &framerates )
                  != DC1394_SUCCESS ) {
                fprintf ( stderr, "%s: dc1394_video_get_supported_framerates "
                    "failed\n", __FUNCTION__ );
              } else {
                for ( j = 0; !matched && j < framerates.num; j++ ) {
                  if ( framerates.framerates[j] == requiredFrameRate ) {
                    thisOne = i;
                    matched = 1;
                  }
                }
              }
            }
          }
        }
      }
    }

    if ( !matched ) {
      fprintf ( stderr, "%s: can't find compatible video mode\n",
          __FUNCTION__ );
      return -1;
    }

    if ( dc1394_video_set_mode ( cameraState.iidcHandle,
        videoModes.modes[ thisOne ]) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_video_set_mode failed\n", __FUNCTION__ );
      return -1;
    }
    if ( dc1394_video_set_framerate ( cameraState.iidcHandle,
        requiredFrameRate ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_video_set_framerate failed\n",
          __FUNCTION__ );
      return -1;
    }
  }

  if (( ret = dc1394_capture_setup ( cameraState.iidcHandle,
      NUM_IIDC_BUFFERS, DC1394_CAPTURE_FLAGS_DEFAULT )) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_capture_setup failed: %d\n", __FUNCTION__,
        ret );
    return -1;
  }

  cameraState.started = 1;
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_ON ) !=
      DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
        __FUNCTION__ );
    return -1;
  }

  return 0;
}


int
oaIIDCCameraStop ( oaCamera* camera )
{
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_OFF )
      != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
      __FUNCTION__ );
    return -1;
  }
  if ( cameraState.started ) {
    if ( dc1394_capture_stop ( cameraState.iidcHandle ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_capture_stop failed\n", __FUNCTION__ );
      return -1;
    }
    cameraState.started = 0;
  }
  return 0;
}


int
oaIIDCCameraReset ( oaCamera* camera )
{
  if ( dc1394_video_set_transmission ( cameraState.iidcHandle, DC1394_OFF )
      != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_set_transmission failed\n",
      __FUNCTION__ );
    return -1;
  }
  if ( cameraState.started ) {
    if ( dc1394_capture_stop ( cameraState.iidcHandle ) != DC1394_SUCCESS ) {
      fprintf ( stderr, "%s: dc1394_capture_stop failed\n", __FUNCTION__ );
      return -1;
    }
    cameraState.started = 0;
  }
  if ( dc1394_capture_setup ( cameraState.iidcHandle, NUM_IIDC_BUFFERS,
      DC1394_CAPTURE_FLAGS_DEFAULT ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_capture_setup failed\n", __FUNCTION__ );
    return -1;
  }
  cameraState.started = 1;
  return 0;
}


const char*
oaIIDCCameraGetMenuString ( oaCamera* camera, int control, int index )
{
  fprintf ( stderr, "%s: control not implemented\n", __FUNCTION__ );
  return "";
}

#endif /* HAVE_LIBDC1394 */
