/*****************************************************************************
 *
 * IIDCgetState.c -- state querying for IEEE1394/IIDC cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBDC1394

#include <dc1394/dc1394.h>
#include <openastro/camera.h>

#include "IIDCoacam.h"
#include "oacamprivate.h"

#define	cameraState		camera->_iidc

int
oaIIDCCameraGetControlRange ( oaCamera* camera, int control, int64_t* min,
    int64_t* max, int64_t* step, int64_t* def )
{
  *min = camera->min [ control ];
  *max = camera->max [ control ];
  *step = camera->step [ control ];
  *def = camera->def [ control ];
  return 0;
}


const FRAMESIZE*
oaIIDCCameraGetFrameSizes ( oaCamera* camera )
{
  return cameraState.resolutions;
}


const FRAMERATE*
oaIIDCCameraGetFrameRates ( oaCamera* camera, int resX, int resY )
{
  dc1394framerates_t    framerates;
  int numRates;
  unsigned int i;

  if ( dc1394_video_get_supported_framerates ( cameraState.iidcHandle,
      cameraState.videoCurrent, &framerates ) != DC1394_SUCCESS ) {
    fprintf ( stderr, "%s: dc1394_video_get_supported_framerates failed\n",
         __FUNCTION__ );
    return 0;
  }
  if ( !framerates.num ) {
    fprintf ( stderr, "%s: dc1394_video_get_supported_framerates returns "
        "no frame rates\n", __FUNCTION__ );
    return 0;
  }

  numRates = 0;
  for ( i = 0; i < framerates.num; i++ ) {
    switch ( framerates.framerates[i] ) {
      case DC1394_FRAMERATE_1_875:
        cameraState.framerates[numRates].numerator = 8;
        cameraState.framerates[numRates].denominator = 15;
        numRates++;
        break;
      case DC1394_FRAMERATE_3_75:
        cameraState.framerates[numRates].numerator = 4;
        cameraState.framerates[numRates].denominator = 15;
        numRates++;
        break;
      case DC1394_FRAMERATE_7_5:
        cameraState.framerates[numRates].numerator = 2;
        cameraState.framerates[numRates].denominator = 15;
        numRates++;
        break;
      case DC1394_FRAMERATE_15:
        cameraState.framerates[numRates].numerator = 1;
        cameraState.framerates[numRates].denominator = 15;
        numRates++;
        break;
      case DC1394_FRAMERATE_30:
        cameraState.framerates[numRates].numerator = 1;
        cameraState.framerates[numRates].denominator = 30;
        numRates++;
        break;
      case DC1394_FRAMERATE_60:
        cameraState.framerates[numRates].numerator = 1;
        cameraState.framerates[numRates].denominator = 60;
        numRates++;
        break;
      case DC1394_FRAMERATE_120:
        cameraState.framerates[numRates].numerator = 1;
        cameraState.framerates[numRates].denominator = 120;
        numRates++;
        break;
      case DC1394_FRAMERATE_240:
        cameraState.framerates[numRates].numerator = 1;
        cameraState.framerates[numRates].denominator = 240;
        numRates++;
        break;
      default:
        fprintf ( stderr, "%s: unknown frame rate %d\n", __FUNCTION__,
            framerates.framerates[i] );
        break;
    }
  }

  if ( !numRates ) {
    return 0;
  }

  cameraState.framerates[numRates].numerator = 0;
  cameraState.framerates[numRates].denominator = 0;

  return cameraState.framerates;
}


const char*
oaIIDCCameraGetName ( oaCamera* camera )
{
  return camera->deviceName;
}


int
oaIIDCCameraGetFramePixelFormat ( oaCamera* camera, int depth )
{
  dc1394color_coding_t  codec;

  if ( dc1394_get_color_coding_from_video_mode ( cameraState.iidcHandle,
      cameraState.videoCurrent, &codec ) == DC1394_SUCCESS ) {

    switch ( codec ) {
      case DC1394_COLOR_CODING_MONO8:
        return OA_PIX_FMT_GREY8;
        break;

      case DC1394_COLOR_CODING_MONO16:
        return OA_PIX_FMT_GBRG16LE;
        break;

      case DC1394_COLOR_CODING_RGB8:
        return OA_PIX_FMT_RGB24;
        break;

      case DC1394_COLOR_CODING_RAW8:
        // FIX ME -- may not be at all, of course
        // There is a way to tell this, but at the moment it only appears
        // to be on a frame-by-frame basis rather than when the camera is
        // configured
        return OA_PIX_FMT_GBRG8;
        break;

      case DC1394_COLOR_CODING_YUV422:
        // normally this would be OA_PIX_FMT_YUYV, but by definition IIDC
        // cameras use UYVY for YUV422
        return OA_PIX_FMT_UYVY;
        break;

      default:
        break;
    }
  }

  fprintf ( stderr, "%s: dc1394_get_color_coding_from_video_mode failed\n",
      __FUNCTION__ );
    
  return OA_PIX_FMT_RGB24;
}

#endif /* HAVE_LIBDC1394 */
