/*****************************************************************************
 *
 * QHY6.c -- QHY6 camera interface
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>
#include <openastro/errno.h>

#include <libusb-1.0/libusb.h>

#include "QHYoacam.h"
#include "QHY6.h"
#include "oacamprivate.h"
#include "QHYusb.h"

#define	QHY6_BUFFERS		5
#define QHY6_SENSOR_WIDTH	800
#define QHY6_SENSOR_HEIGHT	596
#define OFFSET_X		120

#define	DEFAULT_SPEED		1
#define	DEFAULT_GAIN		31
#define DEFAULT_EXPOSURE	500

#define USB_TIMEOUT		5000

#define MSG_SENDREGS		0xb5

#define AMP_MODE_OFF		0
#define AMP_MODE_ON		1
#define AMP_MODE_AUTO		2


static int      _setParameters ( oaCamera* );
static void     _startExposure ( oaCamera* );
static int      _readExposure ( oaCamera*, void*, unsigned int );
static void*    _capture ( void* );
static void	_QHY6InitFunctionPointers ( oaCamera* );
static void     _recalculateSizes ( oaCamera* );


static int      setControl ( oaCamera*, int, oaControlValue* );
static int      getFramePixelFormat ( oaCamera*, int );
static const FRAMESIZE* getFrameSizes ( oaCamera* );
static const char* getName ( oaCamera* );
static int      stop ( oaCamera* );
static int      start ( oaCamera*, START_PARMS* );
static int      getControlRange ( oaCamera*, int, int64_t*, int64_t*,
			int64_t*, int64_t* );
static int      startReadFrame ( oaCamera*, unsigned int );
static int      finishReadFrame ( oaCamera* );
static int      readFrame ( oaCamera*, void** );
static int      closeCamera ( oaCamera* );

/**
 * Initialise a given camera device
 */

int
_QHY6InitCamera ( oaCamera* camera )
{
  int i, j;

  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _QHY6InitFunctionPointers ( camera );

  camera->controls[ OA_CAM_CTRL_GAIN ] = OA_CTRL_TYPE_INT32;
  camera->min[ OA_CAM_CTRL_GAIN ] = 0;
  camera->max[ OA_CAM_CTRL_GAIN ] = 63;
  camera->step[ OA_CAM_CTRL_GAIN ] = 1;
  camera->def[ OA_CAM_CTRL_GAIN ] = DEFAULT_GAIN;

  camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CTRL_TYPE_INT64;
  camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 0;
  camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 0xffffffff;
  camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1;
  camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = DEFAULT_EXPOSURE * 1000;

  camera->controls[ OA_CAM_CTRL_HIGHSPEED ] = OA_CTRL_TYPE_BOOLEAN;
  camera->min[ OA_CAM_CTRL_HIGHSPEED ] = 0;
  camera->max[ OA_CAM_CTRL_HIGHSPEED ] = 1;
  camera->step[ OA_CAM_CTRL_HIGHSPEED ] = 1;
  camera->def[ OA_CAM_CTRL_HIGHSPEED ] = DEFAULT_SPEED;
  cameraState.currentHighSpeed = DEFAULT_SPEED;

  camera->controls[ OA_CAM_CTRL_BINNING ] = OA_CTRL_TYPE_DISCRETE;

  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED_RESET ] = OA_CTRL_TYPE_BUTTON;

  cameraState.maxResolutionX = QHY6_SENSOR_WIDTH;
  cameraState.maxResolutionY = QHY6_SENSOR_HEIGHT;

  cameraState.videoRGB24 = cameraState.videoGrey = 0;
  cameraState.videoGrey16 = 1;
  cameraState.videoCurrent = OA_PIX_FMT_GREY16BE;

  cameraState.resolutions[0].x = cameraState.maxResolutionX;
  cameraState.resolutions[0].y = cameraState.maxResolutionY;
  cameraState.resolutions[1].x = 0;
  cameraState.resolutions[1].y = 0;

  cameraState.binMode = cameraState.horizontalBinMode =
      cameraState.verticalBinMode = OA_BIN_MODE_NONE;

  cameraState.binMode2Resolutions[0].x = QHY6_SENSOR_WIDTH / 2;
  cameraState.binMode2Resolutions[0].y = QHY6_SENSOR_HEIGHT / 2;
  cameraState.binMode2Resolutions[1].x = 0;
  cameraState.binMode2Resolutions[1].y = 0;

  cameraState.xSize = cameraState.maxResolutionX;
  cameraState.ySize = cameraState.maxResolutionY;
  cameraState.xOffset = OFFSET_X;

  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;

  camera->features.ROI = 0;
  camera->features.hasReset = 0;
  cameraState.topOffset = cameraState.bottomOffset = 0;

  _recalculateSizes ( camera );
  if (!( cameraState.xferBuffer = malloc ( cameraState.captureLength ))) {
    fprintf ( stderr, "malloc of transfer buffer failed in %s\n",
        __FUNCTION__ );
    return 0;
  }

  cameraState.currentGain = camera->def[ OA_CAM_CTRL_GAIN ];
  cameraState.currentExposure = DEFAULT_EXPOSURE;
  cameraState.correctedExposureTime = DEFAULT_EXPOSURE - DEFAULT_EXPOSURE / 10;

  // FIX ME -- need to add amp on/off/auto when I understand what it does
  cameraState.requestedAmpMode = AMP_MODE_AUTO;
  if ( cameraState.correctedExposureTime > 550 &&
      AMP_MODE_AUTO == cameraState.requestedAmpMode ) {
    cameraState.currentAmpMode = 1;
  } else {
    cameraState.currentAmpMode = cameraState.requestedAmpMode ? 0 : 1;
  }

  cameraState.imageBufferLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY * 2;
  cameraState.buffers = calloc ( QHY6_BUFFERS, sizeof ( struct QHYbuffer ));
  for ( i = 0; i < QHY6_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      free ( cameraState.xferBuffer );
      return 0;
    }
  }

  return -1;
}


static int
_setParameters ( oaCamera* camera )
{
  unsigned char buff[64];

  OA_CLEAR ( buff );
  buff[0] = cameraState.currentGain;
  buff[1] = cameraState.xOffset;
  buff[2] = ( cameraState.correctedExposureTime >> 16 ) & 0xff;
  buff[3] = ( cameraState.correctedExposureTime >> 8 ) & 0xff;
  buff[4] = cameraState.correctedExposureTime & 0xff;
  buff[5] = cameraState.horizontalBinMode;
  buff[6] = cameraState.verticalBinMode;
  buff[7] = ( cameraState.xSize >> 8 ) & 0xff;
  buff[8] = cameraState.xSize & 0xff;
  buff[9] = ( cameraState.ySize >> 8 ) & 0xff;
  buff[10] = cameraState.ySize & 0xff;
  buff[11] = ( cameraState.topOffset >> 8 ) & 0xff;
  buff[12] = cameraState.topOffset & 0xff;
  buff[13] = ( cameraState.bottomOffset >> 8 ) & 0xff;
  buff[14] = cameraState.bottomOffset & 0xff;

  buff[17] = ( cameraState.transferPadding >> 8 ) & 0xff;
  buff[18] = cameraState.transferPadding & 0xff;

  buff[32] = cameraState.currentAmpMode;
  buff[33] = cameraState.currentHighSpeed;

  buff[46] = 30;

  buff[58] = 100;

  _usbControlMsg ( camera, 0x01, MSG_SENDREGS, 0, 0, buff, sizeof ( buff ),
      USB_TIMEOUT );
  return 0;
}


static void
_startExposure ( oaCamera* camera )
{
  unsigned char buff[8] = "DEADBEEF";
  // FIX ME -- more magic numbers
  buff[0] = 0;
  buff[1] = 100;
  _usbControlMsg ( camera, 0x01, 0xB3, 0, 0, buff, 2, USB_TIMEOUT );
}


static int
_readExposure ( oaCamera* camera, void* buffer, unsigned int readLength )
{
  int ret;
  unsigned int readSize, expectedSize;

  // FIX ME -- references to droppedFrames here should probably be mutexed

  expectedSize = readLength - 256; // Why?
  ret = _usbBulkTransfer ( camera, QHY_BULK_ENDP_IN, buffer, expectedSize,
      &readSize, USB_TIMEOUT );
  if ( ret ) {
    cameraState.droppedFrames++;
    return ret;
  }
  if ( readSize != expectedSize ) {
    fprintf ( stderr, "readExposure: USB bulk transfer was short. %d != %d\n",
        readSize, expectedSize );
    cameraState.droppedFrames++;
    return -1;
  }
  return 0;
}


static int
setControl ( oaCamera* camera, int control, oaControlValue* val )
{
  int32_t	val_s64;
  int		updateSettings = 0;

  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      if ( val->valueType != OA_CTRL_TYPE_INT32 ) {
        fprintf ( stderr, "%s: invalid control type %d where int32 expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      cameraState.currentGain = val->int64;
      updateSettings = 1;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      uint64_t val_u64;

      if ( val->valueType != OA_CTRL_TYPE_INT64 ) {
        fprintf ( stderr, "%s: invalid control type %d where int64 expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      val_s64 = val->int64;
      val_s64 /= 1000;
      if ( val_s64 < 1 ) { val_s64 = 1; }
      val_u64 = ( uint64_t ) val_s64;
      cameraState.currentExposure = val_u64;
      cameraState.correctedExposureTime = val_u64 - val_u64 / 10;
      if ( cameraState.correctedExposureTime > 550 &&
          AMP_MODE_AUTO == cameraState.requestedAmpMode ) {
        cameraState.currentAmpMode = 1;
      } else {
        cameraState.currentAmpMode = cameraState.requestedAmpMode ? 0 : 1;
      }
      updateSettings = 1;
      break;
    }

    case OA_CAM_CTRL_BINNING:
      if ( val->valueType != OA_CTRL_TYPE_DISCRETE ) {
        fprintf ( stderr, "%s: invalid control type %d where discrete "
            "expected\n", __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      switch ( val->discrete ) {
        case OA_BIN_MODE_NONE:
        case OA_BIN_MODE_2x2:
          pthread_mutex_lock ( &cameraState.captureMutex );
          cameraState.binMode = cameraState.horizontalBinMode =
              cameraState.verticalBinMode = val->discrete;
          // _recalculateSizes ( camera );
          pthread_mutex_unlock ( &cameraState.captureMutex );
          // don't need to update the settings here as a resolution
          // change should be forthcoming
          break;
        default:
          return -OA_ERR_OUT_OF_RANGE;
          break;
      }
      break;

    case OA_CAM_CTRL_HIGHSPEED:
      if ( val->valueType != OA_CTRL_TYPE_BOOLEAN ) {
        fprintf ( stderr, "%s: invalid control type %d where bool expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      cameraState.currentHighSpeed = val->boolean;
      updateSettings = 1;
      break;

    case OA_CAM_CTRL_DROPPED_RESET:
      // FIX ME -- droppedFrames should probably be mutexed
      cameraState.droppedFrames = 0;
      break;

    default:
      fprintf ( stderr, "QHY6: %s not yet implemented for control %d\n",
          __FUNCTION__, control );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  if ( updateSettings ) {
    if ( cameraState.captureThreadStarted ) {
      pthread_mutex_lock ( &cameraState.captureMutex );
      cameraState.doSetParameters = 1;
      pthread_mutex_unlock ( &cameraState.captureMutex );
    } else {
      _setParameters ( camera );
    }
  }

  return OA_ERR_NONE;
}


static int
readControl ( oaCamera* camera, int control, oaControlValue* val )
{
  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      val->valueType = OA_CTRL_TYPE_INT32;
      val->int32 = cameraState.currentGain;
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val->valueType = OA_CTRL_TYPE_INT64;
      val->int64 = cameraState.currentExposure * 1000;
      break;

    case OA_CAM_CTRL_DROPPED:
      val->valueType = OA_CTRL_TYPE_READONLY;
      val->readonly = cameraState.droppedFrames;
      break;

    default:
      fprintf ( stderr,
          "Unrecognised control %d in QHY6:%s\n", control, __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return OA_ERR_NONE;
}


static int
testControl ( oaCamera* camera, int control, oaControlValue* val )
{
  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  if ( camera->controls [ control ] != val->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  switch ( control ) {

    case OA_CAM_CTRL_GAIN:
      if ( val->int32 >= camera->min[ control ] &&
          val->int32 <= camera->max[ control ] &&
          ( 0 == ( val->int32 - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      if ( val->int64 >= camera->min[ control ] &&
          val->int64 <= camera->max[ control ] &&
          ( 0 == ( val->int64 - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_BINNING:
      if ( OA_CTRL_TYPE_DISCRETE == val->valueType &&
          ( OA_BIN_MODE_NONE == val->discrete || OA_BIN_MODE_2x2 ==
          val->discrete )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
    case OA_CAM_CTRL_COLOUR_MODE:
      return -OA_ERR_INVALID_CONTROL;
      break;

    default:
      fprintf ( stderr, "QHY6: %s not yet implemented for control %d\n",
          __FUNCTION__, control );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


static int
getFramePixelFormat ( oaCamera* camera, int depth )
{
  if ( depth > 0 && depth != 16 ) {
    return -OA_ERR_INVALID_BIT_DEPTH;
  }
  return OA_PIX_FMT_GREY16BE;
}


const static FRAMESIZE*
getFrameSizes ( oaCamera* camera )
{
  if ( cameraState.binMode == OA_BIN_MODE_2x2 ) {
    return cameraState.binMode2Resolutions;
  } else {
    return cameraState.resolutions;
  }
}


static const char*
getName ( oaCamera* camera )
{
  return "QHY6";
}


static int
stop ( oaCamera* camera )
{
  if ( !cameraState.captureThreadStarted ) {
    return 0;
  }

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.captureThreadExit = 1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  pthread_join ( cameraState.captureThread, 0 );
  cameraState.captureThreadStarted = 0;
  return 0;
}


static int
start ( oaCamera* camera, START_PARMS* params )
{
  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.xSize = params->size.x;
  cameraState.ySize = params->size.y;

  _recalculateSizes ( camera );
  _setParameters ( camera );
  cameraState.imageBufferLength = cameraState.frameSize;
  pthread_create ( &cameraState.captureThread, 0, &_capture, camera );
  cameraState.captureThreadStarted = 1;
  return 0;
}


static void*
_capture ( void* param )
{
  oaCamera*      camera = param;
  int            quitThread, nextBuffer, doSetParameters, rowBytes;
  unsigned int   i, reorderFrame;
  unsigned char* evenSrc;
  unsigned char* oddSrc;
  unsigned char* s1;
  unsigned char* s2;
  unsigned char* t;

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  cameraState.doSetParameters = 0;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  nextBuffer = 0;

  // see the comments further down for more information about these
  evenSrc = cameraState.xferBuffer;
  oddSrc = cameraState.xferBuffer + cameraState.frameSize / 2;
  rowBytes = cameraState.xSize * 2;
  reorderFrame = ( OA_BIN_MODE_2x2 == cameraState.binMode ) ? 0 : 1;

  while ( 1 ) {
    pthread_mutex_lock ( &cameraState.captureMutex );
    quitThread = cameraState.captureThreadExit;
    if ( nextBuffer == cameraState.nextBufferToRead ) {
      nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
    }
    pthread_mutex_unlock ( &cameraState.captureMutex );
    if ( quitThread ) {
      break;
    }

    _startExposure ( camera );
    usleep ( cameraState.currentExposure );
    _readExposure ( camera, cameraState.xferBuffer,
        cameraState.captureLength );

    // If the frame is unbinned it now has to be unpacked.  There are two
    // "half-frames" in the buffer, one of odd-numbered scanlines and one
    // of even

    t = cameraState.buffers[ nextBuffer ].start;
    if ( reorderFrame ) {
      s1 = evenSrc;
      s2 = oddSrc;
      for ( i = 0; i < cameraState.ySize / 2; i++ ) {
        memcpy ( t, s1, rowBytes );
        t += rowBytes;
        s1 += rowBytes;
        memcpy ( t, s2, rowBytes );
        t += rowBytes;
        s2 += rowBytes;
      }
    } else {
      memcpy ( t, cameraState.xferBuffer, cameraState.frameSize );
    }

    pthread_mutex_lock ( &cameraState.captureMutex );
    cameraState.nextBufferToRead = nextBuffer;
    cameraState.haveDataToRead = 1;
    doSetParameters = cameraState.doSetParameters;
    cameraState.doSetParameters = 0;
    pthread_cond_signal ( &cameraState.frameAvailable );
    pthread_mutex_unlock ( &cameraState.captureMutex );
    nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;

    if ( doSetParameters ) {
      _setParameters ( camera );
    }
  }
  return 0;
}


static int
getControlRange ( oaCamera* camera, int control, int64_t* min, int64_t* max,
    int64_t* step, int64_t* def )
{
  if ( !camera->controls[ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  *min = camera->min[ control ];
  *max = camera->max[ control ];
  *step = camera->step[ control ];
  *def = camera->def[ control ];
  return 0;
}


static int
startReadFrame ( oaCamera* camera, unsigned int frameTime )
{
  cameraState.frameTime = frameTime;
  return 0;
}


static int
readFrame ( oaCamera* camera, void** buffer )
{
  int haveData = 0;
  int bufferToRead;
  struct timespec waitUntil;
  struct timeval now;
  unsigned long frameSec = 0;
  unsigned long frameNanoSec = 0;

  frameNanoSec = cameraState.frameTime * 1000000;
  frameSec = cameraState.frameTime / 1000 + 5;

  pthread_mutex_lock ( &cameraState.captureMutex );
  while ( !cameraState.haveDataToRead ) {

    gettimeofday ( &now, 0 );
    waitUntil.tv_sec = now.tv_sec + frameSec;
    if (( now.tv_usec * 1000 + frameNanoSec ) > 1000000000 ) {
      waitUntil.tv_sec++;
    }
    waitUntil.tv_nsec = ( now.tv_usec * 1000 + frameNanoSec ) % 1000000000;

    pthread_cond_timedwait ( &cameraState.frameAvailable,
        &cameraState.captureMutex, &waitUntil );
  }
  haveData = cameraState.haveDataToRead;
  bufferToRead = cameraState.nextBufferToRead;
  cameraState.haveDataToRead = 0;
  pthread_mutex_unlock ( &cameraState.captureMutex );

  *buffer = haveData ? cameraState.buffers[ bufferToRead ].start : 0;
  return ( haveData ? cameraState.imageBufferLength : 0 );
}


static int
finishReadFrame ( oaCamera* camera )
{
  return 0;
}


static int
closeCamera ( oaCamera* camera )
{
  libusb_release_interface ( cameraState.usbHandle, 0 );
  libusb_close ( cameraState.usbHandle );
  libusb_exit ( cameraState.usbContext );
  return 0;
}


static void
_recalculateSizes ( oaCamera* camera )
{
  unsigned int transferBlockSize;

  if ( cameraState.binMode == OA_BIN_MODE_2x2 ) {
    transferBlockSize = 234 * 1024;
  } else {
    transferBlockSize = 932 * 1024;
  }

  cameraState.frameSize = cameraState.xSize * cameraState.ySize * 2;
  cameraState.captureLength = cameraState.frameSize;
  cameraState.transferPadding = transferBlockSize -
      ( cameraState.frameSize % transferBlockSize );
}


static void
_QHY6InitFunctionPointers ( oaCamera* camera )
{
  // Set by QHYinit()
  // camera->funcs.initCamera = oaQHYInitCamera;
  camera->funcs.closeCamera = closeCamera;
  // camera->funcs.resetCamera = reset;
  // camera->funcs.startCapture = start;
  camera->funcs.startCaptureExtended = start;
  camera->funcs.stopCapture = stop;
  // stick with the device versions of these
  // camera->funcs.hasLoadableFirmware = _cameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = getName;
  camera->funcs.getControlRange = getControlRange;
  camera->funcs.readControl = readControl;
  camera->funcs.setControl = setControl;
  camera->funcs.testControl = testControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = startReadFrame;
  camera->funcs.readFrame = readFrame;
  camera->funcs.finishReadFrame = finishReadFrame;
  camera->funcs.enumerateFrameSizes = getFrameSizes;
  camera->funcs.getFramePixelFormat = getFramePixelFormat;
  // camera->funcs.testROISize = testROISize;
  // camera->funcs.getMenuString = _getMenuString;
  // camera->funcs.getAutoWBManualSetting = _getAutoWBManualSetting;
  // camera->funcs.hasFixedFrameRates = _hasFixedFrameRates;
}

