/*****************************************************************************
 *
 * QHYinit.c -- Initialise QHY cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>

#include <libusb-1.0/libusb.h>

#include "QHYoacam.h"
#include "QHY5.h"
#include "QHY6.h"
#include "QHY5LII.h"
#include "unimplemented.h"
#include "oacamprivate.h"


static void _QHYInitFunctionPointers ( oaCamera* );


/**
 * Initialise a given camera device
 */

oaCamera*
oaQHYInitCamera ( oaCameraDevice* device )
{
  oaCamera*				camera;
  int                   		i, matched, ret;
  int					deviceAddr, deviceBus, numUSBDevices;
  libusb_device**			devlist;
  libusb_device*			usbDevice;
  libusb_device_handle*			usbHandle = 0;
  struct libusb_device_descriptor	desc;


  switch ( device->_devType ) {
    case CAM_QHY5:
    case CAM_QHY6_ST4:
    case CAM_QHY5LII:
      break;
    default:
      fprintf ( stderr, "Unsupported camera %ld: %s\n", device->_devType,
          device->deviceName );
      return 0;
      break;
  }

  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    perror ( "malloc oaCamera failed" );
    return 0;
  }

  OA_CLEAR ( cameraState );
  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _oaInitCameraFunctionPointers ( camera );
  _QHYInitFunctionPointers ( camera );

  ( void ) strcpy ( camera->deviceName, device->deviceName );
  cameraState.initialised = 0;
  cameraState.index = -1;

  // FIX ME -- This is a bit ugly.  Much of it is repeated from the
  // getCameras function.  I should join the two together somehow.

  libusb_init ( &cameraState.usbContext );
  // libusb_set_debug ( cameraState.usbContext, LIBUSB_LOG_LEVEL_DEBUG );
  numUSBDevices = libusb_get_device_list ( cameraState.usbContext, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    if ( numUSBDevices ) {
      fprintf ( stderr, "Can't see any USB devices now (list returns -1)\n" );
      free ( camera );
      return 0;
    }
    fprintf ( stderr, "Can't see any USB devices now\n" );
    free ( camera );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  for ( i = 0; i < numUSBDevices && !matched; i++ ) {
    usbDevice = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( usbDevice, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( cameraState.usbContext );
      fprintf ( stderr, "get device descriptor failed\n" );
      free ( camera );
      return 0;
    }
    if ( desc.idVendor == cameraList[ device->_misc ].vendorId &&
        desc.idProduct == cameraList[ device->_misc ].productId &&
        libusb_get_bus_number ( usbDevice ) == deviceBus &&
        libusb_get_device_address ( usbDevice ) == deviceAddr ) {
      // this looks like the one!
      matched = 1;
      libusb_open ( usbDevice, &usbHandle );
    }
  }
  libusb_free_device_list ( devlist, 1 );
  if ( !matched ) {
    fprintf ( stderr, "No matching USB device found!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }
  if ( !usbHandle ) {
    fprintf ( stderr, "Unable to open USB device!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }

  if ( libusb_kernel_driver_active ( usbHandle, 0 )) {
      libusb_detach_kernel_driver( usbHandle, 0 );
  }

  if (( ret = libusb_set_configuration ( usbHandle, 1 ))) {
    fprintf ( stderr, "Can't get configuration for USB device! err = %d\n",
        ret );
    fprintf ( stderr, "Try unplugging and reconnecting the device?\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }
  if ( libusb_claim_interface ( usbHandle, 0 )) {
    fprintf ( stderr, "Unable to claim interface for USB device!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }

  if ( CAM_QHY5 == cameraState.cameraType ) {
    // FIX ME -- may not be required?
    if ( libusb_set_interface_alt_setting ( usbHandle, 0, 0 )) {
      fprintf ( stderr, "Unable to set alternate interface for USB device!\n" );
      libusb_release_interface ( cameraState.usbHandle, 0 );
      libusb_exit ( cameraState.usbContext );
      free ( camera );
      return 0;
    }
  }

  camera->interface = device->interface;
  cameraState.usbHandle = usbHandle;
  cameraState.index = device->_devIndex;
  cameraState.cameraType = device->_devType;
  cameraState.droppedFrames = 0;
  cameraState.isColour = device->_colour;

  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  pthread_mutex_init ( &cameraState.captureMutex, 0 );
  pthread_mutex_init ( &cameraState.usbMutex, 0 );
  pthread_cond_init ( &cameraState.frameAvailable, 0 );

  switch ( cameraState.cameraType ) {
    case CAM_QHY5:
      ret = _QHY5InitCamera ( camera );
      break;
    case CAM_QHY6_ST4:
      ret = _QHY6InitCamera ( camera );
      break;
    case CAM_QHY5LII:
      ret = _QHY5LIIInitCamera ( camera );
      break;
    default:
      fprintf ( stderr, "unsupported camera type\n" );
      libusb_release_interface ( cameraState.usbHandle, 0 );
      libusb_exit ( cameraState.usbContext );
      free ( camera );
      ret = -1;
      break;
  }
  if ( !ret ) {
    free ( camera );
    camera = 0;
  }

  return camera;
}


static void
_QHYInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaQHYInitCamera;
}
