/*****************************************************************************
 *
 * SXcontrol.c -- control functions for Starlight Xpress cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>

#include "oacamprivate.h"
#include "SXoacam.h"

#define	cameraState		camera->_sx

static void*    _capture ( void* );
static void     _startExposure ( oaCamera*, unsigned int );
static int      _readExposure ( oaCamera*, unsigned char*, unsigned int );
static int      _readTemperature ( oaCamera* );
static int	_clearFrame ( oaCamera*, unsigned int );
static int	_latchFrame ( oaCamera*, unsigned int, unsigned int,
		    unsigned int );
static int	_readFrame ( oaCamera*, unsigned char*, unsigned int );


int
oaSXCameraSetControl ( oaCamera* camera, int control, oaControlValue* val )
{
   // FIX ME -- change ranges & return OOB or unsupported control error

  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: control: %s ( %d, ? )\n",
      __FUNCTION__, control );

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      uint64_t exp;

      if ( val->valueType != OA_CTRL_TYPE_INT64 ) {
        fprintf ( stderr, "%s: invalid control type %d where int64 expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      exp = val->int64 / 1000;
      // Probably don't need to do anything else with this because we're
      // effectively free-running.  The camera waits for us to say that
      // the exposure is done.  But I'll set it just in case.
      cameraState.currentExposure = exp;
      _SXsetTimer ( camera, cameraState.currentExposure );
      break;
    }
    case OA_CAM_CTRL_BINNING:
      if ( val->valueType != OA_CTRL_TYPE_DISCRETE ) {
        fprintf ( stderr, "%s: invalid control type %d where discrete "
            "expected\n", __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      cameraState.requestedBinMode = val->discrete;
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return OA_ERR_NONE;
}


int
oaSXCameraReadControl ( oaCamera* camera, int control,
    oaControlValue* val )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: control: %s ( %d )\n",
      __FUNCTION__, control );

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val->valueType = OA_CTRL_TYPE_INT64;
      val->int64 = cameraState.currentExposure * 1000;
      break;

    case OA_CAM_CTRL_BINNING:
      val->valueType = OA_CTRL_TYPE_DISCRETE;
      val->discrete = cameraState.binMode;
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      val->valueType = OA_CTRL_TYPE_DISCRETE;
      val->discrete = cameraState.bytesPerPixel * 8;
      break;

    case OA_CAM_CTRL_TEMPERATURE:
    {
      val->valueType = OA_CTRL_TYPE_READONLY;
      val->readonly = _readTemperature ( camera );
      break;
    }

    case OA_CAM_CTRL_DROPPED:
      val->valueType = OA_CTRL_TYPE_READONLY;
      val->readonly = cameraState.droppedFrames;
      break;

    default:
      fprintf ( stderr,
          "Unrecognised control %d in %s\n", control, __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return OA_ERR_NONE;
}


int
oaSXCameraTestControl ( oaCamera* camera, int control,
    oaControlValue* val )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: control: %s ( %d, ? )\n",
      __FUNCTION__, control );

  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  if ( camera->controls [ control ] != val->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      int64_t val_s64 = val->int64;
      if ( val_s64 >= camera->min[ control ] &&
          val_s64 <= camera->max[ control ] &&
          ( 0 == ( val_s64 - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;
    }
    case OA_CAM_CTRL_BINNING:
    {
      int32_t val_s32 = val->discrete;
      if ( OA_BIN_MODE_NONE == val_s32 || OA_BIN_MODE_2x2 == val_s32 ) {
        return OA_ERR_NONE;
      }
      break;
    }
    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in oaSXCameraSetControl\n",
          control );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
oaSXCameraSetROI ( oaCamera* camera, int x, int y )
{
  cameraState.xSize = x;
  cameraState.ySize = y;
  return 0;
}


int
oaSXCameraStart ( oaCamera* camera, START_PARMS* parms )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: control: %s ([ %d, %d ])\n",
      __FUNCTION__, parms->size.x, parms->size.y );

  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.xSize = parms->size.x;
  cameraState.ySize = parms->size.y;
  cameraState.binMode = cameraState.requestedBinMode;

  cameraState.imageBufferLength = parms->size.x * parms->size.y *
      cameraState.bytesPerPixel;
/*
  if ( OA_BIN_MODE_2x2 == cameraState.binMode ) {
    cameraState.captureLength = cameraState.imageBufferLength / 2;
    cameraState.imageBufferLength /= 4;
  } else {
*/
    cameraState.captureLength = cameraState.imageBufferLength;
/*
  }
*/
  pthread_create ( &cameraState.captureThread, 0, &_capture, camera );
  cameraState.captureThreadStarted = 1;

  return 0;
}


int
oaSXCameraStop ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: command: %s ()\n", __FUNCTION__ );

  if ( !cameraState.captureThreadStarted ) {
    return 0;
  }

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.captureThreadExit = 1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  pthread_join ( cameraState.captureThread, 0 );
  cameraState.captureThreadStarted = 0;
  return 0;
}


// FIX ME -- should this return value be the other way around?
int
oaSXCameraReset ( oaCamera* camera )
{
  unsigned char	buff[ SXUSB_REQUEST_BUFSIZE ];
  int		ret, transferred;

  oacamDebugMsg ( DEBUG_CAM_CTRL, "SX: control: %s ()\n", __FUNCTION__ );

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_SEND;
  buff[ SXUSB_REQ_CMD ] = SXUSB_RESET;
  pthread_mutex_lock ( &cameraState.usbMutex );
  ret = libusb_bulk_transfer ( cameraState.usbHandle, SXUSB_BULK_ENDP_OUT,
      buff, SXUSB_REQUEST_BUFSIZE, &transferred, SXUSB_TIMEOUT );
  pthread_mutex_unlock ( &cameraState.usbMutex );
  if ( ret || transferred != SXUSB_REQUEST_BUFSIZE ) {
    fprintf ( stderr, "send RESET for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, SXUSB_REQUEST_BUFSIZE );
    return -1;
  }

  return 0;
}


static void*
_capture ( void* param )
{
  oaCamera*     camera = param;
  int           quitThread, nextBuffer;
  unsigned int  requestedTime;

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  nextBuffer = 0;

  while ( 1 ) {
    pthread_mutex_lock ( &cameraState.captureMutex );
    quitThread = cameraState.captureThreadExit;
    if ( nextBuffer == cameraState.nextBufferToRead ) {
      nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
    }
    requestedTime = cameraState.currentExposure * 1000;
    pthread_mutex_unlock ( &cameraState.captureMutex );
    if ( quitThread ) {
      break;
    }

    _startExposure ( camera, requestedTime );
    usleep ( requestedTime );
    _readExposure ( camera, cameraState.buffers[ nextBuffer ].start,
        cameraState.captureLength );

    pthread_mutex_lock ( &cameraState.captureMutex );
    cameraState.nextBufferToRead = nextBuffer;
    cameraState.haveDataToRead = 1;
    pthread_cond_signal ( &cameraState.frameAvailable );
    pthread_mutex_unlock ( &cameraState.captureMutex );
    nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;

  }
  return 0;
}


static void
_startExposure ( oaCamera* camera, unsigned int exposureTime )
{
  if ( cameraState.isInterlaced ) {
    _clearFrame ( camera, CCD_EXP_FLAGS_FIELD_EVEN );
    usleep ( 100 );
    _clearFrame ( camera, CCD_EXP_FLAGS_FIELD_ODD );
  } else {
    _clearFrame ( camera, CCD_EXP_FLAGS_FIELD_BOTH );
  }
}


static int
_readExposure ( oaCamera* camera, unsigned char* buffer,
    unsigned int readLength )
{
  unsigned char*	evenFrame;
  unsigned char*	oddFrame;
  unsigned char*	tgt;
  unsigned int		halfFrameSize, rowLength, i, numRows;

  _clearFrame ( camera, CCD_EXP_FLAGS_NOWIPE_FRAME );
 
  usleep ( 3000 );

  if ( cameraState.isInterlaced && OA_BIN_MODE_NONE == cameraState.binMode ) {

    halfFrameSize = readLength / 2;
    evenFrame = cameraState.xferBuffer;
    oddFrame = cameraState.xferBuffer + halfFrameSize;
    rowLength = cameraState.xSize * cameraState.bytesPerPixel;
    numRows = cameraState.ySize / 2;

    pthread_mutex_lock ( &cameraState.usbMutex );

    _latchFrame ( camera, CCD_EXP_FLAGS_FIELD_EVEN, cameraState.xSize,
        numRows );
    _readFrame ( camera, evenFrame, halfFrameSize );
    _latchFrame ( camera, CCD_EXP_FLAGS_FIELD_ODD, cameraState.xSize,
        numRows );
    _readFrame ( camera, oddFrame, halfFrameSize );

    pthread_mutex_unlock ( &cameraState.usbMutex );

    tgt = buffer;
    for ( i = 0; i < numRows; i++ ) {
      memcpy ( tgt, oddFrame, rowLength );
      tgt += rowLength;
      oddFrame += rowLength;
      memcpy ( tgt, evenFrame, rowLength );
      tgt += rowLength;
      evenFrame += rowLength;
    }
  } else {
    _latchFrame ( camera, CCD_EXP_FLAGS_FIELD_BOTH, cameraState.xSize,
        cameraState.ySize );
    _readFrame ( camera, buffer, readLength );
  }

  return 0;
}


static int
_readTemperature ( oaCamera* camera )
{
  int		temp, ret, transferred;
  unsigned char	buff [ SXUSB_REQUEST_BUFSIZE ];

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_REQUEST;
  buff[ SXUSB_REQ_CMD ] = SXUSB_COOLER_TEMPERATURE;
  buff[ SXUSB_REQ_LENGTH_L ] = 2;
  buff[ SXUSB_REQ_LENGTH_H ] = 0;
  pthread_mutex_lock ( &cameraState.usbMutex );

  if (( ret = libusb_bulk_transfer ( cameraState.usbHandle,
      SXUSB_BULK_ENDP_OUT, buff, SXUSB_REQUEST_BUFSIZE, &transferred,
      SXUSB_TIMEOUT )) || transferred != SXUSB_REQUEST_BUFSIZE ) {
    fprintf ( stderr, "request TEMPERATURE for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, SXUSB_REQUEST_BUFSIZE );
    pthread_mutex_unlock ( &cameraState.usbMutex );
    return 0;
  }
  if (( ret = libusb_bulk_transfer ( cameraState.usbHandle,
      SXUSB_BULK_ENDP_IN, buff, 2, &transferred,
      SXUSB_TIMEOUT )) || transferred != 2 ) {
    fprintf ( stderr, "request TEMPERATURE for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, 2 );
    pthread_mutex_unlock ( &cameraState.usbMutex );
    return 0;
  }
  pthread_mutex_unlock ( &cameraState.usbMutex );

  temp = ( buff[0] | ( buff[1] << 8 )) - 273;
  return temp;
}


static int
_clearFrame ( oaCamera* camera, unsigned int flags )
{
  unsigned char	buff [ SXUSB_REQUEST_BUFSIZE ];
  int		ret, transferred;

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_SEND;
  buff[ SXUSB_REQ_CMD ] = SXUSB_CLEAR_PIXELS;
  buff[ SXUSB_REQ_VALUE_L ] = flags & 0xff;
  buff[ SXUSB_REQ_VALUE_H ] = ( flags >> 8 ) & 0xff;
  pthread_mutex_lock ( &cameraState.usbMutex );
  ret = libusb_bulk_transfer ( cameraState.usbHandle, SXUSB_BULK_ENDP_OUT,
      buff, SXUSB_REQUEST_BUFSIZE, &transferred, SXUSB_TIMEOUT );
  pthread_mutex_unlock ( &cameraState.usbMutex );
  if ( ret || transferred != SXUSB_REQUEST_BUFSIZE ) {
    fprintf ( stderr, "send CLEAR PIXELS for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, SXUSB_REQUEST_BUFSIZE );
    return -1;
  }

  return 0;
}


static int
_latchFrame ( oaCamera* camera, unsigned int flags, unsigned int x,
    unsigned int y )
{
  unsigned char buff [ SXUSB_READ_BUFSIZE ];
  int           ret, transferred, xbin, ybin;

  if ( OA_BIN_MODE_2x2 == cameraState.binMode ) {
    xbin = 2;
    ybin = 1;
  } else {
    xbin = ybin = 1;
  }

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_SEND;
  buff[ SXUSB_REQ_CMD ] = SXUSB_READ_PIXELS;
  buff[ SXUSB_REQ_VALUE_L ] = flags & 0xff;
  buff[ SXUSB_REQ_VALUE_H ] = ( flags >> 8 ) & 0xff;
  buff[ SXUSB_REQ_INDEX_L ] = 0;
  buff[ SXUSB_REQ_INDEX_H ] = 0;
  buff[ SXUSB_REQ_LENGTH_L ] = 10;
  buff[ SXUSB_REQ_LENGTH_H ] = 0;
  buff[ SXUSB_REQ_DATA + 0] = 0;
  buff[ SXUSB_REQ_DATA + 1] = 0;
  buff[ SXUSB_REQ_DATA + 2] = 0;
  buff[ SXUSB_REQ_DATA + 3] = 0;
  buff[ SXUSB_REQ_DATA + 4] = x & 0xff;
  buff[ SXUSB_REQ_DATA + 5] = ( x >> 8 ) & 0xff;
  buff[ SXUSB_REQ_DATA + 6] = y & 0xff;
  buff[ SXUSB_REQ_DATA + 7] = ( y >> 8 ) & 0xff;
  buff[ SXUSB_REQ_DATA + 8] = xbin;
  buff[ SXUSB_REQ_DATA + 9] = ybin;

  // mutex is handled by the caller
  if (( ret = libusb_bulk_transfer ( cameraState.usbHandle,
      SXUSB_BULK_ENDP_OUT, buff, SXUSB_READ_BUFSIZE, &transferred,
      SXUSB_TIMEOUT )) || transferred != SXUSB_READ_BUFSIZE ) {
    fprintf ( stderr, "request READ for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, SXUSB_READ_BUFSIZE );
    return -1;
  }

  return 0;
}


static int
_readFrame ( oaCamera* camera, unsigned char* buffer, unsigned int length )
{
  unsigned int	transferred;
  int		ret;

  // mutex is handled by the caller
  if (( ret = libusb_bulk_transfer ( cameraState.usbHandle,
      SXUSB_BULK_ENDP_IN, buffer, length, &transferred,
      SXUSB_FRAME_TIMEOUT ))) {
    fprintf ( stderr, "receive READ for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, length );
    return -1;
  }
  if ( length != transferred ) {
    fprintf ( stderr, "length %d != transferred %d in %s\n", length,
        transferred, __FUNCTION__ );
    return -1;
  }
  return 0;
}


int
_SXsetTimer ( oaCamera* camera, unsigned int milliseconds )
{
  unsigned char buff [ SXUSB_TIMER_BUFSIZE ];
  int           ret, transferred;

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_SEND;
  buff[ SXUSB_REQ_CMD ] = SXUSB_SET_TIMER;
  buff[ SXUSB_REQ_LENGTH_L ] = 4;
  buff[ SXUSB_REQ_LENGTH_H ] = 0;
  buff[ SXUSB_REQ_DATA + 0] = milliseconds & 0xff;
  buff[ SXUSB_REQ_DATA + 1] = ( milliseconds >> 8 ) & 0xff;
  buff[ SXUSB_REQ_DATA + 2] = ( milliseconds >> 16 ) & 0xff;
  buff[ SXUSB_REQ_DATA + 3] = ( milliseconds >> 24 ) & 0xff;

  pthread_mutex_lock ( &cameraState.usbMutex );

  ret = libusb_bulk_transfer ( cameraState.usbHandle, SXUSB_BULK_ENDP_OUT,
      buff, SXUSB_TIMER_BUFSIZE, &transferred, SXUSB_TIMEOUT );
  pthread_mutex_unlock ( &cameraState.usbMutex );
  if ( ret || transferred != SXUSB_TIMER_BUFSIZE ) {
    fprintf ( stderr, "send TIMER for SX failed: ret = %d, "
        "transferred = %d of %d\n", ret, transferred, SXUSB_TIMER_BUFSIZE );
    return -1;
  }

  return 0;
}
