/*****************************************************************************
 *
 * SXinit.c -- Initialise Starlight Xpress cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>

#include <libusb-1.0/libusb.h>

#include "SXoacam.h"
#include "unimplemented.h"
#include "oacamprivate.h"

#define	cameraState		camera->_sx
#define DEFAULT_EXPOSURE	100
#define SX_BUFFERS		5

static void _SXInitFunctionPointers ( oaCamera* );


/**
 * Initialise a given camera device
 */

oaCamera*
oaSXInitCamera ( oaCameraDevice* device )
{
  oaCamera*				camera;
  int                   		i, j, matched, ret, transferred;
  int					deviceAddr, deviceBus, numUSBDevices;
  libusb_device**			devlist;
  libusb_device*			usbDevice;
  libusb_device_handle*			usbHandle = 0;
  struct libusb_device_descriptor	desc;
  unsigned char				buff[ SXUSB_GET_CCD_BUFSIZE ];
  uint32_t				extraCaps, cameraModel;

  switch ( device->_devType ) {
    case CAM_LODESTAR:
    case CAM_COSTAR:
      break;
    default:
      fprintf ( stderr, "Unsupported camera %ld: %s\n", device->_devType,
          device->deviceName );
      return 0;
      break;
  }

  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    perror ( "malloc oaCamera failed" );
    return 0;
  }

  OA_CLEAR ( cameraState );
  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );
  _oaInitCameraFunctionPointers ( camera );
  _SXInitFunctionPointers ( camera );

  ( void ) strcpy ( camera->deviceName, device->deviceName );
  cameraState.initialised = 0;
  cameraState.index = -1;

  // FIX ME -- This is a bit ugly.  Much of it is repeated from the
  // getCameras function.  I should join the two together somehow.

  libusb_init ( &cameraState.usbContext );
  // libusb_set_debug ( cameraState.usbContext, LIBUSB_LOG_LEVEL_DEBUG );
  numUSBDevices = libusb_get_device_list ( cameraState.usbContext, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    if ( numUSBDevices ) {
      fprintf ( stderr, "Can't see any USB devices now (list returns -1)\n" );
      free ( camera );
      return 0;
    }
    fprintf ( stderr, "Can't see any USB devices now\n" );
    free ( camera );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  for ( i = 0; i < numUSBDevices && !matched; i++ ) {
    usbDevice = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( usbDevice, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( cameraState.usbContext );
      fprintf ( stderr, "get device descriptor failed\n" );
      free ( camera );
      return 0;
    }
    if ( desc.idVendor == SXCameraList[ device->_misc ].vendorId &&
        desc.idProduct == SXCameraList[ device->_misc ].productId &&
        libusb_get_bus_number ( usbDevice ) == deviceBus &&
        libusb_get_device_address ( usbDevice ) == deviceAddr ) {
      // this looks like the one!
      matched = 1;
      libusb_open ( usbDevice, &usbHandle );
    }
  }
  libusb_free_device_list ( devlist, 1 );
  if ( !matched ) {
    fprintf ( stderr, "No matching USB device found!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }
  if ( !usbHandle ) {
    fprintf ( stderr, "Unable to open USB device!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }

  if ( libusb_kernel_driver_active ( usbHandle, 1 )) {
    libusb_detach_kernel_driver( usbHandle, 1 );
  }

  if ( libusb_claim_interface ( usbHandle, 1 )) {
    fprintf ( stderr, "Unable to claim interface for USB device!\n" );
    libusb_exit ( cameraState.usbContext );
    free ( camera );
    return 0;
  }

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_REQUEST;
  buff[ SXUSB_REQ_CMD ] = SXUSB_GET_CCD;
  buff[ SXUSB_REQ_LENGTH_L ] = SXUSB_GET_CCD_BUFSIZE;
  if (( ret = libusb_bulk_transfer ( usbHandle, SXUSB_BULK_ENDP_OUT, buff,
      SXUSB_REQUEST_BUFSIZE, &transferred, SXUSB_TIMEOUT ))) {
    fprintf ( stderr, "request GET_CCD for SX failed: %d\n", ret );
    libusb_release_interface ( usbHandle, 1 );
    libusb_close ( usbHandle );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    return 0;
  }
  if (( ret = libusb_bulk_transfer ( usbHandle, SXUSB_BULK_ENDP_IN, buff,
      SXUSB_GET_CCD_BUFSIZE, &transferred, SXUSB_TIMEOUT ))) {
    fprintf ( stderr, "request GET_CCD for SX failed: %d\n", ret );
    libusb_release_interface ( usbHandle, 1 );
    libusb_close ( usbHandle );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    return 0;
  }

  cameraState.horizontalFrontPorch = buff[0];
  cameraState.horizontalBackPorch = buff[1];
  cameraState.maxResolutionX = buff[2] | ( buff[3] << 8 );
  cameraState.verticalFrontPorch = buff[4];
  cameraState.verticalBackPorch = buff[5];
  cameraState.maxResolutionY = buff[6] | ( buff[7] << 8 );
  cameraState.colourMatrix = buff[12] | ( buff[13] << 8 );
  cameraState.bitDepth = buff[14];
  extraCaps = buff[16];

  OA_CLEAR( buff );
  buff[ SXUSB_REQ_CMD_TYPE ] = SXUSB_CMD_REQUEST;
  buff[ SXUSB_REQ_CMD ] = SXUSB_CAMERA_MODEL;
  buff[ SXUSB_REQ_LENGTH_L ] = SXUSB_CAMERA_MODEL_BUFSIZE;
  if (( ret = libusb_bulk_transfer ( usbHandle, SXUSB_BULK_ENDP_OUT, buff,
      SXUSB_REQUEST_BUFSIZE, &transferred, SXUSB_TIMEOUT ))) {
    fprintf ( stderr, "request GET_CCD for SX failed: %d\n", ret );
    libusb_release_interface ( usbHandle, 1 );
    libusb_close ( usbHandle );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    return 0;
  }
  if (( ret = libusb_bulk_transfer ( usbHandle, SXUSB_BULK_ENDP_IN, buff,
      SXUSB_CAMERA_MODEL_BUFSIZE, &transferred, SXUSB_TIMEOUT ))) {
    fprintf ( stderr, "request GET_CCD for SX failed: %d\n", ret );
    libusb_release_interface ( usbHandle, 1 );
    libusb_close ( usbHandle );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    return 0;
  }

  cameraModel = buff[0] | ( buff[1] << 8 );
  cameraState.isInterlaced = ( cameraModel & SX_MODEL_MASK_INTERLACE ) ? 1 : 0;

  camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CTRL_TYPE_INT64;
  camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1000;
  camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 0xffffffff * 1000;
  camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1000;
  camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = DEFAULT_EXPOSURE * 1000;

  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED_RESET ] = OA_CTRL_TYPE_BUTTON;

  // camera->controls[ OA_CAM_CTRL_BINNING ] = OA_CTRL_TYPE_DISCRETE;

  if ( extraCaps & SXUSB_CAPS_COOLER ) {
    // These are just made up as I have no documentation
    camera->controls[ OA_CAM_CTRL_TEMP_SETPOINT ] = OA_CTRL_TYPE_INT32;
    camera->min[ OA_CAM_CTRL_TEMP_SETPOINT ] = -320;
    camera->max[ OA_CAM_CTRL_TEMP_SETPOINT ] = 10000;
    camera->step[ OA_CAM_CTRL_TEMP_SETPOINT ] = 1;
    camera->def[ OA_CAM_CTRL_TEMP_SETPOINT ] = 0;

    camera->controls[ OA_CAM_CTRL_TEMPERATURE ] = OA_CTRL_TYPE_READONLY;
  }

  camera->interface = device->interface;
  cameraState.usbHandle = usbHandle;
  cameraState.index = device->_devIndex;
  cameraState.cameraType = device->_devType;
  cameraState.droppedFrames = 0;
  cameraState.isColour = device->_colour;

  if ( cameraState.bitDepth <= 8 ) {
    cameraState.videoCurrent = OA_PIX_FMT_GREY8;
    cameraState.videoGrey = 1;
    cameraState.videoGrey16 = 0;
    cameraState.bytesPerPixel = 1;
  } else {
    cameraState.videoCurrent = OA_PIX_FMT_GREY16LE;
    cameraState.videoGrey16 = 1;
    cameraState.videoGrey = 0;
    cameraState.bytesPerPixel = 2;
  }
  cameraState.binMode = cameraState.requestedBinMode = OA_BIN_MODE_NONE;

  camera->features.hasReset = 1;

  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;

  // If the camera is interlaced it reports half the number of lines for
  // the full image, so multiply by two
  if ( cameraState.isInterlaced ) {
    cameraState.maxResolutionY *= 2;
  }
  cameraState.captureLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY * cameraState.bytesPerPixel;
  if (!( cameraState.xferBuffer = malloc ( cameraState.captureLength ))) {
    fprintf ( stderr, "malloc of transfer buffer failed in %s\n",
        __FUNCTION__ );
    return 0;
  }

  cameraState.imageBufferLength = cameraState.captureLength;
  cameraState.buffers = calloc ( SX_BUFFERS, sizeof ( struct SXbuffer ));
  for ( i = 0; i < SX_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      free ( cameraState.xferBuffer );
      return 0;
    }
  }

  cameraState.currentExposure = DEFAULT_EXPOSURE;

  cameraState.resolutions[0].x = cameraState.maxResolutionX;
  cameraState.resolutions[0].y = cameraState.maxResolutionY;
  cameraState.resolutions[1].x = 0;
  cameraState.resolutions[1].y = 0;
  cameraState.binMode2Resolutions[0].x = cameraState.maxResolutionX / 2;
  cameraState.binMode2Resolutions[0].y = cameraState.maxResolutionY / 2;
  cameraState.binMode2Resolutions[1].x = 0;
  cameraState.binMode2Resolutions[1].y = 0;

  cameraState.xSize = cameraState.maxResolutionX;
  cameraState.ySize = cameraState.maxResolutionY;

  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  pthread_mutex_init ( &cameraState.captureMutex, 0 );
  pthread_mutex_init ( &cameraState.usbMutex, 0 );
  pthread_cond_init ( &cameraState.frameAvailable, 0 );

  _SXsetTimer ( camera, DEFAULT_EXPOSURE );

  return camera;
}


static void
_SXInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaSXInitCamera;
  camera->funcs.closeCamera = oaSXCloseCamera;
  camera->funcs.resetCamera = oaSXCameraReset;
  // camera->funcs.startCapture = oaSXCameraStart;
  camera->funcs.startCaptureExtended = oaSXCameraStart;
  camera->funcs.stopCapture = oaSXCameraStop;
  camera->funcs.hasLoadableFirmware = oaSXCameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = oaSXCameraGetName;
  camera->funcs.getControlRange = oaSXCameraGetControlRange;
  camera->funcs.readControl = oaSXCameraReadControl;
  camera->funcs.setControl = oaSXCameraSetControl;
  camera->funcs.testControl = oaSXCameraTestControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = oaSXCameraStartReadFrame;
  camera->funcs.readFrame = oaSXCameraReadFrame;
  camera->funcs.finishReadFrame = oaSXCameraFinishReadFrame;
  camera->funcs.enumerateFrameSizes = oaSXCameraGetFrameSizes;
  // camera->funcs.enumerateFrameRates = _enumerateFrameRates;
  camera->funcs.getFramePixelFormat = oaSXCameraGetFramePixelFormat;
/*
  camera->funcs.testROISize = oaSXCameraTestROISize;
  // camera->funcs.getMenuString = _getMenuString;
  // camera->funcs.getAutoWBManualSetting = _getAutoWBManualSetting;
  // camera->funcs.hasFixedFrameRates = _hasFixedFrameRates;
*/
}
