/*****************************************************************************
 *
 * UVCcontrol.c -- control functions for UVC cameras
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBUVC

#include <openastro/camera.h>

#include "UVCoacam.h"

#define	cameraState		camera->_uvc


int
oaUVCCameraSetControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  int found, len;
  uint8_t uvcControl;
  int32_t val_s32;
  int64_t val_s64;
  unsigned int i;

  for ( i = 0, found = 0; !found && i < numPUControls; i++ ) {
    if ( controlData[i].oaControl == control ) {
      uvcControl = controlData[i].uvcControl;
      len = controlData[i].size;
      found = 1;
      // make all non-zero boolean values 1
      if ( OA_CTRL_TYPE_BOOLEAN == controlData[i].oaControlType ) {
        val_s32 = valp->boolean ? 1 : 0;
      } else {
        val_s32 = valp->int32;
      }
    }
  }

  if ( found ) {
    switch ( control ) {

      case OA_CAM_CTRL_BRIGHTNESS:
      case OA_CAM_CTRL_CONTRAST:
      case OA_CAM_CTRL_SATURATION:
      case OA_CAM_CTRL_HUE:
      case OA_CAM_CTRL_GAMMA:
      case OA_CAM_CTRL_GAIN:
      case OA_CAM_CTRL_HUE_AUTO:
      case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
      case OA_CAM_CTRL_SHARPNESS:
      case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
      case OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP:
        return setUVCControl ( cameraState.uvcHandle, cameraState.unitId,
            uvcControl, len, val_s32 );
        break;

      default:
        fprintf ( stderr, "Unrecognised control %d in %s\n", control,
            __FUNCTION__ );
        return -OA_ERR_INVALID_CONTROL;
        break;
    }
  }

  // Now handle the ones that are not in a processing unit, and white
  // component which is something of a special case

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val_s64 = valp->int64 / 1000;
      val_s32 = val_s64 & 0xffffffff;
      if ( uvc_set_exposure_abs ( cameraState.uvcHandle, val_s32 )
          != UVC_SUCCESS ) {
        fprintf ( stderr, "uvc_get_exposure_abs() failed in %s\n",
            __FUNCTION__ );
      }
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
    {
      uint8_t data;

      val_s32 = valp->menu;
      switch ( val_s32 ) {
        case OA_EXPOSURE_AUTO:
          data = 2; break;
        case OA_EXPOSURE_MANUAL:
          data = 1; break;
        case OA_EXPOSURE_SHUTTER_PRIORITY:
          data = 4; break;
        case OA_EXPOSURE_APERTURE_PRIORITY:
          data = 8; break;
      }
      if ( val_s32 < OA_EXPOSURE_TYPE_LAST_P1 ) {
        if ( uvc_set_ae_mode ( cameraState.uvcHandle, data )
            != UVC_SUCCESS ) {
          fprintf ( stderr, "uvc_set_ae_mode( %d ) failed in %s\n", val_s32,
              __FUNCTION__ );
        }
      } else {
        fprintf ( stderr, "uvc_set_ae_mode() value %d invalid in %s\n",
            val_s32, __FUNCTION__ );
      }
      break;
    }

    case OA_CAM_CTRL_BLUE_BALANCE:
      cameraState.componentBalance = ( cameraState.componentBalance &
          ~0xffff ) | ( val_s32 & 0xffff );
        return setUVCControl ( cameraState.uvcHandle, cameraState.unitId,
            UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, 4,
            cameraState.componentBalance );

    case OA_CAM_CTRL_RED_BALANCE:
      cameraState.componentBalance = ( cameraState.componentBalance &
          0xffff ) | (( val_s32 & 0xffff ) <<  16 );
        return setUVCControl ( cameraState.uvcHandle, cameraState.unitId,
            UVC_PU_WHITE_BALANCE_COMPONENT_CONTROL, 4,
            cameraState.componentBalance );

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      break;
  }

  return -OA_ERR_INVALID_CONTROL;
}


int
oaUVCCameraReadControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  int found, len;
  uint8_t uvcControl;
  unsigned int i;

  for ( i = 0, found = 0; !found && i < numPUControls; i++ ) {
    if ( controlData[i].oaControl == control ) {
      uvcControl = controlData[i].uvcControl;
      len = controlData[i].size;
      found = 1;
    }
  }

  if ( found ) {
    switch ( control ) {

      case OA_CAM_CTRL_BRIGHTNESS:
      case OA_CAM_CTRL_CONTRAST:
      case OA_CAM_CTRL_SATURATION:
      case OA_CAM_CTRL_HUE:
      case OA_CAM_CTRL_GAMMA:
      case OA_CAM_CTRL_GAIN:
      case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
      case OA_CAM_CTRL_SHARPNESS:
        valp->valueType = OA_CTRL_TYPE_INT32;
        valp->int32 = getUVCControl ( cameraState.uvcHandle,
            cameraState.unitId, uvcControl, len, UVC_GET_CUR );
        break;
      case OA_CAM_CTRL_HUE_AUTO:
      case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
      case OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP:
        valp->valueType = OA_CTRL_TYPE_BOOLEAN;
        valp->boolean = getUVCControl ( cameraState.uvcHandle,
            cameraState.unitId, uvcControl, len, UVC_GET_CUR );
        break;

      default:
        fprintf ( stderr, "Unrecognised control %d in %s\n", control,
            __FUNCTION__ );
        return -OA_ERR_INVALID_CONTROL;
        break;
    }
    return OA_ERR_NONE;
  }

  // Now handle the ones that are not in a processing unit...

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      uint32_t data32;
      int64_t data64;

      if ( uvc_get_exposure_abs ( cameraState.uvcHandle, &data32, UVC_GET_CUR )
          != UVC_SUCCESS ) {
        fprintf ( stderr, "uvc_get_exposure_abs() failed in %s\n",
            __FUNCTION__ );
      }
      data64 = data32;
      data64 *= 1000; // convert msec to usec
      valp->valueType = OA_CTRL_TYPE_INT64;
      valp->int64 = data64;
      break;
    }
    case OA_CAM_CTRL_AUTO_EXPOSURE:
    {
      uint8_t data;

      if ( uvc_get_ae_mode ( cameraState.uvcHandle, &data, UVC_GET_CUR )
          != UVC_SUCCESS ) {
        fprintf ( stderr, "uvc_get_ae_mode() failed in %s\n",
            __FUNCTION__ );
      }
      switch ( data ) {
        case 2:
          data = OA_EXPOSURE_MANUAL;
          break;
        case 1:
          data = OA_EXPOSURE_AUTO;
          break;
        case 4:
          data = OA_EXPOSURE_SHUTTER_PRIORITY;
          break;
        case 8:
          data = OA_EXPOSURE_APERTURE_PRIORITY;
          break;
        default:
          fprintf ( stderr,
              "uvc_get_ae_mode() returned unexpected value %d in %s\n",
              data, __FUNCTION__ );
          break;
      }
      valp->valueType = OA_CTRL_TYPE_MENU;
      valp->menu = data;
    }
    case OA_CAM_CTRL_BLUE_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = cameraState.componentBalance & 0xffff;
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = cameraState.componentBalance >> 16;
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  return OA_ERR_NONE;
}


int
oaUVCCameraTestControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  int32_t val_s32;
  int64_t val_s64;

  // FIX ME -- this is horribly wrong.  Should be done by control type
  // rather than by the actual control itself

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
    case OA_CAM_CTRL_CONTRAST:
    case OA_CAM_CTRL_SATURATION:
    case OA_CAM_CTRL_HUE:
    case OA_CAM_CTRL_SHARPNESS:
    case OA_CAM_CTRL_GAMMA:
    case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_GAIN:
      val_s32 = valp->int32;
      if ( val_s32 >= camera->min[ control ] &&
          val_s32 <= camera->max[ control ] &&
          ( 0 == ( val_s32 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val_s64 = valp->int64;
      if ( val_s64 >= camera->min[ control ] &&
          val_s64 <= camera->max[ control ] &&
          ( 0 == ( val_s64 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
    case OA_CAM_CTRL_AUTO_WHITE_BALANCE_TEMP:
    case OA_CAM_CTRL_HUE_AUTO:
      return OA_ERR_NONE;
      break;

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      val_s32 = valp->menu;
      if ( val_s32 >= 0 && val_s32 < OA_EXPOSURE_TYPE_LAST_P1 ) {
        return OA_ERR_NONE;
      }
      break;     

    case OA_CAM_CTRL_BIT_DEPTH:
    case OA_CAM_CTRL_BINNING:
      return -OA_ERR_INVALID_CONTROL;
      break;

    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
oaUVCCameraSetROI ( oaCamera* camera, int x, int y )
{
  fprintf ( stderr, "attempt to set ROI of %dx%d in %s\n", x, y,
      __FUNCTION__ );
  return 0;
}


int
oaUVCCameraStart ( oaCamera* camera, START_PARMS* parms )
{
  const uvc_frame_desc_t*	frame;
  int				matched, res, format, multiplier;

  // Need to pick appropriate resolution out of current frame format
  // entry and kick off streaming with that.

  frame = cameraState.videoCurrent->frame_descs;
  matched = 0;
  do {
    if ( frame->wWidth == parms->size.x && frame->wHeight == parms->size.y ) {
      matched = 1;
    } else {
      frame = frame->next;
    }
  } while ( !matched && frame );

  if ( !matched ) {
    fprintf ( stderr, "Requested image size %dx%d not available\n",
        parms->size.x, parms->size.y );
    return -1;
  }

  cameraState.lastUsedBuffer = -1;

  camera->features.frameRates = frame->bFrameIntervalType ? 1 : 0;
  cameraState.frameInterval = frame->dwDefaultFrameInterval;

  multiplier = 1;
  // FIX ME -- work out frame format nicely here
  format = UVC_FRAME_FORMAT_GRAY8;
  if ( cameraState.isColour ) {
    if ( cameraState.videoRaw ) {
      format = cameraState.videoRaw;
    } else {
      format = UVC_FRAME_FORMAT_YUYV;
      multiplier = 2;
    }
  }

  cameraState.imageBufferLength = parms->size.x * parms->size.y * multiplier;

  res = uvc_get_stream_ctrl_format_size ( cameraState.uvcHandle,
      &cameraState.streamControl, format, parms->size.x, parms->size.y,
      parms->rate.denominator / parms->rate.numerator );

  if ( res < 0 ) {
    fprintf ( stderr, "uvc_get_stream_ctrl_format_size returned %d\n", res );
    return -1;
  }

  if (( res = uvc_start_streaming ( cameraState.uvcHandle,
      &cameraState.streamControl, frameCallback, camera, 0 )) < 0 ) {
    fprintf ( stderr, "uvc_start_streaming returned %d\n", res );
    return -1;
  }

  return 0;
}


int
oaUVCCameraStop ( oaCamera* camera )
{
  uvc_stop_streaming ( cameraState.uvcHandle );
  return 0;
}


int
oaUVCCameraReset ( oaCamera* camera )
{
  // uvc_device_t* dev = uvc_get_device ( cameraState.uvcHandle );
  uvc_stop_streaming ( cameraState.uvcHandle );
  // uvc_close ( cameraState.uvcHandle );
  // uvc_open ( dev, &cameraState.uvcHandle );
  // now set up all the formats and cameraState.videoCurrent again...
  return 0;
}


const char*
oaUVCCameraGetMenuString ( oaCamera* camera, int control, int index )
{
  if ( control != OA_CAM_CTRL_AUTO_EXPOSURE ) {
    fprintf ( stderr, "%s: control not implemented\n", __FUNCTION__ );
    return "";
  }

  switch ( index ) {
    case 1:
      return "Manual";
      break;
    case 2:
      return "Auto";
      break;
    case 4:
      return "Shutter Priority";
      break;
    case 8:
      return "Aperture Priority";
      break;
  }

  return "Unknown";
}

#endif /* HAVE_LIBUVC */
