/*****************************************************************************
 *
 * V4L2control.c -- control functions for V4L2 cameras
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#if HAVE_LIBV4L2

#include <openastro/camera.h>

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <sys/mman.h>
#include <libv4l2.h>

#include "V4L2oacam.h"
#include "v4l2ioctl.h"
#include "v4l2cameras.h"
#include "oacamprivate.h"

#define	cameraState		camera->_v4l2


int	_setUserControl ( int, int, int );
int	_getUserControl ( int, int );
int	_setExtendedControl ( int, int, oaControlValue* );
int	_getExtendedControl ( int, int, oaControlValue* );
int	_startCamera ( oaCamera*, int, int );


int
oaV4L2CameraSetControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  int32_t val_s32;

  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }
  if ( camera->controls [ control ] != valp->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  // FIX ME -- make a map between control names to simplify this switch
  // statement

  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_BRIGHTNESS, val_s32 );
      break;

    case OA_CAM_CTRL_CONTRAST:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_CONTRAST, val_s32 );
      break;

    case OA_CAM_CTRL_SATURATION:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_SATURATION, val_s32 );
      break;

    case OA_CAM_CTRL_HUE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_HUE, val_s32 );
      break;

    case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
      // These are the only two we currently allow in init()
      if ( camera->controls[ OA_CAM_CTRL_AUTO_WHITE_BALANCE ] ==
          OA_CTRL_TYPE_MENU ) {
        val_s32 = valp->menu;
      } else {
        val_s32 = valp->boolean;
      }
      return _setUserControl ( cameraState.fd, V4L2_CID_AUTO_WHITE_BALANCE,
          val_s32 );
      break;

    case OA_CAM_CTRL_WHITE_BALANCE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_DO_WHITE_BALANCE,
          val_s32 );
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_BLUE_BALANCE, val_s32 );
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_RED_BALANCE, val_s32 );
      break;

    case OA_CAM_CTRL_GAMMA:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_GAMMA, val_s32 );
      break;

    case OA_CAM_CTRL_EXPOSURE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_EXPOSURE, val_s32 );
      break;

    case OA_CAM_CTRL_AUTO_GAIN:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_AUTOGAIN, val_s32 );
      break;

    case OA_CAM_CTRL_GAIN:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_GAIN, val_s32 );
      break;

    case OA_CAM_CTRL_HFLIP:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_HFLIP, val_s32 );
      break;

    case OA_CAM_CTRL_VFLIP:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_VFLIP, val_s32 );
      break;

    case OA_CAM_CTRL_POWER_LINE_FREQ:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_POWER_LINE_FREQUENCY,
          val_s32 );
      break;

    case OA_CAM_CTRL_HUE_AUTO:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_HUE_AUTO, val_s32 );
      break;

    case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd,
          V4L2_CID_WHITE_BALANCE_TEMPERATURE, val_s32 );
      break;

    case OA_CAM_CTRL_SHARPNESS:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_SHARPNESS, val_s32 );
      break;

    case OA_CAM_CTRL_BACKLIGHT_COMPENSATION:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd,
          V4L2_CID_BACKLIGHT_COMPENSATION, val_s32 );
      break;

    case OA_CAM_CTRL_CHROMA_AGC:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_CHROMA_AGC, val_s32 );
      break;

    case OA_CAM_CTRL_COLOUR_KILLER:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_COLOR_KILLER, val_s32 );
      break;

    case OA_CAM_CTRL_COLOURFX:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_COLORFX, val_s32 );
      break;

    case OA_CAM_CTRL_AUTO_BRIGHTNESS:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_BRIGHTNESS, val_s32 );
      break;

    case OA_CAM_CTRL_BAND_STOP_FILTER:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_BAND_STOP_FILTER,
          val_s32 );
      break;

    case OA_CAM_CTRL_ROTATE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_ROTATE, val_s32 );
      break;

    case OA_CAM_CTRL_BG_COLOUR:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_BG_COLOR, val_s32 );
      break;

    case OA_CAM_CTRL_CHROMA_GAIN:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_CHROMA_GAIN, val_s32 );
      break;

    case OA_CAM_CTRL_MIN_BUFFERS_FOR_CAPTURE:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd,
          V4L2_CID_MIN_BUFFERS_FOR_CAPTURE, val_s32 );
      break;

#ifdef V4L2_CID_ALPHA_COMPONENT
    case OA_CAM_CTRL_ALPHA_COMPONENT:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_ALPHA_COMPONENT,
          val_s32 );
      break;
#endif

#ifdef V4L2_CID_COLORFX_CBCR
    case OA_CAM_CTRL_COLOURFX_CBCR:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, V4L2_CID_COLORFX_CBCR, val_s32 );
      break;
#endif

    // end of the standard V4L2 controls.  Now the extended ones

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_EXPOSURE_AUTO,
          valp );
      break;     

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      // have to convert from usec to units of 100 usec here, but we don't
      // know which one we're using
      oaControlValue newVal;
      if (( newVal.valueType = valp->valueType == OA_CTRL_TYPE_INT32 )) {
        newVal.int32 = valp->int32 / 100;
      } else {
        newVal.int64 = valp->int64 / 100;
      }
      return _setExtendedControl ( cameraState.fd, V4L2_CID_EXPOSURE_ABSOLUTE,
        &newVal );
      break;
    }
    case OA_CAM_CTRL_PAN_RELATIVE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_PAN_RELATIVE,
        valp );
      break;

    case OA_CAM_CTRL_TILT_RELATIVE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_TILT_RELATIVE,
        valp );
      break;

    case OA_CAM_CTRL_PAN_RESET:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_PAN_RESET,
        valp );
      break;

    case OA_CAM_CTRL_TILT_RESET:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_TILT_RESET,
        valp );
      break;

    case OA_CAM_CTRL_PAN_ABSOLUTE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_PAN_ABSOLUTE,
        valp );
      break;

    case OA_CAM_CTRL_TILT_ABSOLUTE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_TILT_ABSOLUTE,
        valp );
      break;

    case OA_CAM_CTRL_ZOOM_ABSOLUTE:
      return _setExtendedControl ( cameraState.fd, V4L2_CID_ZOOM_ABSOLUTE,
        valp );
      break;

    case OA_CAM_CTRL_AUTO_CONTOUR:
      val_s32 = valp->boolean;
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(autocontour),
          val_s32 );
      break;

    case OA_CAM_CTRL_CONTOUR:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(contour),
          val_s32 );
      break;

    case OA_CAM_CTRL_NOISE_REDUCTION:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(noise_reduction),
          val_s32 );
      break;

    case OA_CAM_CTRL_AUTO_WB_SPEED:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(awb_speed),
          val_s32 );
      break;

    case OA_CAM_CTRL_AUTO_WB_DELAY:
      val_s32 = valp->int32;
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(awb_delay),
          val_s32 );
      break;

    case OA_CAM_CTRL_SAVE_USER:
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(save_user), 1 );
      break;

    case OA_CAM_CTRL_RESTORE_USER:
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(restore_user),
          1 );
      break;

    case OA_CAM_CTRL_RESTORE_FACTORY:
      return _setUserControl ( cameraState.fd, PWC_CID_CUSTOM(restore_factory),
          1 );
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return -OA_ERR_INVALID_CONTROL;
}


int
oaV4L2CameraReadControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  switch ( control ) {

    case OA_CAM_CTRL_BRIGHTNESS:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_BRIGHTNESS );
      break;

    case OA_CAM_CTRL_CONTRAST:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_CONTRAST );
      break;

    case OA_CAM_CTRL_SATURATION:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_SATURATION );
      break;

    case OA_CAM_CTRL_HUE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_HUE );
      break;

    case OA_CAM_CTRL_AUTO_WHITE_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_AUTO_WHITE_BALANCE );
      break;

    case OA_CAM_CTRL_WHITE_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_DO_WHITE_BALANCE );
      break;

    case OA_CAM_CTRL_BLUE_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_BLUE_BALANCE );
      break;

    case OA_CAM_CTRL_RED_BALANCE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_RED_BALANCE );
      break;

    case OA_CAM_CTRL_GAMMA:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_GAMMA );
      break;

    case OA_CAM_CTRL_EXPOSURE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_EXPOSURE );
      break;

    case OA_CAM_CTRL_AUTO_GAIN:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_AUTOGAIN );
      break;

    case OA_CAM_CTRL_GAIN:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_GAIN );
      break;

    case OA_CAM_CTRL_HFLIP:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_HFLIP );
      break;

    case OA_CAM_CTRL_VFLIP:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_VFLIP );
      break;

    case OA_CAM_CTRL_POWER_LINE_FREQ:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_POWER_LINE_FREQUENCY );
      break;

    case OA_CAM_CTRL_HUE_AUTO:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_HUE_AUTO );
      break;

    case OA_CAM_CTRL_WHITE_BALANCE_TEMP:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_WHITE_BALANCE_TEMPERATURE );
      break;

    case OA_CAM_CTRL_SHARPNESS:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_SHARPNESS );
      break;

    case OA_CAM_CTRL_BACKLIGHT_COMPENSATION:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_BACKLIGHT_COMPENSATION );
      break;

    case OA_CAM_CTRL_CHROMA_AGC:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_CHROMA_AGC );
      break;

    case OA_CAM_CTRL_COLOUR_KILLER:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_COLOR_KILLER );
      break;

    case OA_CAM_CTRL_COLOURFX:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_COLORFX );
      break;

    case OA_CAM_CTRL_AUTO_BRIGHTNESS:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_BRIGHTNESS );
      break;

    case OA_CAM_CTRL_BAND_STOP_FILTER:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_BAND_STOP_FILTER );
      break;

    case OA_CAM_CTRL_ROTATE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_ROTATE );
      break;

    case OA_CAM_CTRL_BG_COLOUR:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_BG_COLOR );
      break;

    case OA_CAM_CTRL_CHROMA_GAIN:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_CHROMA_GAIN );
      break;

    case OA_CAM_CTRL_MIN_BUFFERS_FOR_CAPTURE:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_MIN_BUFFERS_FOR_CAPTURE );
      break;

#ifdef V4L2_CID_ALPHA_COMPONENT
    case OA_CAM_CTRL_ALPHA_COMPONENT:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          V4L2_CID_ALPHA_COMPONENT );
      break;
#endif

#ifdef V4L2_CID_COLORFX_CBCR
    case OA_CAM_CTRL_COLOURFX_CBCR:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, V4L2_CID_COLORFX_CBCR );
      break;
#endif

    // end of the standard V4L2 controls.  Now the extended ones

    case OA_CAM_CTRL_AUTO_EXPOSURE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_AUTO_EXPOSURE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_EXPOSURE_AUTO, valp );
      break;

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_EXPOSURE_ABSOLUTE, valp );
      // convert 100 usec interval to 1 usec
      if ( valp->valueType == OA_CTRL_TYPE_INT32 ) {
        valp->int32 *= 100;
      } else {
        valp->int64 *= 100;
      }
      break;

    case OA_CAM_CTRL_PAN_RELATIVE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_PAN_RELATIVE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_PAN_RELATIVE, valp );
      break;

    case OA_CAM_CTRL_TILT_RELATIVE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_TILT_RELATIVE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_TILT_RELATIVE, valp );
      break;

    case OA_CAM_CTRL_PAN_ABSOLUTE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_PAN_ABSOLUTE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_PAN_ABSOLUTE, valp );
      break;

    case OA_CAM_CTRL_TILT_ABSOLUTE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_TILT_ABSOLUTE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_TILT_ABSOLUTE, valp );
      break;

    case OA_CAM_CTRL_ZOOM_ABSOLUTE:
      valp->valueType = camera->controls[ OA_CAM_CTRL_ZOOM_ABSOLUTE ];
      _getExtendedControl ( cameraState.fd, V4L2_CID_ZOOM_ABSOLUTE, valp );
      break;

    case OA_CAM_CTRL_CONTOUR:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd, PWC_CID_CUSTOM(contour));
      break;

    case OA_CAM_CTRL_AUTO_CONTOUR:
      valp->valueType = OA_CTRL_TYPE_BOOLEAN;
      valp->boolean = _getUserControl ( cameraState.fd,
          PWC_CID_CUSTOM(autocontour));
      break;

    case OA_CAM_CTRL_NOISE_REDUCTION:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          PWC_CID_CUSTOM(noise_reduction));
      break;

    case OA_CAM_CTRL_AUTO_WB_SPEED:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          PWC_CID_CUSTOM(awb_speed));
      break;

    case OA_CAM_CTRL_AUTO_WB_DELAY:
      valp->valueType = OA_CTRL_TYPE_INT32;
      valp->int32 = _getUserControl ( cameraState.fd,
          PWC_CID_CUSTOM(awb_delay));
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  return OA_ERR_NONE;
}


int
oaV4L2CameraTestControl ( oaCamera* camera, int control, oaControlValue* valp )
{
  int32_t      val_s32;
  int64_t      val_s64;

  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  if ( camera->controls [ control ] != valp->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  // FIX ME -- this is horribly wrong.  Should be done by control type
  // rather than by the actual control itself

  switch ( valp->valueType ) {

    case OA_CTRL_TYPE_INT32:
      val_s32 = valp->int32;
      if ( val_s32 >= camera->min[ control ] &&
          val_s32 <= camera->max[ control ] &&
          ( 0 == ( val_s32 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CTRL_TYPE_DISCRETE:
      val_s32 = valp->discrete;
      if ( val_s32 >= camera->min[ control ] &&
          val_s32 <= camera->max[ control ] &&
          ( 0 == ( val_s32 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CTRL_TYPE_MENU:
      val_s32 = valp->menu;
      if ( val_s32 >= camera->min[ control ] &&
          val_s32 <= camera->max[ control ] &&
          ( 0 == ( val_s32 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CTRL_TYPE_BOOLEAN:
      return OA_ERR_NONE;
      break;

    case OA_CTRL_TYPE_INT64:
      val_s64 = valp->int64;
      if ( val_s64 >= camera->min[ control ] &&
          val_s64 <= camera->max[ control ] &&
          ( 0 == ( val_s64 - camera->min[ control ] ) %
          camera->step[ control ] )) {
        return OA_ERR_NONE;
      }
      break;

    case OA_CTRL_TYPE_BUTTON:
      return OA_ERR_NONE;
      break;

    case OA_CTRL_TYPE_STRING:
      return OA_ERR_NONE;
      break;

    default:
      fprintf ( stderr, "%s: unhandled value type %d\n", __FUNCTION__,
          valp->valueType );
      return -OA_ERR_INVALID_CONTROL_TYPE;
      break;
  }

  // And if we reach here it's because the value wasn't valid
  return -OA_ERR_OUT_OF_RANGE;
}


int
_setUserControl ( int fd, int id, int32_t value )
{
  struct v4l2_control        control;

  OA_CLEAR ( control );
  control.id = id;
  control.value = value;
  if ( v4l2ioctl ( fd, VIDIOC_S_CTRL, &control )) {
    perror ("VIDIOC_S_CTRL");
    return -OA_ERR_SYSTEM_ERROR;
  }
  return OA_ERR_NONE;
}


int
_getUserControl ( int fd, int id )
{
  struct v4l2_control        control;

  OA_CLEAR ( control );
  control.id = id;
  if ( v4l2ioctl ( fd, VIDIOC_G_CTRL, &control )) {
    perror ("VIDIOC_G_CTRL");
    return -OA_ERR_SYSTEM_ERROR;
  }
  return control.value;
}


int
_setExtendedControl ( int fd, int id, oaControlValue* valp )
{
  struct v4l2_ext_control    extControl[1];
  struct v4l2_ext_controls   controls;

  OA_CLEAR ( controls );
  OA_CLEAR ( extControl );

  switch ( valp->valueType ) { 
    case OA_CTRL_TYPE_INT32:
      extControl[0].value = valp->int32;
      break;
    case OA_CTRL_TYPE_DISCRETE:
      extControl[0].value = valp->discrete;
      break;
    case OA_CTRL_TYPE_BOOLEAN:
      extControl[0].value = valp->boolean;
      break;
    case OA_CTRL_TYPE_MENU:
      extControl[0].value = valp->menu;
      break;
    case OA_CTRL_TYPE_INT64:
      extControl[0].value64 = valp->int64;
      break;
    case OA_CTRL_TYPE_BUTTON:
      extControl[0].value = 1;
      break;
    case OA_CTRL_TYPE_STRING:
      extControl[0].string = ( char* ) valp->string;
      break;
    default:
      fprintf ( stderr, "%s: unhandled value type %d\n", __FUNCTION__,
          valp->valueType );
      return -OA_ERR_INVALID_CONTROL_TYPE;
  }
  controls.ctrl_class = V4L2_CTRL_ID2CLASS ( id );
  // controls.ctrl_class = V4L2_CTRL_CLASS_CAMERA;
  controls.count = 1;
  controls.controls = extControl;
  extControl[0].id = id;
  if ( v4l2ioctl ( fd, VIDIOC_S_EXT_CTRLS, &controls )) {
    perror ("setExtendedControl: VIDIOC_S_EXT_CTRLS");
    return -OA_ERR_SYSTEM_ERROR;
  }
  return OA_ERR_NONE;
}


int
_getExtendedControl ( int fd, int id, oaControlValue* valp )
{
  struct v4l2_ext_control    extControl[1];
  struct v4l2_ext_controls   controls;

  OA_CLEAR ( controls );
  OA_CLEAR ( extControl );
  controls.ctrl_class = V4L2_CTRL_ID2CLASS ( id );
  controls.ctrl_class = V4L2_CTRL_CLASS_CAMERA;
  controls.count = 1;
  controls.controls = extControl;
  extControl[0].id = id;
  if ( v4l2ioctl ( fd, VIDIOC_G_EXT_CTRLS, &controls )) {
    perror ("setExtendedControl: VIDIOC_G_EXT_CTRLS");
    return -OA_ERR_SYSTEM_ERROR;
  }

  switch ( valp->valueType ) {
    case OA_CTRL_TYPE_INT32:
      valp->int32 = extControl[0].value;
      break;
    case OA_CTRL_TYPE_DISCRETE:
      valp->discrete = extControl[0].value;
      break;
    case OA_CTRL_TYPE_BOOLEAN:
      valp->boolean = extControl[0].value;
      break;
    case OA_CTRL_TYPE_MENU:
      valp->menu = extControl[0].value;
      break;
    case OA_CTRL_TYPE_INT64:
      valp->int64 = extControl[0].value64;
      break;
    case OA_CTRL_TYPE_STRING:
      valp->string = extControl[0].string;
      break;
    default:
      fprintf ( stderr, "%s: unhandled value type %d\n", __FUNCTION__,
          valp->valueType );
      return -OA_ERR_INVALID_CONTROL_TYPE;
  }
  return OA_ERR_NONE;
}


int
oaV4L2CameraSetROI ( oaCamera* camera, int x, int y )
{
  fprintf ( stderr, "attempt to set ROI of %dx%d in V4L2\n", x, y );
  return 0;
}


int
oaV4L2CameraSetFrameInterval ( oaCamera* camera, int numerator,
    int denominator )
{
  struct v4l2_streamparm     parm;
  enum v4l2_buf_type         type;

  return 0;

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_STREAMOFF, &type ) < 0 )  {
    perror ( "VIDIOC_STREAMOFF" );
  } 
  OA_CLEAR( parm );
  parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  parm.parm.capture.capability = V4L2_CAP_TIMEPERFRAME;
  parm.parm.capture.timeperframe.numerator = numerator;
  parm.parm.capture.timeperframe.denominator = denominator;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_S_PARM, &parm )) {
    perror ( "VIDIOC_S_PARM xioctl failed" );
  }
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_STREAMON, &type ) < 0 )  {
    if ( -ENOSPC == errno ) {
      fprintf ( stderr, "Insufficient bandwidth for camera on the USB bus\n" );
      return -1;
    }
    perror ( "VIDIOC_STREAMON" );
  } 
  return 0;
}


int
oaV4L2CameraStart ( oaCamera* camera, START_PARMS* parms )
{
  struct v4l2_format         fmt;
  struct v4l2_buffer         buf;
  struct v4l2_requestbuffers req;
  struct v4l2_streamparm     parm;
  enum v4l2_buf_type         type;
  unsigned int               m, n;

  OA_CLEAR ( fmt );
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = parms->size.x;
  fmt.fmt.pix.height = parms->size.y;
  fmt.fmt.pix.pixelformat = cameraState.videoCurrent;
  fmt.fmt.pix.field = V4L2_FIELD_NONE;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_S_FMT, &fmt )) {
    perror ( "VIDIOC_S_FMT v4l2ioctl failed" );
    return -1;
  }

  if ( fmt.fmt.pix.pixelformat != cameraState.videoCurrent ) {
    fprintf ( stderr, "Can't get expected video format: %c%c%c%c\n",
        cameraState.videoCurrent & 0xff, cameraState.videoCurrent >> 8 & 0xff,
        cameraState.videoCurrent >> 16 & 0xff, cameraState.videoCurrent >> 24 );
    return -1;
  }

  if (( fmt.fmt.pix.width != parms->size.x ) || ( fmt.fmt.pix.height !=
      parms->size.y )) {
    fprintf ( stderr, "Requested image size %dx%d, offered as %dx%d\n",
        parms->size.x, parms->size.y, fmt.fmt.pix.width, fmt.fmt.pix.height );
  }

  OA_CLEAR( parm );
  parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_G_PARM, &parm )) {
    if ( errno != EINVAL ) {
      perror ( "VIDIOC_G_PARM v4l2ioctl failed" );
      return -1;
    }
  }

  // FIX ME -- changing frameRates here may have unpleasant knock-on
  // effects

  if ( V4L2_CAP_TIMEPERFRAME == parm.parm.capture.capability ) {
    OA_CLEAR( parm );
    parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    parm.parm.capture.capability = V4L2_CAP_TIMEPERFRAME;
    parm.parm.capture.timeperframe.numerator = parms->rate.numerator;
    parm.parm.capture.timeperframe.denominator = parms->rate.denominator;
    camera->features.frameRates = 1;
    if ( v4l2ioctl ( cameraState.fd, VIDIOC_S_PARM, &parm )) {
      perror ( "VIDIOC_S_PARM v4l2ioctl failed" );
      return -OA_ERR_SYSTEM_ERROR;
    }
  } else {
    camera->features.frameRates = 0;
  }

  OA_CLEAR( req );
  req.count = V4L2_BUFFERS;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  if ( v4l2ioctl( cameraState.fd, VIDIOC_REQBUFS, &req )) {
    perror ( "VIDIOC_REQBUFS v4l2ioctl failed" );
    return -1;
  }

  cameraState.configuredBuffers = 0;
  cameraState.lastUsedBuffer = -1;
  cameraState.buffers = calloc( req.count, sizeof ( struct buffer ));
  for ( n = 0; n < req.count; n++ ) {
    OA_CLEAR ( buf );
    buf.type = req.type;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n;
    if ( v4l2ioctl ( cameraState.fd, VIDIOC_QUERYBUF, &buf )) {
      perror ( "VIDIOC_QUERYBUF v4l2ioctl failed" );
      return -1;
    }
    cameraState.buffers[ n ].length = buf.length;
    if ( MAP_FAILED == ( cameraState.buffers[ n ].start = v4l2_mmap ( 0,
        buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, cameraState.fd,
        buf.m.offset ))) {
      perror ( "mmap" );
      if ( n ) {
        for ( m = 0; m < n; m++ ) {
          v4l2_munmap ( cameraState.buffers[ m ].start,
              cameraState.buffers[ m ].length );
        }
      }
      return -1;
    }
    cameraState.configuredBuffers++;
  }

  for ( n = 0; n < req.count; n++ ) {
    OA_CLEAR( buf );
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n;
    if ( v4l2ioctl ( cameraState.fd, VIDIOC_QBUF, &buf ) < 0 ) {
      perror ( "init VIDIOC_QBUF" );
      for ( m = 0; m < cameraState.configuredBuffers; m++ ) {
        v4l2_munmap ( cameraState.buffers[ m ].start,
            cameraState.buffers[ m ].length );
      }
      return -1;
    }
  }

  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_STREAMON, &type ) < 0 )  {
    if ( -ENOSPC == errno ) {
      fprintf ( stderr, "Insufficient bandwidth for camera on the USB bus\n" );
    }
    perror ( "VIDIOC_STREAMON" );
    for ( m = 0; m < cameraState.configuredBuffers; m++ ) {
      v4l2_munmap ( cameraState.buffers[ m ].start,
          cameraState.buffers[ m ].length );
    }
    return -1;
  }

  return 0;
}


int
oaV4L2CameraStop ( oaCamera* camera )
{
  unsigned int       n;
  enum v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

  if ( v4l2ioctl ( cameraState.fd, VIDIOC_STREAMOFF, &type ) < 0 ) {
    perror ( "VIDIOC_STREAMOFF" );
  }

  // FIX ME -- refactor buffer release

  if ( cameraState.configuredBuffers ) {
    for ( n = 0; n < cameraState.configuredBuffers; n++ ) {
      v4l2_munmap ( cameraState.buffers[ n ].start,
          cameraState.buffers[ n ].length );
    }
  }

  return 0;
}


int
oaV4L2CameraReset ( oaCamera* camera )
{
  // FIX ME -- should check here that buffers have been released etc.?

  v4l2_close ( cameraState.fd );
  cameraState.lastUsedBuffer = -1;
  if (( cameraState.fd = v4l2_open ( cameraState.devicePath,
      O_RDWR | O_NONBLOCK, 0 )) < 0 ) {
    fprintf ( stderr, "cannot open video device %s\n",
        cameraState.devicePath );
    // errno should be set?
    free (( void * ) camera );
    return 0;
  }

  return -1;
}


int
oaV4L2CameraSetBitDepth ( oaCamera* camera, int depth )
{
  fprintf ( stderr, "%s: unsupported attempt to set depth of %d\n",
      __FUNCTION__, depth );
  return -1;
}


const char*
oaV4L2CameraGetMenuString ( oaCamera* camera, int control, int index )
{
  struct v4l2_querymenu menuItem;
  static char buff[40];

  // FIX ME -- need map of OA_CTRL to V4L2_CID values for this really
  if ( control != OA_CAM_CTRL_AUTO_WHITE_BALANCE &&
      control != OA_CAM_CTRL_AUTO_EXPOSURE ) {
    fprintf ( stderr, "%s: control not implemented\n", __FUNCTION__ );
    return "";
  }

  OA_CLEAR( menuItem );
  if ( OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
    menuItem.id = V4L2_CID_EXPOSURE_AUTO;
  }
  if ( OA_CAM_CTRL_AUTO_WHITE_BALANCE == control ) {
    menuItem.id = V4L2_CID_AUTO_WHITE_BALANCE;
  }
  menuItem.index = index;
  if ( v4l2ioctl ( cameraState.fd, VIDIOC_QUERYMENU, &menuItem )) {
    perror ("VIDIOC_QUERYMENU");
    fprintf ( stderr, "%s: control: %s, index %d\n", __FUNCTION__,
        menuItem.id == V4L2_CID_EXPOSURE_AUTO ? "auto-exposure" : "auto-wb",
        index );
    if ( OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
      return oaCameraAutoExposureLabel[ index ];
    }
  }
  strncpy ( buff, ( char* ) menuItem.name, 40 ); // <=32 per V4L2 spec
  return ( const char* ) buff;
}


// Used when the WB setting is a menu so we don't have a programmatically
// defined way to be sure of turning the white balance off

int
oaV4L2CameraGetAutoWBManualSetting ( oaCamera* camera )
{
  if ( !cameraState.haveWhiteBalanceManual ) {
    fprintf ( stderr, "%s: have no manual value for white balance menu\n",
        __FUNCTION__ );
  }
  return cameraState.autoWhiteBalanceOff;
}

#endif /* HAVE_LIBV4L2 */
