/*****************************************************************************
 *
 * atikSerialcontrol-ftdi.c -- control functions for Atik serial cameras
 * using libftdi1
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>

#include "oacamprivate.h"
#include "atikSerialoacam.h"

#define	cameraState		camera->_atikSerial

static void*    _capture ( void* );
static void     _startExposure ( oaCamera*, unsigned int );
static int      _readExposure ( oaCamera*, unsigned char*, unsigned int );

#define READ_COMMAND		3
#define	READ_EXPOSURE_TIMING	4
#define	READ_X_BINNING		5
#define	READ_Y_BINNING		6
#define	READ_ROI_X_LO		7
#define	READ_ROI_X_HI		8
#define	READ_ROI_Y_LO		9
#define	READ_ROI_Y_HI		10
#define	READ_SIZE_X_LO		11
#define	READ_SIZE_X_HI		12
#define	READ_SIZE_Y_LO		13
#define	READ_SIZE_Y_HI		14
#define	READ_CONFIG_FLAGS_LO	15
#define	READ_CONFIG_FLAGS_HI	16
#define	READ_BUFFER_LENGTH	17

#define	EXPOSURE_TIMING_EXTERNAL	0
#define	NULL_READS			20
#define OPTIMAL_READ_SIZE		3968

int
oaAtikSerialCameraSetControl ( oaCamera* camera, int control, oaControlValue* val )
{
  // FIX ME -- change ranges & return OOB or unsupported control error

  oacamDebugMsg ( DEBUG_CAM_CTRL, "atikSerial: control: %s ( %d, ? )\n",
      __FUNCTION__, control );

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      uint64_t exp;

      if ( val->valueType != OA_CTRL_TYPE_INT64 ) {
        fprintf ( stderr, "%s: invalid control type %d where int64 expected\n",
            __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      exp = val->int64 / 1000;
      cameraState.currentExposure = exp;
      break;
    }
    case OA_CAM_CTRL_BINNING:
      if ( val->valueType != OA_CTRL_TYPE_DISCRETE ) {
        fprintf ( stderr, "%s: invalid control type %d where discrete "
            "expected\n", __FUNCTION__, val->valueType );
        return -OA_ERR_INVALID_CONTROL_TYPE;
      }
      cameraState.binMode = val->discrete;
      break;

    default:
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  return OA_ERR_NONE;
}


int
oaAtikSerialCameraReadControl ( oaCamera* camera, int control,
    oaControlValue* val )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "atikSerial: control: %s ( %d )\n",
      __FUNCTION__, control );

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      val->valueType = OA_CTRL_TYPE_INT64;
      val->int64 = cameraState.currentExposure * 1000;
      break;

    case OA_CAM_CTRL_BINNING:
      val->valueType = OA_CTRL_TYPE_DISCRETE;
      val->discrete = cameraState.binMode;
      break;

    case OA_CAM_CTRL_BIT_DEPTH:
      val->valueType = OA_CTRL_TYPE_DISCRETE;
      val->discrete = 16;
      break;

    case OA_CAM_CTRL_DROPPED:
      val->valueType = OA_CTRL_TYPE_READONLY;
      val->readonly = cameraState.droppedFrames;
      break;

    default:
      fprintf ( stderr,
          "Unrecognised control %d in %s\n", control, __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }

  return OA_ERR_NONE;
}


int
oaAtikSerialCameraTestControl ( oaCamera* camera, int control,
    oaControlValue* val )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "atikSerial: control: %s ( %d, ? )\n",
      __FUNCTION__, control );

  if ( !camera->controls [ control ] ) {
    return -OA_ERR_INVALID_CONTROL;
  }

  if ( camera->controls [ control ] != val->valueType ) {
    return -OA_ERR_INVALID_CONTROL_TYPE;
  }

  switch ( control ) {

    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
    {
      int64_t val_s64 = val->int64;
      if ( val_s64 >= camera->min[ control ] &&
          val_s64 <= camera->max[ control ] &&
          ( 0 == ( val_s64 - camera->min[ control ]) %
          camera->step[ control ])) {
        return OA_ERR_NONE;
      }
      break;
    }
    case OA_CAM_CTRL_BINNING:
    {
      // For this one we allow up to the minimum of the binning values
      // for the X and Y directions
      uint32_t val_u32 = ( uint32_t ) val->discrete;
      if ( cameraState.maxBinningX >= val_u32 &&
          cameraState.maxBinningY >= val_u32 ) {
        return OA_ERR_NONE;
      }
      break;
    }
    default:
      // If we reach here it's because we don't recognise the control
      // FIX ME -- make this debug only
      fprintf ( stderr, "Unrecognised control %d in %s\n", control,
          __FUNCTION__ );
      return -OA_ERR_INVALID_CONTROL;
      break;
  }
  // And if we reach here it's because the value wasn't valid

  return -OA_ERR_OUT_OF_RANGE;
}


int
oaAtikSerialCameraSetROI ( oaCamera* camera, int x, int y )
{
  cameraState.xSize = x;
  cameraState.ySize = y;
  return 0;
}


int
oaAtikSerialCameraStart ( oaCamera* camera, START_PARMS* parms )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "AtikSerial: control: %s ([ %d, %d ])\n",
      __FUNCTION__, parms->size.x, parms->size.y );

  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  cameraState.xSize = parms->size.x;
  cameraState.ySize = parms->size.y;

  cameraState.imageBufferLength = parms->size.x * parms->size.y * 2 /
      cameraState.binMode;

  cameraState.captureLength = cameraState.imageBufferLength;
  pthread_create ( &cameraState.captureThread, 0, &_capture, camera );
  cameraState.captureThreadStarted = 1;

  return 0;
}


int
oaAtikSerialCameraStop ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "AtikSerial: command: %s ()\n", __FUNCTION__ );

  if ( !cameraState.captureThreadStarted ) {
    return 0;
  }

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.captureThreadExit = 1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  pthread_join ( cameraState.captureThread, 0 );
  cameraState.captureThreadStarted = 0;
  return 0;
}


// FIX ME -- should this return value be the other way around?
int
oaAtikSerialCameraReset ( oaCamera* camera )
{
  oacamDebugMsg ( DEBUG_CAM_CTRL, "AtikSerial: control: %s ()\n", __FUNCTION__ );

  fprintf ( stderr, "%s needs implementing\n", __FUNCTION__ );
  return 0;
}


static void*
_capture ( void* param )
{
  oaCamera*     camera = param;
  int           quitThread, nextBuffer;
  unsigned int  requestedTime;
  unsigned char ampOnCmd[5] = { 'C', 'M', 'D', ATIK_CMD_SET_AMP, 1 };

  pthread_mutex_lock ( &cameraState.captureMutex );
  cameraState.haveDataToRead = 0;
  cameraState.nextBufferToRead = -1;
  pthread_mutex_unlock ( &cameraState.captureMutex );
  nextBuffer = 0;

  while ( 1 ) {
    pthread_mutex_lock ( &cameraState.captureMutex );
    quitThread = cameraState.captureThreadExit;
    if ( nextBuffer == cameraState.nextBufferToRead ) {
      nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;
    }
    requestedTime = cameraState.currentExposure;
    pthread_mutex_unlock ( &cameraState.captureMutex );
    if ( quitThread ) {
      break;
    }

    // ATIK_SERIAL_MAX_SHORT_EXPOSURE is the maximum exposure that can
    // be timed on-camera.  If we're doing an exposure timed by the
    // camera hardware then there's no point doing the stuff in the
    // _startExposure() function because the exposure doesn't actually
    // take place until we get to  _readExposure() where we revert some of
    // the stuff done in _startExposure()

    if ( requestedTime > ATIK_SERIAL_MAX_SHORT_EXPOSURE ) {
      // FIX ME -- does _startExposure actually need to turn off guiding, or
      // could it be done here as it happens for both exposures timed on-
      // camera and externally?
      _startExposure ( camera, requestedTime );
      requestedTime -= 300; // 300ms is some sort of magic number here
      usleep ( requestedTime * 1000 );

      if ( _atikSerialCamWrite ( cameraState.ftdiContext, ampOnCmd, 4 )) {
        fprintf ( stderr, "%s: write error on amp on\n", __FUNCTION__ );
      }
      usleep ( 100000 );

    }
    _readExposure ( camera, cameraState.buffers[ nextBuffer ].start,
        cameraState.captureLength );

    pthread_mutex_lock ( &cameraState.captureMutex );
    cameraState.nextBufferToRead = nextBuffer;
    cameraState.haveDataToRead = 1;
    pthread_cond_signal ( &cameraState.frameAvailable );
    pthread_mutex_unlock ( &cameraState.captureMutex );
    nextBuffer = ( nextBuffer + 1 ) % cameraState.configuredBuffers;

  }
  return 0;
}


static void
_startExposure ( oaCamera* camera, unsigned int exposureTime )
{
  unsigned char		guideOffCmd[4] = { 'C', 'M', 'D',
			    ATIK_CMD_DISABLE_GUIDE };
  unsigned char		ampOffCmd[5] = { 'C', 'M', 'D', ATIK_CMD_SET_AMP, 0 };
  unsigned char		clearCCDCmd[4] = { 'C', 'M', 'D', ATIK_CMD_CLEAR_CCD };
  unsigned char		shutterCmd[4] = { 'C', 'M', 'D',
			    ATIK_CMD_START_EXPOSURE };

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, ampOffCmd, 5 )) {
    fprintf ( stderr, "%s: write error on amp off\n", __FUNCTION__ );
    return;
  }
  usleep ( 100000 );

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, clearCCDCmd, 4 )) {
    fprintf ( stderr, "%s: write error on clear CCD\n", __FUNCTION__ );
    return;
  }
  usleep ( 100000 );

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, shutterCmd, 4 )) {
    fprintf ( stderr, "%s: write error on shutter\n", __FUNCTION__ );
    return;
  }
  usleep ( 100000 );

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, guideOffCmd, 4 )) {
    fprintf ( stderr, "%s: write error on guide off\n", __FUNCTION__ );
    return;
  }
  usleep ( 100000 );
}


static int
_readExposure ( oaCamera* camera, unsigned char* buffer,
    unsigned int readLength )
{
  unsigned char         guideOffCmd[4] = { 'C', 'M', 'D',
                            ATIK_CMD_DISABLE_GUIDE };
  unsigned char		ampOffCmd[5] = { 'C', 'M', 'D', ATIK_CMD_SET_AMP, 0 };
  unsigned char         readCCDCmd[20] = { 'C', 'M', 'D', ATIK_CMD_READ_CCD };
  unsigned char*	p = buffer;
  unsigned int		bytesToRead = readLength;
  int			allowedEmptyReads, numRead;

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, guideOffCmd, 4 )) {
    fprintf ( stderr, "%s: write error on guide off\n", __FUNCTION__ );
    return -1;
  }
  usleep ( 100000 );

  if ( cameraState.currentExposure < 10 ) {
    readCCDCmd[ READ_EXPOSURE_TIMING ] = cameraState.currentExposure;
    // This may need to be the other way around for some of the cameras
    // cameraState.ccdReadFlags |= ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
    cameraState.ccdReadFlags &= ~ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
  } else {
    if ( cameraState.currentExposure <= ATIK_SERIAL_MAX_SHORT_EXPOSURE ) {
      readCCDCmd[ READ_EXPOSURE_TIMING ] =
          cameraState.currentExposure / 10 + 9;
      // This may need to be the other way around for some of the cameras
      // cameraState.ccdReadFlags |= ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
      cameraState.ccdReadFlags &= ~ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
    } else {
      readCCDCmd[ READ_EXPOSURE_TIMING ] = EXPOSURE_TIMING_EXTERNAL;
      // This may need to be the other way around for some of the cameras
      // cameraState.ccdReadFlags &= ~ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
      cameraState.ccdReadFlags |= ATIK_SERIAL_READ_FLAGS_TIMER_AMP_ON;
    }
  }

  readCCDCmd[ READ_X_BINNING ] = cameraState.binMode;
  readCCDCmd[ READ_Y_BINNING ] = cameraState.binMode;
  readCCDCmd[ READ_ROI_X_LO ] = 0;
  readCCDCmd[ READ_ROI_X_HI ] = 0;
  readCCDCmd[ READ_ROI_Y_LO ] = 0;
  readCCDCmd[ READ_ROI_Y_HI ] = 0;
  readCCDCmd[ READ_SIZE_X_LO ] = cameraState.xSize & 0xff;
  readCCDCmd[ READ_SIZE_X_HI ] = cameraState.xSize >> 8;
  readCCDCmd[ READ_SIZE_Y_LO ] = cameraState.ySize & 0xff;
  readCCDCmd[ READ_SIZE_Y_HI ] = cameraState.ySize >> 8;

  if ( cameraState.ccdReadFlags & ATIK_SERIAL_READ_FLAGS_DEINTERLACE ) {
    fprintf ( stderr, "%s: Help!  can't handle interlaced camera yet.  Add code\n",
        __FUNCTION__ );
  }

  readCCDCmd[ READ_CONFIG_FLAGS_LO ] = cameraState.ccdReadFlags & 0xff;
  readCCDCmd[ READ_CONFIG_FLAGS_HI ] = cameraState.ccdReadFlags >> 8;

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, readCCDCmd,
      READ_BUFFER_LENGTH )) {
    fprintf ( stderr, "%s: write error on read cmd\n", __FUNCTION__ );
    return -1;
  }
  usleep ( 100000 );

  allowedEmptyReads = NULL_READS;
  while ( bytesToRead && allowedEmptyReads ) {
    numRead = _atikSerialCamReadBlock ( cameraState.ftdiContext, p,
      bytesToRead >= OPTIMAL_READ_SIZE ? OPTIMAL_READ_SIZE : bytesToRead );
    if ( !numRead ) {
      allowedEmptyReads--;
    } else {
      if ( numRead > 0 ) {
        p += numRead;
        bytesToRead -= numRead;
        allowedEmptyReads = NULL_READS;
      }
    }
    // I determined this delay experimentally to allow something close to
    // OPTIMAL_READ_SIZE byte to be available at the time of the above read
    usleep ( 20000 );
  }

  if ( bytesToRead ) {
    fprintf ( stderr, "short read.  Expected %d, got %lu\n", readLength,
        p - buffer );
    cameraState.droppedFrames++;
  }

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, ampOffCmd, 5 )) {
    fprintf ( stderr, "%s: write error on amp off\n", __FUNCTION__ );
    return -1;
  }
  usleep ( 100000 );

  return 0;
}
