/*****************************************************************************
 *
 * atikSerialinit-ftdi.c -- Initialise Atik serial cameras using libftdi1
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <errno.h>
#ifdef HAVE_LIBFTDI1_FTDI_H
#include <libftdi1/ftdi.h>
#endif
#ifdef HAVE_LIBFTDI_FTDI_H
#include <libftdi/ftdi.h>
#endif
#include <openastro/camera.h>

#include "atikSerialoacam.h"
#include "unimplemented.h"
#include "oacamprivate.h"

#define	cameraState	camera->_atikSerial

static void _atikSerialInitFunctionPointers ( oaCamera* );

#define	CAPS_FLAGS_LO		0
#define	CAPS_FLAGS_HI		1
#define	CAPS_TOTAL_PIXELS_X_LO	2
#define	CAPS_TOTAL_PIXELS_X_HI	3
#define	CAPS_TOTAL_PIXELS_Y_LO	4
#define	CAPS_TOTAL_PIXELS_Y_HI	5
#define	CAPS_PIXEL_SIZE_X_LO	6
#define	CAPS_PIXEL_SIZE_X_HI	7
#define	CAPS_PIXEL_SIZE_Y_LO	8
#define	CAPS_PIXEL_SIZE_Y_HI	9
#define	CAPS_MAX_BIN_X_LO	10
#define	CAPS_MAX_BIN_X_HI	11
#define	CAPS_MAX_BIN_Y_LO	12
#define	CAPS_MAX_BIN_Y_HI	13
#define	CAPS_WELL_DEPTH_LO	14
#define	CAPS_WELL_DEPTH_HI	15

#define	DEFAULT_EXPOSURE	500

#define ATIK_SERIAL_BUFFERS	5

/**
 * Initialise a given camera device
 */

oaCamera*
oaAtikSerialInitCamera ( oaCameraDevice* device )
{
  oaCamera*		camera;
  int			numRead, i, j, useFIFO, matched, ret;
  int                   deviceAddr, deviceBus, numUSBDevices;
  unsigned char		buffer[64];
  unsigned char		capsCmd[4] = { 'C', 'M', 'D', ATIK_CMD_QUERY_CAPS };
  unsigned char		pingCmd[4] = { 'C', 'M', 'D', ATIK_CMD_PING };
  unsigned char		serialCmd[4] = { 'C', 'M', 'D',
                            ATIK_CMD_QUERY_SERIAL_NO };
  unsigned char		fifoCmd[4] = { 'C', 'M', 'D', ATIK_CMD_QUERY_FIFO };
  unsigned char		extCmd[8] = { 'C', 'M', 'D', ATIK_CMD_SEND_EXTERNAL,
			    0xc9, 0xf4, 0xa9, 0xdb };
  libusb_device**       devlist;
  libusb_device*        usbDevice;
  struct libusb_device_descriptor desc;


  if (!( camera = ( oaCamera* ) malloc ( sizeof ( oaCamera )))) {
    perror ( "malloc oaCamera failed" );
    return 0;
  }

  OA_CLEAR ( cameraState );
  OA_CLEAR ( camera->controls );
  OA_CLEAR ( camera->features );

  pthread_mutex_init ( &cameraState.serialMutex, 0 );

  ( void ) strcpy ( camera->deviceName, device->deviceName );
  ( void ) strcpy ( cameraState.devicePath, device->sysPath );
  cameraState.initialised = 0;
  cameraState.index = -1;

  libusb_init ( &cameraState.usbContext );
  numUSBDevices = libusb_get_device_list ( cameraState.usbContext, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( cameraState.usbContext );
    if ( numUSBDevices ) {
      fprintf ( stderr, "Can't see any USB devices now (list returns -1)\n" );
      free ( camera );
      return 0;
    }
    fprintf ( stderr, "Can't see any USB devices now\n" );
    free ( camera );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  for ( i = 0; i < numUSBDevices && !matched; i++ ) {
    usbDevice = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( usbDevice, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( cameraState.usbContext );
      fprintf ( stderr, "get device descriptor failed\n" );
      free ( camera );
      return 0;
    }

    if ( desc.idVendor == device->_vendorId &&
        desc.idProduct == device->_productId &&
        libusb_get_bus_number ( usbDevice ) == deviceBus &&
        libusb_get_device_address ( usbDevice ) == deviceAddr ) {
      // this looks like the one!
      matched = 1;
    }
  }

  if ( matched ) {
    if (!( cameraState.ftdiContext = ftdi_new())) {
      fprintf ( stderr, "ftdi_new() failed\n" );
      matched = 0;
    } else {
      if (( ret = ftdi_init ( cameraState.ftdiContext ))) {
        fprintf ( stderr, "can't initialise ftdi context, err = %d\n", ret );
        matched = 0;
      } else {
        ftdi_set_interface ( cameraState.ftdiContext, INTERFACE_ANY );
        if ( ftdi_usb_open_dev ( cameraState.ftdiContext, usbDevice ) < 0 ) {
          fprintf ( stderr, "FTDI open of serial device failed\n" );
          matched = 0;
        }
      }
    }
  }

  // now the libusb stuff can be released

  libusb_free_device_list ( devlist, 0 );

  if ( !matched ) {
    fprintf ( stderr, "no matching camera\n" );
    free ( camera );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  if (( ret = ftdi_set_baudrate ( cameraState.ftdiContext, 115200 )) < 0 ) {
    fprintf ( stderr, "set baud rate for FTDI serial device failed, "
        "err = %d\n", ret );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  if ( ftdi_set_line_property ( cameraState.ftdiContext, BITS_8, STOP_BIT_1,
      NONE ) < 0 ) {
    fprintf ( stderr, "set 8N1 for device FTDI serial device failed\n" );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  camera->interface = device->interface;
  cameraState.index = device->_devIndex;
  cameraState.cameraType = device->_devType;

  // Send a PING command to the PIC.  No idea if this is really required.
  if ( _atikSerialCamWrite ( cameraState.ftdiContext, pingCmd, 4 )) {
    fprintf ( stderr, "%s: write error on ping\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }
  usleep ( 100000 );
  memset ( buffer, 0, 16 );
  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext,
      buffer, 1 ) != 1 )) {
    fprintf ( stderr, "%s: read error on ping\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, capsCmd, 4 )) {
    fprintf ( stderr, "%s: write error on query caps\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  memset ( buffer, 0, 16 );
  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext,
      buffer, 2 )) != 2 ) {
    fprintf ( stderr, "%s: read error 1 on query caps (%d)\n", __FUNCTION__,
        numRead );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  /*
  fprintf ( stderr, "%s: camera protocol version %hhu.%hhu\n", __FUNCTION__,
      buffer[1], buffer[0] );
   */

  memset ( buffer, 0, 16 );
  if (( numRead = _atikSerialCamReadToZero ( cameraState.ftdiContext,
      buffer )) < 1 ) {
    fprintf ( stderr, "%s: read error 2 on query caps\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  // fprintf ( stderr, "%s: camera id '%s'\n", __FUNCTION__, buffer );

  memset ( buffer, 0, 16 );
  if (( numRead = _atikSerialCamReadToZero ( cameraState.ftdiContext,
      buffer )) < 1 ) {
    fprintf ( stderr, "%s: read error 3 on query caps\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  // fprintf ( stderr, "%s: manufacturer '%s'\n", __FUNCTION__, buffer );

  memset ( buffer, 0, 16 );
  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext, buffer,
      16 ) != 16 )) {
    fprintf ( stderr, "%s: read error 4 on query caps\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  cameraState.cameraFlags = buffer[ CAPS_FLAGS_LO ] |
      ( buffer[ CAPS_FLAGS_HI ] << 8 );
  cameraState.maxResolutionX = buffer[ CAPS_TOTAL_PIXELS_X_LO ] |
      ( buffer[ CAPS_TOTAL_PIXELS_X_HI ] << 8 );
  cameraState.maxResolutionY = buffer[ CAPS_TOTAL_PIXELS_Y_LO ] |
      ( buffer[ CAPS_TOTAL_PIXELS_Y_HI ] << 8 );
  cameraState.pixelSizeX = buffer[ CAPS_PIXEL_SIZE_X_LO ] |
      ( buffer[ CAPS_PIXEL_SIZE_X_HI ] << 8 );
  cameraState.pixelSizeY = buffer[ CAPS_PIXEL_SIZE_Y_LO ] |
      ( buffer[ CAPS_PIXEL_SIZE_Y_HI ] << 8 );
  cameraState.maxBinningX = buffer[ CAPS_MAX_BIN_X_LO ] |
      ( buffer[ CAPS_MAX_BIN_X_HI ] << 8 );
  cameraState.maxBinningY = buffer[ CAPS_MAX_BIN_Y_LO ] |
      ( buffer[ CAPS_MAX_BIN_Y_HI ] << 8 );
  cameraState.wellDepth = buffer[ CAPS_WELL_DEPTH_LO ] |
      ( buffer[ CAPS_WELL_DEPTH_HI ] << 8 );

  /*
  fprintf ( stderr, "flags: %04x\n", cameraState.cameraFlags );
  fprintf ( stderr, "pixels X: %d\n", cameraState.maxResolutionX );
  fprintf ( stderr, "pixels Y: %d\n", cameraState.maxResolutionY );
  fprintf ( stderr, "pixel width: %d\n", cameraState.pixelSizeX );
  fprintf ( stderr, "pixel height: %d\n", cameraState.pixelSizeY );
  fprintf ( stderr, "max X binning: %d\n", cameraState.maxBinningX );
  fprintf ( stderr, "max Y binning: %d\n", cameraState.maxBinningY );
  fprintf ( stderr, "well depth: %d\n", cameraState.wellDepth * 1000 );
   */

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, serialCmd, 4 )) {
    fprintf ( stderr, "%s: write error on query serial no\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }
  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext,
      buffer, 7 ) != 7 )) {
    fprintf ( stderr, "%s: read error on query serial no\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  /*
  fprintf ( stderr, "serial no: %.6s\n", buffer );
  fprintf ( stderr, "hardware type: %s\n", buffer[6] ? "production" : "beta" );
   */

  cameraState.hardwareType = buffer[6];

  if ( _atikSerialCamWrite ( cameraState.ftdiContext, fifoCmd, 4 )) {
    fprintf ( stderr, "%s: write error on query fifo\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }
  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext,
      buffer, 1 ) != 1 )) {
    fprintf ( stderr, "%s: read error on query fifo\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }
  // fprintf ( stderr, "have fifo: %s\n", buffer[0] ? "yes" : "no" );
  cameraState.haveFIFO = buffer[0];

  // Not sure what this does, but my camera won't work without it.
  if ( _atikSerialCamWrite ( cameraState.ftdiContext, extCmd, 8 )) {
    fprintf ( stderr, "%s: write error on ext port\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  if (( numRead = _atikSerialCamRead ( cameraState.ftdiContext,
      buffer, 1 ) != 1 )) {
    fprintf ( stderr, "%s: read error on ext port\n", __FUNCTION__ );
    ftdi_usb_close ( cameraState.ftdiContext );
    ftdi_free ( cameraState.ftdiContext );
    return 0;
  }

  _oaInitCameraFunctionPointers ( camera );
  _atikSerialInitFunctionPointers ( camera );

  camera->controls[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = OA_CTRL_TYPE_INT64;
  camera->min[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1000;
  camera->max[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1800000000; // I made this up
  camera->step[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = 1000;
  camera->def[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = DEFAULT_EXPOSURE * 1000;

  camera->controls[ OA_CAM_CTRL_DROPPED ] = OA_CTRL_TYPE_READONLY;
  camera->controls[ OA_CAM_CTRL_DROPPED_RESET ] = OA_CTRL_TYPE_BUTTON;

  cameraState.resolutions[0].x = cameraState.maxResolutionX;
  cameraState.resolutions[0].y = cameraState.maxResolutionY;
  cameraState.resolutions[1].x = 0;
  cameraState.resolutions[1].y = 0;

  cameraState.buffers = 0;
  cameraState.configuredBuffers = 0;
  cameraState.droppedFrames = 0;

  cameraState.captureLength = cameraState.maxResolutionX *
      cameraState.maxResolutionY * 2;
  if (!( cameraState.xferBuffer = malloc ( cameraState.captureLength ))) {
    fprintf ( stderr, "malloc of transfer buffer failed in %s\n",
        __FUNCTION__ );
    return 0;
  }

  cameraState.imageBufferLength = cameraState.captureLength;
  cameraState.buffers = calloc ( ATIK_SERIAL_BUFFERS,
      sizeof ( struct atikSerialbuffer ));
  for ( i = 0; i < ATIK_SERIAL_BUFFERS; i++ ) {
    void* m = malloc ( cameraState.imageBufferLength );
    if ( m ) {
      cameraState.buffers[i].start = m;
      cameraState.configuredBuffers++;
    } else {
      fprintf ( stderr, "%s malloc failed\n", __FUNCTION__ );
      if ( i ) {
        for ( j = 0; j < i; j++ ) {
          free (( void* ) cameraState.buffers[j].start );
        }
      }
      free ( cameraState.xferBuffer );
      return 0;
    }
  }

  cameraState.currentExposure = DEFAULT_EXPOSURE;
  cameraState.captureThreadStarted = 0;
  cameraState.captureThreadExit = 0;
  pthread_mutex_init ( &cameraState.captureMutex, 0 );
  pthread_cond_init ( &cameraState.frameAvailable, 0 );

  switch ( cameraState.cameraType ) {
    case CAM_ATK16C:
    case CAM_ATK16HRC:
    case CAM_ATK16ICC:
    case CAM_ATK16ICSC:
      cameraState.colour = 1;
      break;
    default:
      cameraState.colour = 0;
      break;
  }

  cameraState.binMode = cameraState.horizontalBinMode =
      cameraState.verticalBinMode = OA_BIN_MODE_NONE;
  cameraState.ccdReadFlags = ATIK_SERIAL_READ_FLAGS_CTP_BOTH |
      ATIK_SERIAL_READ_FLAGS_IPCS_MODE;
  if ( cameraState.cameraFlags & ATIK_SERIAL_FLAGS_INTERLACED ) {
    cameraState.ccdReadFlags |= ATIK_SERIAL_READ_FLAGS_DEINTERLACE;
  }

  useFIFO = 0;
  /*
   * Even though my camera claims to have a FIFO, Artemis Capture doesn't
   * appear to use it.  Not that it means we shouldn't, I guess :)
   */
  if ( cameraState.cameraFlags & ATIK_SERIAL_FLAGS_HAVE_FIFO ) {
    useFIFO = 1;
  } else {
    if ( cameraState.hardwareType ) {
      useFIFO = 1;
    } else {
      if ( cameraState.haveFIFO ) {
        useFIFO = 1;
      }
    }
  }
  if ( useFIFO ) {
    cameraState.ccdReadFlags |= ATIK_SERIAL_READ_FLAGS_USE_FIFO;
  }

  return camera;
}


static void
_atikSerialInitFunctionPointers ( oaCamera* camera )
{
  camera->funcs.initCamera = oaAtikSerialInitCamera;
  camera->funcs.closeCamera = oaAtikSerialCloseCamera;
/*
  camera->funcs.resetCamera = oaAtikSerialCameraReset;
*/
  // camera->funcs.startCapture = oaAtikSerialCameraStart;
  camera->funcs.startCaptureExtended = oaAtikSerialCameraStart;
  camera->funcs.stopCapture = oaAtikSerialCameraStop;
  camera->funcs.hasLoadableFirmware = oaAtikSerialCameraHasLoadableFirmware;
  // camera->funcs.isFirmwareLoaded = _cameraIsFirmwareLoaded;
  // camera->funcs.loadFirmware = _cameraLoadFirmware;
  camera->funcs.getName = oaAtikSerialCameraGetName;
  camera->funcs.getControlRange = oaAtikSerialCameraGetControlRange;
  camera->funcs.readControl = oaAtikSerialCameraReadControl;
  camera->funcs.setControl = oaAtikSerialCameraSetControl;
  camera->funcs.testControl = oaAtikSerialCameraTestControl;
  // camera->funcs.hasAuto = oaHasAuto;
  // camera->funcs.isAuto = _isAuto;
  // camera->funcs.setControlMulti = _setControlMulti;
  camera->funcs.startReadFrame = oaAtikSerialCameraStartReadFrame;
  camera->funcs.readFrame = oaAtikSerialCameraReadFrame;
  camera->funcs.finishReadFrame = oaAtikSerialCameraFinishReadFrame;
  camera->funcs.enumerateFrameSizes = oaAtikSerialCameraGetFrameSizes;
  // camera->funcs.enumerateFrameRates = oaAtikSerialCameraGetFrameRates;
  camera->funcs.getFramePixelFormat = oaAtikSerialCameraGetFramePixelFormat;
  // camera->funcs.testROISize = oaAtikSerialCameraTestROISize;
  // camera->funcs.getMenuString = oaAtikSerialCameraGetMenuString;
  // camera->funcs.getAutoWBManualSetting = oaAtikSerialCameraGetAutoWBManualSetting;
  camera->funcs.hasFixedFrameRates = oaAtikSerialCameraHasFixedFrameRates;
}


int
oaAtikSerialCloseCamera ( oaCamera* camera )
{
  return 0;
}
