/*****************************************************************************
 *
 * atikSerialoacam-udev.c -- main entrypoint for Atik serial support via udev
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>
#include <openastro/camera.h>

#ifdef HAVE_LIBFTDI1_FTDI_H
#include <libftdi1/ftdi.h>
#endif
#ifdef HAVE_LIBFTDI_FTDI_H
#include <libftdi/ftdi.h>
#endif

#include "atikSerialoacam.h"
#include "unimplemented.h"

struct atikSerialCam atikCameraList[] =
{
  // ?
  { ATIK_FTDI_VENDOR_ID, 0xdf28, "Artemis", CAM_ARTEMIS },
  // 782x582, 8.3um pixels, USB1, 16-bit mono
  { ATIK_FTDI_VENDOR_ID, 0xdf2c, "Atik ATK-16IC-S", CAM_ATK16ICS },
  // 782x582, 8.3um pixels, USB1, 16-bit colour
  { ATIK_FTDI_VENDOR_ID, 0xdf2d, "Atik ATK-16IC-S-C", CAM_ATK16ICSC },
  // 782x582, 8.6x8.3um pixels, 16-bit, mono
  { ATIK_FTDI_VENDOR_ID, 0xdf30, "Atik ATK-16", CAM_ATK16 },
  { ATIK_FTDI_VENDOR_ID, 0xdf31, "Atik ATK-16HR", CAM_ATK16HR },
  { ATIK_FTDI_VENDOR_ID, 0xdf32, "Atik ATK-16C", CAM_ATK16HR },
  { ATIK_FTDI_VENDOR_ID, 0xdf33, "Atik ATK-16HRC", CAM_ATK16HRC },
  { ATIK_FTDI_VENDOR_ID, 0xdf34, "Mini Artemis", CAM_MINI_ARTEMIS },
  // 659x494, 7.4um pixels, USB1, mono
  { ATIK_FTDI_VENDOR_ID, 0xdf35, "Atik ATK-16IC", CAM_ATK16IC },
  // 659x494, 7.4um pixels, USB1, colour
  { ATIK_FTDI_VENDOR_ID, 0xdf36, "Atik ATK-16IC-C", CAM_ATK16ICC }
};

#define	cameraState	camera->_atikSerial


int
oaAtikSerialGetCameras ( oaCameraDevice** deviceList )
{
  // want to find all the usb serial devices and see which have
  // a VID:PID of 0403:6001 or VID:PID 0403:6015.  Further check
  // that the serial number for the device is either A4008T44 or
  // DA001CEZ

  int current = 0, numFound = 0, numUSBDevices, i, j, numRead, ret;
  int matchedVidPid, matchedUSBStrings, entry, numCameras;
  libusb_context*                       ctx = 0;
  libusb_device**                       devlist;
  libusb_device*                        device;
  libusb_device_handle*                 handle;
  struct libusb_device_descriptor       desc;
  oaCameraDevice*                       camera;
  unsigned short                        busNum, addr;
  unsigned int                          index;
  unsigned char                         cmd[3];
  struct ftdi_context*                  ftdiCtx;

  while ( deviceList[ current ] ) {
    current++;
  }
  if ( current >= OA_MAX_DEVICES ) {
    return 0;
  }

  numCameras = sizeof ( atikCameraList ) / sizeof ( struct atikSerialCam );

  libusb_init ( &ctx );
  // libusb_set_debug ( ctx, LIBUSB_LOG_LEVEL_DEBUG );
  numUSBDevices = libusb_get_device_list ( ctx, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( ctx );
    if ( numUSBDevices ) {
      return -1;
    }
    return 0;
  }

  if (!( ftdiCtx = ftdi_new())) {
    fprintf ( stderr, "can't connect to ftdi\n" );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( ctx );
  }

  if (( ret = ftdi_init ( ftdiCtx ))) {
    fprintf ( stderr, "can't initialise ftdi context, err = %d\n", ret );
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( ctx );
  }

  ftdi_set_interface ( ftdiCtx, INTERFACE_ANY );

  for ( i = 0; i < numUSBDevices; i++ ) {
    device = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( device, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( ctx );
      return -1;
    }
    busNum = libusb_get_bus_number ( device );
    addr = libusb_get_device_address ( device );
    index = ( busNum << 8 ) | addr;
    matchedVidPid = -1;
    for ( j = 0; j < numCameras && matchedVidPid == -1; j++ ) {
      if ( desc.idVendor == atikCameraList[j].vendorId &&
          desc.idProduct == atikCameraList[j].productId ) {
        matchedVidPid = j;
      }
    }
    if ( matchedVidPid >= 0 ) {

      if (!( camera = malloc ( sizeof ( oaCameraDevice )))) {
        libusb_free_device_list ( devlist, 1 );
        libusb_exit ( ctx );
        ftdi_free ( ftdiCtx );
        return -OA_ERR_SYSTEM_ERROR;
      }

      _oaInitCameraDeviceFunctionPointers ( camera );
      camera->interface = OA_CAM_IF_ATIK_SERIAL;
      camera->_devType = atikCameraList[ matchedVidPid ].devType;
      ( void ) strcpy ( camera->deviceName,
          atikCameraList[ matchedVidPid ].name );
      camera->_devIndex = ( busNum << 8 ) | addr;
      camera->_vendorId = atikCameraList[ matchedVidPid ].vendorId;
      camera->_productId = atikCameraList[ matchedVidPid ].productId;
      camera->hasLoadableFirmware = oaAtikSerialCameraDeviceHasLoadableFirmware;
      camera->initCamera = oaAtikSerialInitCamera;
      deviceList [ current++ ] = camera;
      numFound++;
    }
  }

  libusb_free_device_list ( devlist, 1 );
  libusb_exit ( ctx );

  return numFound;
}
