/*****************************************************************************
 *
 * xagylInit-ftdi.c -- Initialise Xagyl filter wheels (libftdi)
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <errno.h>
#include <openastro/filterwheel.h>
#include <libusb-1.0/libusb.h>
#include <libftdi1/ftdi.h>

#include "xagylfw.h"
#include "oafwprivate.h"
#include "unimplemented.h"

#define	wheelState		wheel->_xagyl


static void _xagylInitFunctionPointers ( oaFilterWheel* );
static int  _getNumSlots ( oaFilterWheel* );
static int  _doReset ( oaFilterWheel*, const char*, int );

/**
 * Initialise a given filter wheel device
 */

oaFilterWheel*
oaXagylInitFilterWheel ( oaFilterWheelDevice* device )
{
  oaFilterWheel*			wheel;
  int                                   i, j, matched, ret, transferred;
  int                                   deviceAddr, deviceBus, numUSBDevices;
  libusb_context*			usbContext;
  libusb_device**                       devlist;
  libusb_device*                        usbDevice;
  libusb_device_handle*                 usbHandle = 0;
  struct libusb_device_descriptor       desc;

  if (!( wheel = ( oaFilterWheel* ) malloc ( sizeof ( oaFilterWheel )))) {
    perror ( "malloc oaFilterWheel failed" );
    return 0;
  }

  OA_CLEAR ( wheelState );
  OA_CLEAR ( wheel->features );

  pthread_mutex_init ( &wheelState.serialMutex, 0 );

  ( void ) strcpy ( wheel->deviceName, device->deviceName );
  wheelState.initialised = 0;
  wheelState.index = -1;

  libusb_init ( &usbContext );
  numUSBDevices = libusb_get_device_list ( usbContext, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( usbContext );
    if ( numUSBDevices ) {
      fprintf ( stderr, "Can't see any USB devices now (list returns -1)\n" );
      free ( wheel );
      return 0;
    }
    fprintf ( stderr, "Can't see any USB devices now\n" );
    free ( wheel );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  for ( i = 0; i < numUSBDevices && !matched; i++ ) {
    usbDevice = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( usbDevice, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( usbContext );
      fprintf ( stderr, "get device descriptor failed\n" );
      free ( wheel );
      return 0;
    }
    if ( desc.idVendor == XAGYL_VID &&
        ( desc.idProduct == XAGYL_FILTERWHEEL_PID1 ||
        desc.idProduct == XAGYL_FILTERWHEEL_PID2 ) &&
        libusb_get_bus_number ( usbDevice ) == deviceBus &&
        libusb_get_device_address ( usbDevice ) == deviceAddr ) {
      // this looks like the one!
      matched = 1;
    }
  }

  if ( matched ) {
    if (!( wheelState.ftdiContext = ftdi_new())) {
      fprintf ( stderr, "ftdi_new() failed\n" );
      matched = 0;
    } else {
      if (( ret = ftdi_init ( wheelState.ftdiContext ))) {
        fprintf ( stderr, "can't initialise ftdi context, err = %d\n", ret );
        matched = 0;
      } else {
        ftdi_set_interface ( wheelState.ftdiContext, INTERFACE_ANY );
        if ( ftdi_usb_open_dev ( wheelState.ftdiContext, usbDevice ) < 0 ) {
          fprintf ( stderr, "FTDI open of serial device failed\n" );
          matched = 0;
        }
      }
    }
  }

  // now the libusb stuff can be released

  libusb_free_device_list ( devlist, 0 );
  // libusb_exit ( usbContext );

  if ( !matched ) {
    fprintf ( stderr, "no matching filter wheel\n" );
    free ( wheel );
    ftdi_free ( wheelState.ftdiContext );
    return 0;
  }

  if (( ret = ftdi_set_baudrate ( wheelState.ftdiContext, 9600 )) < 0 ) {
    fprintf ( stderr, "set baud rate for FTDI serial device failed, "
        "err = %d\n", ret );
    ftdi_usb_close ( wheelState.ftdiContext );
    ftdi_free ( wheelState.ftdiContext );
    return 0;
  }

  if ( ftdi_set_line_property ( wheelState.ftdiContext, BITS_8, STOP_BIT_1,
      NONE ) < 0 ) {
    fprintf ( stderr, "set 8N1 for device FTDI serial device failed\n" );
    ftdi_usb_close ( wheelState.ftdiContext );
    ftdi_free ( wheelState.ftdiContext );
    return 0;
  }


  wheel->interface = device->interface;
  wheelState.index = device->_devIndex;
  wheelState.wheelType = device->_devType;
  wheelState.currentPosition = 1;

  oaXagylWheelWarmReset ( wheel, 0 );

  if (( wheelState.numSlots = _getNumSlots ( wheel )) < 1 ) {
    fprintf ( stderr, "%s: invalid number of slots in filter wheel %s\n",
        __FUNCTION__, device->sysPath );
    return 0;
  }

  _oaInitFilterWheelFunctionPointers ( wheel );
  _xagylInitFunctionPointers ( wheel );

  wheel->features.coldReset = 1;
  wheel->features.warmReset = 1;
  wheel->features.speedControl = 1;

  return wheel;
}


static void
_xagylInitFunctionPointers ( oaFilterWheel* wheel )
{
  wheel->funcs.initWheel = oaXagylInitFilterWheel;
  wheel->funcs.getName = oaXagylWheelGetName;
  wheel->funcs.closeWheel = oaXagylWheelClose;
/*
  wheel->funcs.moveForward = _moveForward;
  wheel->funcs.moveBackward = _moveBackward;
*/
  wheel->funcs.moveTo = oaXagylMoveTo;
  wheel->funcs.warmReset = oaXagylWheelWarmReset;
  wheel->funcs.coldReset = oaXagylWheelColdReset;
  wheel->funcs.numSlots = oaXagylNumSlots;

  wheel->funcs.getSpeed = oaXagylGetWheelSpeed;
  wheel->funcs.setSpeed = oaXagylSetWheelSpeed;

}


int
oaXagylWheelWarmReset ( oaFilterWheel* wheel, int nodelay )
{
  return _doReset ( wheel, "r1", nodelay );
}


int
oaXagylWheelColdReset ( oaFilterWheel* wheel, int nodelay )
{
  return _doReset ( wheel, "r0", nodelay );
}


static int
_doReset ( oaFilterWheel* wheel, const char* cmd, int nodelay )
{
  /*
   * After sending "r0" or "r1" here there appears to be some variation
   * in output.
   * eg.
   *
   * Xagyl FW5125V2
   * FW 2.1.7
   * Initializing
   * Calibrating
   * P1
   *
   * Xagyl FW5125V1
   * FW 2.2.0
   * Initializing
   * Calibrating
   * RightCal -2
   * Found Sensor
   * Found Null
   * Found Midrange 514
   * Done
   * P1
   *
   * Restart
   * FW5125V2
   * FW 3.3.1
   * Initializing
   * P1
   *
   * I think it makes sense just to wait for "P1\n" to be returned by the
   * wheel to decide when the reset is complete
   */

  pthread_mutex_lock ( &wheelState.serialMutex );

  if ( nodelay ) {
    fprintf ( stderr, "%s: reset ignoring nodelay\n", __FUNCTION__ );
  }

  const char* expectedWords[] = { "Restart", "Xagyl", "FW", "Initializing",
      "Calibrating", "RightCal", "Found", "\n", 0
  };
  char buffer[50];
  int done = 0, error = 0, ret = 0, numRead;

  if ( _xagylWheelWrite ( wheelState.ftdiContext, cmd, 2 )) {
    fprintf ( stderr, "%s: write error on reset '%s' command\n", __FUNCTION__,
        cmd );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  do {
    if (( numRead = _xagylWheelRead ( wheelState.ftdiContext, buffer,
        50 )) < 0 ) {
      fprintf ( stderr, "%s: no data read from wheel interface\n",
          __FUNCTION__ );
      pthread_mutex_unlock ( &wheelState.serialMutex );
      return -1;
    }

    if ( numRead >= 5 && !strncmp ( buffer, "ERROR", 5 )) {
      done = error = 1;
    } else {
      if ( numRead == 3 && !strncmp ( buffer, "P1\n", 3 )) {
        done = 1;
      } else {
        int i = 0, found = 0;
        while ( expectedWords[i] && !found ) {
          if ( !strncmp ( buffer, expectedWords[i],
              strlen ( expectedWords[i] ))) {
            found = 1;
          }
          i++;
        }
        if ( !found ) {
          buffer [ numRead ] = 0;
          fprintf ( stderr, "%s: unexpected string '%s' (len %d) read\n",
              __FUNCTION__, buffer, numRead );
        }
      }
    }
  } while ( !done );

  if ( error ) {
    buffer[ numRead ] = 0;
    fprintf ( stderr, "%s: wheel returned error '%s'\n", __FUNCTION__,
        buffer );
    ret = -1;
  }
  
  pthread_mutex_unlock ( &wheelState.serialMutex );

  return ret;
}


int
_getNumSlots ( oaFilterWheel* wheel )
{
  char buffer[50];
  int numRead, numSlots;

  pthread_mutex_lock ( &wheelState.serialMutex );

  if ( _xagylWheelWrite ( wheelState.ftdiContext, "i8", 2 )) {
    fprintf ( stderr, "%s: write error on i8 command\n", __FUNCTION__ );
    return 0;
  }

  numRead = _xagylWheelRead ( wheelState.ftdiContext, buffer, 50 );
  pthread_mutex_unlock ( &wheelState.serialMutex );

  if ( numRead > 0 ) {
    if ( strncmp ( buffer, "FilterSlots ", 12 )) {
      fprintf ( stderr, "%s: failed to match expecting string 'FilterSlots '"
           ", got '%40s'\n", __FUNCTION__, buffer );
      return 0;
    }
    if ( sscanf ( buffer, "FilterSlots %d", &numSlots ) != 1 ) {
      fprintf ( stderr, "%s: Failed to match number of slots in '%s'\n",
          __FUNCTION__, buffer );
      return 0;
    }
    return numSlots;
  }

  fprintf ( stderr, "%s: no data read from wheel interface\n",
      __FUNCTION__ );
  return 0;
}


int
oaXagylWheelClose ( oaFilterWheel* wheel )
{
  wheelState.initialised = 0;
  ftdi_usb_close ( wheelState.ftdiContext );
  ftdi_free ( wheelState.ftdiContext );
  wheelState.ftdiContext = 0;
  return 0;
}
