/*****************************************************************************
 *
 * xagylInit-udev.c -- Initialise Xagyl filter wheels (udev)
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>
#include <openastro/filterwheel.h>

#include "xagylfw.h"
#include "oafwprivate.h"
#include "unimplemented.h"

#define	wheelState		wheel->_xagyl


static void _xagylInitFunctionPointers ( oaFilterWheel* );
static int  _getNumSlots ( oaFilterWheel* );
static int  _doReset ( oaFilterWheel*, const char*, int );

/**
 * Initialise a given filter wheel device
 */

oaFilterWheel*
oaXagylInitFilterWheel ( oaFilterWheelDevice* device )
{
  oaFilterWheel*	wheel;
  int			fwDesc;
  struct termios	tio;

  if (!( wheel = ( oaFilterWheel* ) malloc ( sizeof ( oaFilterWheel )))) {
    perror ( "malloc oaFilterWheel failed" );
    return 0;
  }

  OA_CLEAR ( wheelState );
  OA_CLEAR ( wheel->features );

  pthread_mutex_init ( &wheelState.serialMutex, 0 );

  ( void ) strcpy ( wheel->deviceName, device->deviceName );
  ( void ) strcpy ( wheelState.devicePath, device->sysPath );
  wheelState.initialised = 0;
  wheelState.index = -1;

  if (( fwDesc = open ( device->sysPath, O_RDWR | O_NOCTTY )) < 0 ) {
    fprintf ( stderr, "Can't open %s read-write, errno = %d\n",
        device->sysPath, errno );
    return 0;
  }

  if ( ioctl ( fwDesc, TIOCEXCL )) {
    int errnoCopy = errno;
    errno = 0;
    while (( close ( fwDesc ) < 0 ) && EINTR == errno );
    fprintf ( stderr, "%s: can't get lock on %s, errno = %d\n", __FUNCTION__,
        device->sysPath, errnoCopy );
    return 0;
  }

  if ( tcgetattr ( fwDesc, &tio )) {
    int errnoCopy = errno;
    errno = 0;
    while (( close ( fwDesc ) < 0 ) && EINTR == errno );
    fprintf ( stderr, "%s: can't get termio on %s, errno = %d\n", __FUNCTION__,
        device->sysPath, errnoCopy );
    return 0;
  }

  tio.c_cflag &= ~PARENB; // no parity
  tio.c_cflag &= ~CSTOPB; // 1 stop bit
  tio.c_cflag |= CLOCAL | CREAD | CS8;
  cfsetospeed ( &tio, B9600 );
  cfsetispeed ( &tio, B9600 );
  tcflush ( fwDesc, TCIFLUSH );
  if ( tcsetattr ( fwDesc, TCSANOW, &tio )) {
    int errnoCopy = errno;
    errno = 0;
    while (( close ( fwDesc ) < 0 ) && EINTR == errno );
    fprintf ( stderr, "%s: can't set termio on %s, errno = %d\n", __FUNCTION__,
        device->sysPath, errnoCopy );
    return 0;
  }

  wheelState.fd = fwDesc;
  wheel->interface = device->interface;
  wheelState.index = device->_devIndex;
  wheelState.wheelType = device->_devType;
  wheelState.currentPosition = 1;

  oaXagylWheelWarmReset ( wheel, 0 );

  if (( wheelState.numSlots = _getNumSlots ( wheel )) < 1 ) {
    fprintf ( stderr, "%s: invalid number of slots in filter wheel %s\n",
        __FUNCTION__, device->sysPath );
    return 0;
  }

  _oaInitFilterWheelFunctionPointers ( wheel );
  _xagylInitFunctionPointers ( wheel );

  wheel->features.coldReset = 1;
  wheel->features.warmReset = 1;
  wheel->features.speedControl = 1;

  return wheel;
}


static void
_xagylInitFunctionPointers ( oaFilterWheel* wheel )
{
  wheel->funcs.initWheel = oaXagylInitFilterWheel;
  wheel->funcs.getName = oaXagylWheelGetName;
  wheel->funcs.closeWheel = oaXagylWheelClose;
/*
  wheel->funcs.moveForward = _moveForward;
  wheel->funcs.moveBackward = _moveBackward;
*/
  wheel->funcs.moveTo = oaXagylMoveTo;
  wheel->funcs.warmReset = oaXagylWheelWarmReset;
  wheel->funcs.coldReset = oaXagylWheelColdReset;
  wheel->funcs.numSlots = oaXagylNumSlots;

  wheel->funcs.getSpeed = oaXagylGetWheelSpeed;
  wheel->funcs.setSpeed = oaXagylSetWheelSpeed;

}


int
oaXagylWheelWarmReset ( oaFilterWheel* wheel, int nodelay )
{
  return _doReset ( wheel, "r1", nodelay );
}


int
oaXagylWheelColdReset ( oaFilterWheel* wheel, int nodelay )
{
  return _doReset ( wheel, "r0", nodelay );
}


static int
_doReset ( oaFilterWheel* wheel, const char* cmd, int nodelay )
{
  /*
   * After sending "r0" or "r1" here there appears to be some variation
   * in output.
   * eg.
   *
   * Xagyl FW5125V2
   * FW 2.1.7
   * Initializing
   * Calibrating
   * P1
   *
   * Xagyl FW5125V1
   * FW 2.2.0
   * Initializing
   * Calibrating
   * RightCal -2
   * Found Sensor
   * Found Null
   * Found Midrange 514
   * Done
   * P1
   *
   * Restart
   * FW5125V2
   * FW 3.3.1
   * Initializing
   * P1
   *
   * I think it makes sense just to wait for "P1\n" to be returned by the
   * wheel to decide when the reset is complete
   */

  pthread_mutex_lock ( &wheelState.serialMutex );

  if ( nodelay ) {
    fprintf ( stderr, "%s: reset ignoring nodelay\n", __FUNCTION__ );
  }

  const char* expectedWords[] = { "Restart", "Xagyl", "FW", "Initializing",
      "Calibrating", "RightCal", "Found", "\n", 0
  };
  char buffer[50];
  int done = 0, error = 0, ret = 0, numRead;

  tcflush ( wheelState.fd, TCIFLUSH );

  if ( _xagylWheelWrite ( wheelState.fd, cmd, 2 )) {
    fprintf ( stderr, "%s: write error on reset '%s' command\n", __FUNCTION__,
        cmd );
    pthread_mutex_unlock ( &wheelState.serialMutex );
    return -1;
  }

  do {
    if (( numRead = _xagylWheelRead ( wheelState.fd, buffer, 50 )) < 0 ) {
      fprintf ( stderr, "%s: no data read from wheel interface\n",
          __FUNCTION__ );
      pthread_mutex_unlock ( &wheelState.serialMutex );
      return -1;
    }

    if ( numRead >= 5 && !strncmp ( buffer, "ERROR", 5 )) {
      done = error = 1;
    } else {
      if ( numRead == 3 && !strncmp ( buffer, "P1\n", 3 )) {
        done = 1;
      } else {
        int i = 0, found = 0;
        while ( expectedWords[i] && !found ) {
          if ( !strncmp ( buffer, expectedWords[i],
              strlen ( expectedWords[i] ))) {
            found = 1;
          }
          i++;
        }
        if ( !found ) {
          buffer [ numRead ] = 0;
          fprintf ( stderr, "%s: unexpected string '%s' (len %d) read\n",
              __FUNCTION__, buffer, numRead );
        }
      }
    }
  } while ( !done );

  if ( error ) {
    buffer[ numRead ] = 0;
    fprintf ( stderr, "%s: wheel returned error '%s'\n", __FUNCTION__,
        buffer );
    ret = -1;
  }
  
  pthread_mutex_unlock ( &wheelState.serialMutex );

  return ret;
}


int
_getNumSlots ( oaFilterWheel* wheel )
{
  char buffer[50];
  int numRead, numSlots;

  pthread_mutex_lock ( &wheelState.serialMutex );

  tcflush ( wheelState.fd, TCIFLUSH );

  if ( _xagylWheelWrite ( wheelState.fd, "i8", 2 )) {
    fprintf ( stderr, "%s: write error on i8 command\n", __FUNCTION__ );
    return 0;
  }

  numRead = _xagylWheelRead ( wheelState.fd, buffer, 50 );
  pthread_mutex_unlock ( &wheelState.serialMutex );

  if ( numRead > 0 ) {
    if ( strncmp ( buffer, "FilterSlots ", 12 )) {
      fprintf ( stderr, "%s: failed to match expecting string 'FilterSlots '"
           ", got '%40s'\n", __FUNCTION__, buffer );
      return 0;
    }
    if ( sscanf ( buffer, "FilterSlots %d", &numSlots ) != 1 ) {
      fprintf ( stderr, "%s: Failed to match number of slots in '%s'\n",
          __FUNCTION__, buffer );
      return 0;
    }
    return numSlots;
  }

  fprintf ( stderr, "%s: no data read from wheel interface\n",
      __FUNCTION__ );
  return 0;
}


int
oaXagylWheelClose ( oaFilterWheel* wheel )
{
  wheelState.initialised = 0;
  return close ( wheelState.fd );
}
