/*****************************************************************************
 *
 * camera.cc -- camera interface class
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include "camera.h"
#include "configuration.h"
#include "state.h"
#include "version.h"

#define	DEFAULT_FRAME_TIME	100

#define cameraFuncs	cameraContext->funcs
#define cameraControls	cameraContext->controls
#define cameraFeatures	cameraContext->features

Camera::Camera()
{
  initialised = 0;
  captureThreadStarted = 0;
  configUpdated = 0;
  configFlags[0] = 0;
  configFlags[1] = 0;
  configFlags[2] = 0;
  configROIX = configROIY = configFPSNumerator = configFPSDenominator = 0;
  pthread_mutex_init ( &configMutex, 0 );
  pthread_cond_init ( &stopCapture, 0 );
  frameTime = DEFAULT_FRAME_TIME;
  framePixelFormat = OA_PIX_FMT_RGB24; // the default
}


Camera::~Camera()
{
  stop();
  // FIX ME -- wait for mutex?
  disconnect();
  pthread_mutex_destroy ( &configMutex );
  pthread_cond_destroy ( &stopCapture );
}


void
Camera::populateControlValue ( oaControlValue* cp, uint32_t c, int64_t v )
{
  cp->valueType = cameraControls[ c ];
  switch ( cameraControls[ c ] ) {
    case OA_CTRL_TYPE_INT32:
      cp->int32 = v & 0xffffffff;
      break;
    case OA_CTRL_TYPE_BOOLEAN:
      cp->boolean = v ? 1 : 0;
      break;
    case OA_CTRL_TYPE_MENU:
      cp->menu = v & 0xffffffff;
      break;
    case OA_CTRL_TYPE_BUTTON:
      break;
    case OA_CTRL_TYPE_INT64:
      cp->int64 = v;
      break;
    case OA_CTRL_TYPE_DISCRETE:
      cp->discrete = v & 0xffffffff;
      break;
    default:
      qWarning() << __FUNCTION__ << " called with invalid control type " <<
        cameraControls[ c ] << " for control " << c;
  }
}


int64_t
Camera::unpackControlValue ( oaControlValue *cp )
{
  int64_t res;

  switch ( cp->valueType ) {
    case OA_CTRL_TYPE_INT32:
      res = cp->int32;
      break;
    case OA_CTRL_TYPE_BOOLEAN:
      res = cp->boolean;
      break;
    case OA_CTRL_TYPE_MENU:
      res = cp->menu;
      break;
    case OA_CTRL_TYPE_READONLY:
      res = cp->readonly;
      break;
    case OA_CTRL_TYPE_INT64:
      res = cp->int64;
      // FIX ME -- because at the moment Qt controls can only handle 32-bit
      // values
      res &= 0xffffffff;
      break;
    case OA_CTRL_TYPE_DISCRETE:
      res = cp->discrete;
      break;
    default:
      qWarning() << __FUNCTION__ << " called with invalid control type " <<
        cameraControls[ cp->valueType ];
      res = -1;
  }

  return res;
}


int
Camera::has16Bit ( void )
{
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  // FIX ME -- this assumes that we don't have a 16-bit mode if the bit
  // depth can't be changed which may not be true at all.

  if ( !cameraControls[ OA_CAM_CTRL_BIT_DEPTH ] ) {
    return 0;
  }
  populateControlValue ( &v, OA_CAM_CTRL_BIT_DEPTH, 16 );
  return ( OA_ERR_NONE == cameraFuncs.testControl ( cameraContext,
    OA_CAM_CTRL_BIT_DEPTH, &v )) ? 1 : 0;
}


int
Camera::hasBinning ( int64_t factor )
{
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  if ( !cameraControls[ OA_CAM_CTRL_BINNING ] ) {
    return 0;
  }
  populateControlValue ( &v, OA_CAM_CTRL_BINNING, factor );
  return ( OA_ERR_NONE == cameraFuncs.testControl ( cameraContext,
      OA_CAM_CTRL_BINNING, &v )) ? 1 : 0;
}


int
Camera::testROISize ( unsigned int xRes, unsigned int yRes,
    unsigned int* suggX, unsigned int* suggY )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }

  return cameraFuncs.testROISize ( cameraContext, xRes, yRes, suggX, suggY );
}


int
Camera::hasROI ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return cameraFeatures.ROI;
}


int
Camera::hasFrameRateSupport ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return cameraFeatures.frameRates;
}


int
Camera::hasFixedFrameRates ( int xRes, int yRes )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }

  return cameraFuncs.hasFixedFrameRates ( cameraContext, xRes, yRes );
}


int
Camera::hasControl ( int control )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }

  if ( control >= 0 && control < OA_CAM_CTRL_LAST_P1 ) {
    return cameraControls[ control ];
  }
  qWarning() << __FUNCTION__ << " unrecognised control" << control;
  return 0;
}


int
Camera::hasAuto ( int control )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }

  if ( control >= 0 && control < OA_CAM_CTRL_LAST_P1 ) {
    return cameraFuncs.hasAuto ( cameraContext, control );
  }
  qWarning() << __FUNCTION__ << " unrecognised control";
  return 0;
}


int
Camera::isAuto ( int control )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  } 

  if ( control >= 0 && control < OA_CAM_CTRL_LAST_P1 ) {
    return cameraFuncs.isAuto ( cameraContext, control );
  }
  qWarning() << __FUNCTION__ << " unrecognised control";
  return 0;
}


int
Camera::isColour ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return ( cameraFeatures.rawMode | cameraFeatures.demosaicMode );
}


float
Camera::getTemperature ( void )
{
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  if ( !cameraControls[ OA_CAM_CTRL_TEMPERATURE ]) {
    return -273.15;
  }
  cameraFuncs.readControl ( cameraContext, OA_CAM_CTRL_TEMPERATURE, &v );
  float tempReading = v.int32;
  tempReading /= 10;
  return tempReading;
}


const char*
Camera::name()
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return "";
  }
  return cameraFuncs.getName ( cameraContext );
}


void
Camera::controlRange ( int control, int64_t* min, int64_t* max,
    int64_t* step, int64_t* def )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
  } else {
    cameraFuncs.getControlRange ( cameraContext, control, min, max, step, def );
  }
  // FIX ME -- this is because the sliders can't handle more than 32-bit
  // values
  *min &= 0xffffffff;
  *max &= 0xffffffff;
  *step &= 0xffffffff;
  *def &= 0xffffffff;
  return;
}


// FIX ME -- might be nice to make this a tidier type at some point.  vector?
const FRAMESIZE*
Camera::frameSizes ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return cameraFuncs.enumerateFrameSizes ( cameraContext );
}


// FIX ME -- might be nice to make this a tidier type at some point.  vector?
const FRAMERATE*
Camera::frameRates ( int xRes, int yRes)
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return cameraFuncs.enumerateFrameRates ( cameraContext, xRes, yRes );
}


// FIX ME -- might be nice to make this a tidier type at some point.  vector?
int
Camera::listConnected ( oaCameraDevice** devs )
{
  return ( oaGetCameras ( devs ));
}


// FIX ME -- handling of the return code here is really scabby.  Do it
// properly at some point

int
Camera::initialise ( oaCameraDevice* device )
{
  int ret;

  disconnect();

  if ( !device ) {
    qWarning() << "device is null!";
    return -1;
  }

  // At the moment this is really just for OSX
  if ( device->hasLoadableFirmware ( device ) &&
      !device->isFirmwareLoaded ( device )) {

    QMessageBox* loading = new QMessageBox ( QMessageBox::NoIcon,
        APPLICATION_NAME, tr ( "Attempting to load camera firmware" ),
        QMessageBox::NoButton, state.mainWindow );
    QAbstractButton* b = loading->button ( QMessageBox::Ok );
    if ( b ) {
      loading->removeButton ( b );
    }
    QApplication::setOverrideCursor ( Qt::WaitCursor );
    loading->show();
    ret = device->loadFirmware ( device );
    QApplication::restoreOverrideCursor();
    loading->hide();
    delete loading;

    if ( ret != OA_ERR_NONE ) {
      switch ( ret ) {

        case OA_ERR_RESCAN_BUS:
          return 1;
          break;

        case OA_ERR_MANUAL_FIRMWARE:
          QMessageBox::warning ( state.mainWindow, APPLICATION_NAME,
              tr ( "The firmware could not be loaded.  It must be loaded "
              "manually." ));
          break;

        case OA_ERR_FXLOAD_NOT_FOUND:
          QMessageBox::warning ( state.mainWindow, APPLICATION_NAME,
              tr ( "The firmware could not be loaded.  The fxload utility "
              "was not found." ));
          break;

        case OA_ERR_FXLOAD_ERROR:
          QMessageBox::warning ( state.mainWindow, APPLICATION_NAME,
              tr ( "The firmware could not be loaded.  The fxload utility "
              "returned an error." ));
          break;

        case OA_ERR_FIRMWARE_UNKNOWN:
          QMessageBox::warning ( state.mainWindow, APPLICATION_NAME,
              tr ( "The firmware could not be found." ));
          break;

        default:
          QMessageBox::warning ( state.mainWindow, APPLICATION_NAME,
              tr ( "Unexpected error loading firmware." ));
          break;
      }

      return -1;
    }
  }

  if (( cameraContext = device->initCamera ( device ))) {
    initialised = 1;
    framePixelFormat = cameraFuncs.getFramePixelFormat ( cameraContext, 0 );
    return 0;
  }
  return -1;
}


void
Camera::disconnect ( void )
{
  if ( initialised ) {
    cameraFuncs.closeCamera ( cameraContext );
    initialised = 0;
    captureThreadStarted = 0;
    frameTime = DEFAULT_FRAME_TIME;
  }
}


int
Camera::setControl ( int control, int64_t value )
{
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
  }

  // Some cameras get upset if we try to change the controls before
  // the camera is fully configured so we just store them here
  // for later use (or when the capture thread gets around to looking
  // at them
  // The exception to this rule is binning, which is forced to the
  // camera straight away because it can affect the information we
  // subsequently want back from the camera library.

  configValues[ control ] = value;
  if ( OA_CAM_CTRL_BINNING == control ) {
    populateControlValue ( &v, OA_CAM_CTRL_BINNING, value );
    cameraFuncs.setControl ( cameraContext, control, &v );
  } else {
    pthread_mutex_lock ( &configMutex );
    setConfigUpdateFlag ( control );
    if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == control ) {
      frameTime = value;
    }
    pthread_mutex_unlock ( &configMutex );
  }
  return 0;
}


int64_t
Camera::readControl ( int control )
{
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }

  cameraFuncs.readControl ( cameraContext, control, &v );
  return unpackControlValue ( &v );
}


int
Camera::setROI ( int x, int y )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }

  int ret = 0;
  int restart = captureThreadStarted;
  stop();
  configROIX = x;
  configROIY = y;
  flushSettings();
  if ( cameraFeatures.hasReset ) {
    cameraFuncs.resetCamera ( cameraContext );
  }
  if ( restart ) {
    ret = start();
  }
  frameRateChangeDelayed = 0;
  return ret;
}


int
Camera::setFrameInterval ( int numerator, int denominator )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }

  int restart = 0;
  if ( !frameRateChangeDelayed ) {
    restart = captureThreadStarted;
    stop();
  }
  if ( denominator ) {
    frameTime = 1000000 * (( float ) numerator / ( float ) denominator );
  }
  configFPSNumerator = numerator;
  configFPSDenominator = denominator;
  if ( !frameRateChangeDelayed ) {
    if ( cameraFeatures.hasReset ) {
      cameraFuncs.resetCamera ( cameraContext );
    }
    if ( restart ) {
      return start();
    }
  }
  return 0;
}


void
Camera::setConfigUpdateFlag ( int control )
{
  control--;
  int i = control / 32;
  control %= 32;
  if ( i < 3 ) {
    configFlags[i] |= 1 << control;
    configUpdated = 1;
  }
}


int
Camera::clearConfigUpdateFlag ( int control )
{
  control--;
  int i = control / 32;
  control %= 32;
  if ( i < 3 ) {
    configFlags[i] &= ( 0xffffffff ^ ( 1 << control ));
  }
  return ( 0 == configFlags[0] && 0 == configFlags[1] ) ? 0 : 1;
}


int
Camera::testConfigFlag ( int control )
{
  int res = 0, i;

  control--;
  i = control / 32;
  control %= 32;
  if ( i < 3 ) {
    res = configFlags[i] & ( 1 << control );
  }
  return res ? 1 : 0;
}

int
Camera::start ( void )
{
  int		ret;
  START_PARMS	parms;
  
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }
  parms.size.x = configROIX;
  parms.size.y = configROIY;
  parms.rate.numerator = configFPSNumerator;
  parms.rate.denominator = configFPSDenominator;

  if (( ret = cameraFuncs.startCaptureExtended ( cameraContext, &parms ))) {
    return ret;
  }

  // now set each control that we have flagged as changed

  flushSettings();

  // now create a new thread and continue

  captureThreadStarted = 0;
  captureThreadExit = 0;
  captureThreadDataInUse = 1;
  pthread_create ( &captureThread, 0, &Camera::capture, this );

  captureThreadStarted = 1;

  return 0;
}


void
Camera::stop ( void )
{
  // We don't care if the camera isn't initialised for this function.
  // Nothing will happen anyhow

  if ( !captureThreadStarted ) {
    return;
  }

  pthread_mutex_lock ( &configMutex );
  captureThreadExit = 1;
  pthread_cond_signal ( &stopCapture );
  pthread_mutex_unlock ( &configMutex );
  pthread_join ( captureThread, 0 );
  captureThreadStarted = 0;

  cameraFuncs.stopCapture ( cameraContext );
}


void*
Camera::capture ( void *param )
{
  int		quitThread;
  Camera*	camera = ( Camera* ) param;
  void*		imageBuffer;

  while ( true ) {
    pthread_mutex_lock ( &camera->configMutex );
    quitThread = camera->captureThreadExit;
    pthread_mutex_unlock ( &camera->configMutex );
    if ( quitThread ) {
      break;
    }

    camera->cameraFuncs.startReadFrame ( camera->cameraContext,
        camera->frameTime );

    while ( true ) {
      int ret;
      ret = camera->cameraFuncs.readFrame ( camera->cameraContext,
          &imageBuffer );
      // -1 == error, 0 = timeout, > 0 ok
      if ( -1 == ret ) {
        if ( errno != EINTR ) {
          break;
        }
      } else {
        if ( ret ) {
          emit camera->frameGrabbed ( imageBuffer, ret );

          pthread_mutex_lock ( &camera->configMutex );
          while ( camera->captureThreadDataInUse &&
              !camera->captureThreadExit ) {
            pthread_cond_wait ( &camera->stopCapture, &camera->configMutex );
          }
          camera->captureThreadDataInUse = 1;
          pthread_mutex_unlock ( &camera->configMutex );
          camera->cameraFuncs.finishReadFrame ( camera->cameraContext );
        }
        break;
      }
    }

    // process any configuration updates

    pthread_mutex_lock ( &camera->configMutex );
    if ( camera->configUpdated ) {
      int c, r = 0;
      oaControlValue v;

      for ( c = 1; 0 == r && ( c < OA_CAM_CTRL_LAST_P1 ); c++ ) {
        if ( camera->testConfigFlag ( c )) {
          camera->populateControlValue ( &v, c, camera->configValues[c] );
          camera->cameraFuncs.setControl ( camera->cameraContext, c, &v );
          r = camera->clearConfigUpdateFlag ( c );
        }
      }
/*
      camera->flushSettings();
*/
      camera->configUpdated = 0;
    }
    pthread_mutex_unlock ( &camera->configMutex );
  }
  return 0;
}


void
Camera::releaseImageData ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
  }
  if ( captureThreadStarted ) {
    pthread_mutex_lock ( &configMutex );
    captureThreadDataInUse = 0;
    pthread_cond_signal ( &stopCapture );
    pthread_mutex_unlock ( &configMutex );
  }
}


void
Camera::delayFrameRateChanges ( void )
{
  frameRateChangeDelayed = 1;
}


int
Camera::captureRunning ( void )
{
  return captureThreadStarted;
}


int
Camera::videoFramePixelFormat ( void )
{
  return framePixelFormat;
}


void
Camera::flushSettings ( void )
{
  int c;
  oaControlValue v;

  for ( c = 1; c < OA_CAM_CTRL_LAST_P1; c++ ) {
    if ( testConfigFlag ( c )) {
      populateControlValue ( &v, c, configValues[c] );
      cameraFuncs.setControl ( cameraContext, c, &v );
      clearConfigUpdateFlag ( c );
    }
  }
  configUpdated = 0;
}


int
Camera::setBitDepth ( int depth )
{
  int ret = 0;
  int restart = captureThreadStarted;
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }

  stop();
  populateControlValue ( &v, OA_CAM_CTRL_BIT_DEPTH, depth );
  cameraFuncs.setControl ( cameraContext, OA_CAM_CTRL_BIT_DEPTH, &v );
  flushSettings();
  if ( cameraFeatures.hasReset ) {
    cameraFuncs.resetCamera ( cameraContext );
  }
  framePixelFormat = cameraFuncs.getFramePixelFormat ( cameraContext, 0 );
  if ( restart ) {
    ret = start();
  }
  return ret;
}


int
Camera::setRawMode ( int enabled )
{
  int ret = 0;
  int restart = captureThreadStarted;
  oaControlValue v;

  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }

  stop();
  populateControlValue ( &v, OA_CAM_CTRL_COLOUR_MODE,
      enabled ? OA_COLOUR_MODE_RAW : OA_COLOUR_MODE_NONRAW );
  cameraFuncs.setControl ( cameraContext, OA_CAM_CTRL_COLOUR_MODE, &v );
  flushSettings();
  if ( cameraFeatures.hasReset ) {
    cameraFuncs.resetCamera ( cameraContext );
  }
  framePixelFormat = cameraFuncs.getFramePixelFormat ( cameraContext, 0 );
  if ( restart ) {
    ret = start();
  }
  return ret;
}


int
Camera::videoFramePixelFormat ( int depth )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return -1;
  }
  return cameraFuncs.getFramePixelFormat ( cameraContext, depth );
}


int
Camera::isInitialised ( void )
{
  return initialised;
}


int
Camera::hasRawMode ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
  }
  return initialised ? cameraFeatures.rawMode : 0;
}


int
Camera::hasDemosaicMode ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
  }
  return initialised ? cameraFeatures.demosaicMode : 0;
}


const char*
Camera::getMenuString ( int control, int index )
{
 if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return "";
  }
  return cameraFuncs.getMenuString ( cameraContext, control, index );
}


int
Camera::getAWBManualSetting ( void )
{
 if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with camera uninitialised";
    return 0;
  }
  return cameraFuncs.getAutoWBManualSetting ( cameraContext );
}
