/*****************************************************************************
 *
 * controlWidget.cc -- class for the control widget in the UI
 *
 * Copyright 2013,2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <QtGui>

extern "C" {
#include <openastro/camera.h>
}

#include "configuration.h"
#include "controlWidget.h"
#include "state.h"


ControlWidget::ControlWidget ( QWidget* parent ) : QGroupBox ( parent )
{
  gainLabel = new QLabel ( tr ( "Gain" ), this );
  exposureLabel = new QLabel ( tr ( "Exposure (ms)" ), this );
  framerateLabel = new QLabel ( tr ( "Framerate (fps)" ), this );
  expRangeLabel = new QLabel ( tr ( "Exp. Range (ms)" ), this );
  autoLabel = new QLabel ( tr ( "Auto" ), this );

  exposureSlider = new QSlider ( Qt::Horizontal, this );
  framerateSlider = new QSlider ( Qt::Horizontal, this );

  exposureSpinbox = new QSpinBox ( this );
  framerateMenu = new QComboBox ( this );

  selectableControlMenu[0] = new QComboBox ( this );
  selectableControlMenu[1] = new QComboBox ( this );
  selectableControlMenu[0]->setToolTip ( tr ( "Select control to display" ));
  selectableControlMenu[1]->setToolTip ( tr ( "Select control to display" ));

  selectableControlsAllowed[0][0] = config.selectableControl[0];
  selectableControlsAllowed[1][0] = config.selectableControl[1];
  selectableControlIndexes[0][ config.selectableControl[0]] = 0;
  selectableControlIndexes[1][ config.selectableControl[1]] = 1;

  if ( config.selectableControl[0] >= 0 ) {
    selectableControlMenu[0]->addItem (
        tr ( oaCameraControlLabel[ config.selectableControl[0]] ));
    if ( config.selectableControl[1] >= 0 ) {
      selectableControlMenu[1]->addItem (
          tr ( oaCameraControlLabel[ config.selectableControl[1]] ));
    }
  }

  connect ( selectableControlMenu[0], SIGNAL( currentIndexChanged ( int )),
      this, SLOT ( setSelectableControl1 ( int )));
  connect ( selectableControlMenu[1], SIGNAL( currentIndexChanged ( int )),
      this, SLOT ( setSelectableControl2 ( int )));
  ignoreSelectableControlChanges = 1;

  expMenu = new QComboBox ( this );
  QStringList exposures;
  exposures << "0 - 5" << "5 - 30" << "30 - 100" << "100 - 1000";
  exposures << "1000 - 10000";
  minSettings[0] = 0;
  minSettings[1] = 5;
  minSettings[2] = 30;
  minSettings[3] = 100;
  maxSettings[0] = 5;
  maxSettings[1] = 30;
  maxSettings[2] = 100;
  maxSettings[3] = 1000;
  expMenu->addItems ( exposures );
  expMenu->setCurrentIndex ( config.exposureMenuOption );
  usingAbsoluteExposure = 0;
  ignoreExposureMenuChanges = 0;
  useExposureDropdown = 1;

  grid = new QGridLayout ( this );

  grid->addWidget ( autoLabel, 0, 1 );

  grid->addWidget ( gainLabel, 1, 0 );
  grid->addWidget ( selectableControlMenu[0], 2, 0 );
  if ( -1 == config.selectableControl[0] ) {
    selectableControlMenu[0]->hide();
  }
  grid->addWidget ( selectableControlMenu[1], 3, 0 );
  if ( -1 == config.selectableControl[1] ) {
    selectableControlMenu[1]->hide();
  }
  grid->addWidget ( framerateLabel, 4, 0 );
  grid->addWidget ( exposureLabel, 5, 0 );
  grid->addWidget ( expRangeLabel, 6, 0 );
  
  grid->addWidget ( framerateSlider, 4, 2 );
  grid->addWidget ( exposureSlider, 5, 2 );
  grid->addWidget ( expMenu, 6, 2 );

  grid->addWidget ( framerateMenu, 4, 3 );
  grid->addWidget ( exposureSpinbox, 5, 3 );

  grid->setColumnStretch ( 2, 4 );

  // Now for the selectable slider we create a label, slider and spinbox
  // for each control relevant control.  When the control is selected we'll
  // display those control widgets.

  sliderSignalMapper = new QSignalMapper ( this );
  checkboxSignalMapper = new QSignalMapper ( this );
  for ( uint8_t c = 1; c < OA_CAM_CTRL_LAST_P1; c++ ) {
    selectableControlSlider[ c ] = 0;
    selectableControlSpinbox[ c ] = 0;
    selectableControlCheckbox[ c ] = 0;

    // This is a bit hacky.  We create the control sliders and checkboxes
    // on the assumption that they will exist when a camera is connected
    // later, but we've no idea if that is actually the case or not

    if ( c != OA_CAM_CTRL_EXPOSURE && c != OA_CAM_CTRL_EXPOSURE_ABSOLUTE ) {
      selectableControlSlider[ c ] = new QSlider ( Qt::Horizontal, this );
      selectableControlSlider[ c ]->setFocusPolicy ( Qt::TabFocus );
      selectableControlSpinbox[ c ] = new QSpinBox ( this );
      if ( c != config.selectableControl[0] &&
          c != config.selectableControl[1] ) {
        selectableControlSlider[ c ]->hide();
        selectableControlSpinbox[ c ]->hide();
      }

      sliderSignalMapper->setMapping ( selectableControlSpinbox[c], c );

      connect ( selectableControlSlider[ c ], SIGNAL( sliderMoved ( int )),
          selectableControlSpinbox[ c ], SLOT( setValue( int )));
      connect ( selectableControlSlider[ c ], SIGNAL( valueChanged ( int )),
          selectableControlSpinbox[ c ], SLOT( setValue( int )));
      connect ( selectableControlSpinbox[ c ], SIGNAL( valueChanged ( int )),
          selectableControlSlider[ c ], SLOT( setValue( int )));

      connect ( selectableControlSpinbox[ c ], SIGNAL( valueChanged ( int )),
          sliderSignalMapper, SLOT( map()));
    }

    if ( oaGetAutoForControl ( c )) {
      selectableControlCheckbox[ c ] = new QCheckBox ( "", this );
      selectableControlCheckbox[ c ]->hide();
      checkboxSignalMapper->setMapping ( selectableControlCheckbox[ c ], c );
      connect ( selectableControlCheckbox[ c ], SIGNAL (
          stateChanged ( int )), checkboxSignalMapper, SLOT( map()));
    }
  }
  connect ( sliderSignalMapper, SIGNAL( mapped ( int )), this,
      SLOT ( updateSelectableControl ( int )));
  connect ( checkboxSignalMapper, SIGNAL( mapped ( int )), this,
      SLOT ( updateSelectableCheckbox ( int )));

  if ( selectableControlCheckbox[ OA_CAM_CTRL_GAIN ] ) {
    grid->addWidget (
        selectableControlCheckbox[ OA_CAM_CTRL_GAIN ], 1, 1 );
  }
  if ( selectableControlSlider[ OA_CAM_CTRL_GAIN ] ) {
    grid->addWidget ( selectableControlSlider[ OA_CAM_CTRL_GAIN ], 1, 2 );
    grid->addWidget ( selectableControlSpinbox[ OA_CAM_CTRL_GAIN ], 1, 3 );
  }

  if ( config.selectableControl[0] >= 0 ) {
    if ( selectableControlCheckbox[ config.selectableControl[0]] ) {
      grid->addWidget (
          selectableControlCheckbox[ config.selectableControl[0]], 2, 1 );
    }
    grid->addWidget (
        selectableControlSlider[ config.selectableControl[0]], 2, 2 );
    grid->addWidget (
        selectableControlSpinbox[ config.selectableControl[0]], 2, 3 );
  }
  if ( config.selectableControl[1] >= 0 ) {
    if ( selectableControlCheckbox[ config.selectableControl[1]] ) {
      grid->addWidget (
          selectableControlCheckbox[ config.selectableControl[1]], 3, 1 );
    }
    grid->addWidget (
        selectableControlSlider[ config.selectableControl[1]], 3, 2 );
    grid->addWidget (
        selectableControlSpinbox[ config.selectableControl[1]], 3, 3 );
  }

  framerateLabel->hide();
  framerateSlider->hide();
  framerateMenu->hide();

  if ( !config.dockableControls ) {
    setTitle ( tr ( "Control" ));
  }
  setLayout ( grid );

  connect ( selectableControlSlider[ OA_CAM_CTRL_GAIN ], SIGNAL(
      sliderMoved ( int )), selectableControlSpinbox[ OA_CAM_CTRL_GAIN ],
      SLOT( setValue( int )));
  connect ( selectableControlSlider[ OA_CAM_CTRL_GAIN ], SIGNAL(
      valueChanged ( int )), selectableControlSpinbox[ OA_CAM_CTRL_GAIN ],
      SLOT( setValue( int )));
  connect ( selectableControlSpinbox[ OA_CAM_CTRL_GAIN ], SIGNAL(
      valueChanged ( int )), selectableControlSlider[ OA_CAM_CTRL_GAIN ],
      SLOT( setValue( int )));
  connect ( selectableControlSpinbox[ OA_CAM_CTRL_GAIN ], SIGNAL(
      valueChanged ( int )), this, SLOT( updateGain ( int )));
  selectableControlSlider[ OA_CAM_CTRL_GAIN ]->setFocusPolicy ( Qt::TabFocus );

  exposureSlider->setFocusPolicy ( Qt::TabFocus );
  connect ( exposureSlider, SIGNAL( sliderMoved ( int )), exposureSpinbox,
      SLOT( setValue( int )));
  connect ( exposureSlider, SIGNAL( valueChanged ( int )), exposureSpinbox,
      SLOT( setValue( int )));
  connect ( exposureSpinbox, SIGNAL( valueChanged ( int )), exposureSlider,
      SLOT( setValue( int )));
  connect ( exposureSpinbox, SIGNAL( valueChanged ( int )), this,
      SLOT( updateExposure( int )));

  framerateSlider->setFocusPolicy ( Qt::TabFocus );
  connect ( framerateSlider, SIGNAL( sliderMoved ( int )), framerateMenu,
      SLOT( setCurrentIndex( int )));
  connect ( framerateSlider, SIGNAL( valueChanged ( int )), framerateMenu,
      SLOT( setCurrentIndex( int )));
  connect ( framerateSlider, SIGNAL( sliderReleased()), this,
      SLOT( frameRateChanged()));
  connect ( framerateMenu, SIGNAL( currentIndexChanged ( int )),
      framerateSlider, SLOT( setValue( int )));
  connect ( framerateMenu, SIGNAL( currentIndexChanged ( int )), this,
      SLOT( frameRateChanged()));

  connect ( expMenu, SIGNAL( currentIndexChanged ( int )), this,
      SLOT( exposureMenuChanged ( int )));

  ignoreFrameRateChanges = ignoreSelectableControlChanges = 0;
}


ControlWidget::~ControlWidget()
{
  state.mainWindow->destroyLayout (( QLayout* ) grid );
}


void
ControlWidget::configure ( void )
{
  int64_t	min, max, step, def;

#define DISABLE_AUTO
#ifdef DISABLE_AUTO
  // These ones we just want off all the time
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_GAIN )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_GAIN, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_GAIN ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_GAIN, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_HUE_AUTO )) {
    state.camera->setControl ( OA_CAM_CTRL_HUE_AUTO, 0 );
    config.controlValues[ OA_CAM_CTRL_HUE_AUTO ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_HUE_AUTO, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_BRIGHTNESS )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_BRIGHTNESS, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_BRIGHTNESS ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_BRIGHTNESS, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_EXPOSURE )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_EXPOSURE, OA_EXPOSURE_MANUAL );
    config.controlValues[ OA_CAM_CTRL_AUTO_EXPOSURE ] = OA_EXPOSURE_MANUAL;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_EXPOSURE, OA_EXPOSURE_MANUAL );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_GAMMA )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_GAMMA, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_GAMMA ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_GAMMA, 0 );
  }
  int AWBtype = state.camera->hasControl ( OA_CAM_CTRL_AUTO_WHITE_BALANCE );
  if ( AWBtype ) {
    int AWBManual = 0;
    if ( OA_CTRL_TYPE_MENU == AWBtype ) {
      AWBManual = state.camera->getAWBManualSetting();
    }
    state.camera->setControl ( OA_CAM_CTRL_AUTO_WHITE_BALANCE, AWBManual );
    config.controlValues[ OA_CAM_CTRL_AUTO_WHITE_BALANCE ] = AWBManual;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_WHITE_BALANCE, AWBManual );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_RED_BALANCE )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_RED_BALANCE, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_RED_BALANCE ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_RED_BALANCE, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_BLUE_BALANCE )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_BLUE_BALANCE, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_BLUE_BALANCE ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_BLUE_BALANCE, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_USBTRAFFIC )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_USBTRAFFIC, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_USBTRAFFIC ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_USBTRAFFIC, 0 );
  }
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_CONTRAST )) {
    state.camera->setControl ( OA_CAM_CTRL_AUTO_CONTRAST, 0 );
    config.controlValues[ OA_CAM_CTRL_AUTO_CONTRAST ] = 0;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_AUTO_CONTRAST, 0 );
  }
#endif /* DISABLE_AUTO */

  // Now we need to set up the menus for the two selectable sliders
  // from the controls available.
  //
  // Dump the existing menus, then create new ones based on the available
  // INT32/64 controls.  Use the existing controls for the current settings
  // if they're still present, otherwise default to the first available.
  // Make the selected controls in each menu unavailable in the other.
  //
  // Also need to hide the old sliders/spinboxes and remove them from
  // the grid widget, adding the new ones and showing them.

  // Step 1.  Clear the existing data

  ignoreSelectableControlChanges = 1;

  for ( int c = 0; c < 2; c++ ) {
    if ( config.selectableControl[c] >= 0 ) {
      selectableControlMenu[c]->hide();
      selectableControlMenu[c]->clear();
      if ( selectableControlCheckbox[ config.selectableControl[c]] ) {
        selectableControlCheckbox[ config.selectableControl[c]]->hide();
      }
      selectableControlSlider[ config.selectableControl[c]]->hide();
      selectableControlSpinbox[ config.selectableControl[c]]->hide();
    }
  }

  // Step 2.  Find out if we can re-use the existing controls
  int replaceSelectable1 = 0;
  int type;
  if ( config.selectableControl[0] == -1 ) {
    // we can't allow this one
    replaceSelectable1 = -1;
  }
  type = 0;
  if ( config.selectableControl[0] >= 0 ) {
    type = state.camera->hasControl ( config.selectableControl[0]);
  }
  if ( type != OA_CTRL_TYPE_INT32 && type != OA_CTRL_TYPE_INT64 ) {
    // we can't allow this one
    replaceSelectable1 = -1;
  }

  int replaceSelectable2 = 0;
  if ( config.selectableControl[1] == -1 ||
      config.selectableControl[1] == config.selectableControl[0] ) {
    // we can't allow this one
    replaceSelectable2 = -1;
  }
  type = 0;
  if ( config.selectableControl[1] >= 0 ) {
    type = state.camera->hasControl ( config.selectableControl[1]);
  }
  if ( type != OA_CTRL_TYPE_INT32 && type != OA_CTRL_TYPE_INT64 ) {
    // we can't allow this one
    replaceSelectable2 = -1;
  }

  // Step 3.  Rebuild the menus whilst assigning new controls if required
  int c1, c2;
  c1 = c2 = 0;
  for ( int c = 0; c < OA_CAM_CTRL_LAST_P1; c++ ) {
    selectableControlsAllowed[0][ c ] = -1;
    selectableControlsAllowed[1][ c ] = -1;
    selectableControlIndexes[0][ c ] = selectableControlIndexes[1][ c ] = -1;
    if ( c != OA_CAM_CTRL_GAIN && c != OA_CAM_CTRL_EXPOSURE &&
        c != OA_CAM_CTRL_EXPOSURE_ABSOLUTE ) {
      type = state.camera->hasControl ( c );
      if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
        if ( replaceSelectable1 < 0 ) {
          replaceSelectable1 = config.selectableControl[0] = c;
          if ( c == config.selectableControl[1] ) {
            replaceSelectable2 = -1;
          }
        } else {
          if ( replaceSelectable2 < 0 && c != config.selectableControl[0] ) {
            replaceSelectable2 = config.selectableControl[1] = c;
          }
        }
        if ( c != config.selectableControl[1] ) {
          selectableControlMenu[0]->addItem ( tr ( oaCameraControlLabel[c] ));
          selectableControlIndexes[0][ c ] = c1;
          selectableControlsAllowed[0][ c1 ] = c;
          c1++;
        }
        if ( c != config.selectableControl[0]) {
          selectableControlMenu[1]->addItem ( tr ( oaCameraControlLabel[c] ));
          selectableControlIndexes[1][ c ] = c2;
          selectableControlsAllowed[1][ c2 ] = c;
          c2++;
        }

        state.camera->controlRange ( c, &min, &max, &step, &def );
        selectableControlSlider[ c ]->setRange ( min, max );
        selectableControlSlider[ c ]->setSingleStep ( step );
        selectableControlSlider[ c ]->setValue ( config.controlValues[ c ] );
        selectableControlSpinbox[ c ]->setRange ( min, max );
        selectableControlSpinbox[ c ]->setSingleStep ( step );
        selectableControlSpinbox[ c ]->setValue ( config.controlValues[ c ] );
      }
    }
  }

  // Step 4, remove the current widgets from the grid and display the
  // correct ones

  grid->removeItem ( grid->itemAtPosition ( 2, 1 ));
  grid->removeItem ( grid->itemAtPosition ( 2, 2 ));
  grid->removeItem ( grid->itemAtPosition ( 2, 3 ));
  int autoctrl;
  if ( replaceSelectable1 >= 0 ) {
    selectableControlMenu[0]->show();
    autoctrl = oaGetAutoForControl ( config.selectableControl[0] );
    if ( selectableControlCheckbox[ config.selectableControl[0]] &&
        autoctrl >= 0 && state.camera->hasControl ( autoctrl ) ==
        OA_CTRL_TYPE_BOOLEAN ) {
      grid->addWidget (
          selectableControlCheckbox[ config.selectableControl[0]], 2, 1 );
      selectableControlCheckbox[ config.selectableControl[0]]->show();
    }
    grid->addWidget (
        selectableControlSlider[ config.selectableControl[0]], 2, 2 );
    selectableControlSlider[ config.selectableControl[0]]->show();
    grid->addWidget (
        selectableControlSpinbox[ config.selectableControl[0]], 2, 3 );
    selectableControlSpinbox[ config.selectableControl[0]]->show();
  }

  grid->removeItem ( grid->itemAtPosition ( 3, 1 ));
  grid->removeItem ( grid->itemAtPosition ( 3, 2 ));
  grid->removeItem ( grid->itemAtPosition ( 3, 3 ));
  if ( replaceSelectable2 >= 0 ) {
    selectableControlMenu[1]->show();
    autoctrl = oaGetAutoForControl ( config.selectableControl[1] );
    if ( selectableControlCheckbox[ config.selectableControl[1]] &&
        autoctrl >= 0 && state.camera->hasControl ( autoctrl ) ==
        OA_CTRL_TYPE_BOOLEAN ) {
      grid->addWidget (
          selectableControlCheckbox[ config.selectableControl[1]], 3, 1 );
      selectableControlCheckbox[ config.selectableControl[1]]->show();
    }
    grid->addWidget (
        selectableControlSlider[ config.selectableControl[1]], 3, 2 );
    selectableControlSlider[ config.selectableControl[1]]->show();
    grid->addWidget (
        selectableControlSpinbox[ config.selectableControl[1]], 3, 3 );
    selectableControlSpinbox[ config.selectableControl[1]]->show();
  }

  // Step 5, set the current items to be the visible ones in the menu

  if ( replaceSelectable1 >= 0 ) {
    selectableControlMenu[0]->setCurrentIndex (
        selectableControlIndexes[0][ config.selectableControl[0]]);
    if ( replaceSelectable2 >= 0 ) {
      selectableControlMenu[1]->setCurrentIndex (
          selectableControlIndexes[1][ config.selectableControl[1]]);
    } else {
      config.selectableControl[1] = -1;
    }
  } else {
    config.selectableControl[0] = -1;
    config.selectableControl[1] = -1;
  }

  // And we're all done with the selectable controls

  ignoreSelectableControlChanges = 0;

  type = state.camera->hasControl ( OA_CAM_CTRL_GAIN );
  if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
    ignoreGainChanges = 1;
    config.controlValues[ OA_CAM_CTRL_GAIN ] =
        state.camera->readControl ( OA_CAM_CTRL_GAIN );
    state.camera->controlRange ( OA_CAM_CTRL_GAIN, &min, &max, &step, &def );
    selectableControlSlider[ OA_CAM_CTRL_GAIN ]->setRange ( min, max );
    selectableControlSlider[ OA_CAM_CTRL_GAIN ]->setSingleStep ( step );
    // selectableControlSlider[ OA_CAM_CTRL_GAIN ]->setValue (
    //     config.controlValues[ OA_CAM_CTRL_GAIN ] );
    selectableControlSpinbox[ OA_CAM_CTRL_GAIN ]->setRange ( min, max );
    selectableControlSpinbox[ OA_CAM_CTRL_GAIN ]->setSingleStep ( step );
    ignoreGainChanges = 0;
    selectableControlSpinbox[ OA_CAM_CTRL_GAIN ]->setValue (
        config.controlValues[ OA_CAM_CTRL_GAIN ] );
    selectableControlSlider[ OA_CAM_CTRL_GAIN ]->show();
    selectableControlSpinbox[ OA_CAM_CTRL_GAIN ]->show();
    if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_GAIN ) ==
        OA_CTRL_TYPE_BOOLEAN ) {
      selectableControlCheckbox[ OA_CAM_CTRL_GAIN ]->show();
    }
  } else {
    selectableControlSlider[ OA_CAM_CTRL_GAIN ]->hide();
    selectableControlSpinbox[ OA_CAM_CTRL_GAIN ]->hide();
    selectableControlCheckbox[ OA_CAM_CTRL_GAIN ]->hide();
  }

  // Exposure is a bit more messy.  If we have an absolute exposure time
  // then prefer that.  If we don't, but have some sort of exposure control
  // use that, but the labels will need adjusting.  If we have neither then
  // the controls need to be disabled.
  // At the same time we need to create a new set of labels for the
  // exposure combobox.

  int updateExposureControls = 0;
  int setting;

  type = state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE );
  if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
    state.camera->controlRange ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, &min, &max,
      &step, &def );
    usingAbsoluteExposure = 1;
    // The settings from libopenastro are supposed to be in units of 1us
    // so we need to convert those to milliseconds by dividing by 1000.
    // This may mean we lose a bit of resolution at the bottom end.  So be it.
    min /= 1000;
    if ( min < 1 ) { min = 1; }
    max /= 1000;
    if ( max < 1 ) { max = 1; }
    step /= 1000;
    if ( step < 1 ) { step = 1; }
    def /= 1000;
    if ( def < 1 ) { def = 1; }
    // setting = state.camera->readControl (
    //     OA_CAM_CTRL_EXPOSURE_ABSOLUTE ) / 1000;
    setting = config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
    if ( setting < min ) {
      setting = min;
    } else {
      if ( setting > max ) {
        setting = max;
      }
    }
    // config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = setting;
    updateExposureControls = 1;
    exposureLabel->setText ( tr ( "Exposure (ms)" ));
    expRangeLabel->setText ( tr ( "Exp. Range (ms)" ));
  } else {
    type = state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE );
    if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
      usingAbsoluteExposure = 0;
      state.camera->controlRange ( OA_CAM_CTRL_EXPOSURE, &min, &max, &step,
          &def );
      // setting = state.camera->readControl ( OA_CAM_CTRL_EXPOSURE );
      // config.controlValues[ OA_CAM_CTRL_EXPOSURE ] = setting;
      setting = config.controlValues[ OA_CAM_CTRL_EXPOSURE ];
      if ( setting < min ) {
        setting = min;
      } else {
        if ( setting > max ) {
          setting = max;
        }
      }
      updateExposureControls = 1;
      exposureLabel->setText ( tr ( "Exposure" ));
      expRangeLabel->setText ( tr ( "Exp. Range" ));
    } else {
      exposureSlider->hide();
      exposureSpinbox->hide();
      expRangeLabel->hide();
      expMenu->hide();
    }
  }

  if ( updateExposureControls ) {

    // now we need to work out the ranges for the combobox
    // if the range is no greater than 256 we hide the combobox and
    // set the slider to the appropriate range.
    // otherwise for non-absolute exposures we divide the range up into
    // five ranges and for absolute have increasing wider scales

    int diff = max - min;
    int showMin = 1, showMax = 1;
    if ( usingAbsoluteExposure || ( diff / step + 1 ) > 256 ) {

      QStringList exposures;
      int v1 = min;
      int v2;
      int items = 0, showItem = 0;
      do {
        if ( usingAbsoluteExposure ) {
          v2 = v1;
          if ( !v2 ) { v2++; }
          v2 *= 10;
          if ( v2 > max ) { v2 = max; }
        } else {
         v2 = min + diff / step / 5 + 1;
        }
        exposures << QString::number ( v1 ) + " - " + QString::number ( v2 );
        if ( setting >= v1 && setting <= v2 ) {
          showMin = v1;
          showMax = v2;
          showItem = items;
        }
        minSettings[ items ] = v1;
        maxSettings[ items ] = v2;
        items++;
        v1 = v2;
        // we have to stop this somewhere...
      } while ( items < 7 && v2 < max );

      // that should be the string list done.  now delete all the old items
      // and add the new string list and set the current item

      ignoreExposureMenuChanges = 1;
      expMenu->clear();
      expMenu->addItems ( exposures );
      config.exposureMenuOption = showItem;
      ignoreExposureMenuChanges = 0;
      expMenu->setCurrentIndex ( showItem );
      expRangeLabel->show();
      expMenu->show();
      useExposureDropdown = 1;
    } else {
      showMin = min;
      showMax = max;
      expRangeLabel->hide();
      expMenu->hide();
      useExposureDropdown = 0;
    }

    // finally we can set the sliders up

    exposureSlider->setRange ( showMin, showMax );
    exposureSlider->setSingleStep ( step );
    exposureSlider->setValue ( setting );
    exposureSpinbox->setRange ( showMin, showMax );
    exposureSpinbox->setSingleStep ( step );
    exposureSpinbox->setValue ( setting );
    exposureSlider->show();
    exposureSpinbox->show();

    if ( usingAbsoluteExposure ) {
      if ( setting ) {
        state.mainWindow->showFPSMaxValue ( 1000 / setting );
      }
    } else {
      state.mainWindow->clearFPSMaxValue();
    }
  }

  grid->removeItem ( grid->itemAtPosition ( 5, 1 ));
  if ( state.camera->hasControl ( OA_CAM_CTRL_AUTO_EXPOSURE ) ==
      OA_CTRL_TYPE_BOOLEAN ) {
    if ( state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE )) {
      grid->addWidget (
          selectableControlCheckbox[ OA_CAM_CTRL_EXPOSURE ], 5, 1 );
      selectableControlCheckbox[ OA_CAM_CTRL_EXPOSURE ]->show();
    } else {
      if ( state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE )) {
        grid->addWidget (
            selectableControlCheckbox[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ], 5, 1 );
        selectableControlCheckbox[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ]->show();
      }
    }
  }

  updateFrameRates();
}


void
ControlWidget::updateFrameRates ( void )
{
  if ( !state.camera->isInitialised()) {
    return;
  }
  if ( !state.camera->hasFrameRateSupport()) {
    framerateLabel->hide();
    framerateSlider->hide();
    framerateMenu->hide();
    return;
  }

  if ( state.camera->hasFixedFrameRates ( config.imageSizeX,
      config.imageSizeY )) {

    // Ugly sorting of the frame rates here.  We don't know what
    // order we'll get them in so we have to sort them into rate
    // order for the slider to work.

    const FRAMERATE* rates = state.camera->frameRates ( config.imageSizeX,
        config.imageSizeY );

    // don't use the frame rate slider if there's only one frame rate
    if ( !rates[1].denominator ) {
      framerateLabel->hide();
      framerateSlider->hide();
      framerateMenu->hide();
      config.frameRateNumerator = rates[ 0 ].numerator;
      config.frameRateDenominator = rates[ 0 ].denominator;
      return;
    }

    QMap<float,QString> sortMap;
    QMap<float,int> numeratorMap;
    QMap<float,int> denominatorMap;
    int i;
    for ( i = 0; i < OA_MAX_FRAMERATES && rates[i].denominator; i++ ) {
      float frameTime = float ( rates[i].numerator ) /
          float ( rates[i].denominator );
      QString rateStr;
      if ( 1 == rates[i].numerator ) {
        rateStr = QString::number ( rates[i].denominator );
      } else {
        rateStr = QString::number ( rates[i].denominator ) + "/" +
            QString::number ( rates[i].numerator );
      }
      sortMap [ frameTime ] = rateStr;
      numeratorMap [ frameTime ] = rates[i].numerator;
      denominatorMap [ frameTime ] = rates[i].denominator;
    }

    QStringList rateList;
    QMap<float,QString>::iterator j;
    QMap<float,int>::iterator n, d;
    frameRateNumerator.clear();
    frameRateDenominator.clear();
    int numItems = 0, showItem = 0;
    for ( j = sortMap.begin(), n = numeratorMap.begin(),
        d = denominatorMap.begin(); j != sortMap.end(); j++, n++, d++ ) {
      rateList << *j;
      frameRateNumerator << *n;
      frameRateDenominator << *d;
      if ( *n == config.frameRateNumerator &&
          *d == config.frameRateDenominator ) {
        showItem = numItems;
      }
      numItems++;
    }

    ignoreFrameRateChanges = 1;
    framerateMenu->clear();
    framerateMenu->addItems ( rateList );
    framerateSlider->setRange ( 0, numItems - 1 );
    framerateSlider->setSingleStep ( 1 );
    config.frameRateNumerator = frameRateNumerator[ showItem ];
    config.frameRateDenominator = frameRateDenominator[ showItem ];
    ignoreFrameRateChanges = 0;
    framerateMenu->setCurrentIndex ( showItem );
    // unfortunately the above statement may not change the frame rate even
    // though the rate has changed, because its position in the menu might
    // be the same, so we have to do the update manually;
    frameRateChanged();

    framerateLabel->show();
    framerateSlider->show();
    framerateMenu->show();

  } else {
    fprintf ( stderr, "Camera::hasFixedFrameRates failed\n" );
    framerateLabel->hide();
    framerateSlider->hide();
    framerateMenu->hide();
  }
}


void
ControlWidget::exposureMenuChanged ( int index )
{
  if ( ignoreExposureMenuChanges ) {
    return;
  }

  int newMin = minSettings[ index ];
  int newMax = maxSettings[ index ];
  int newSetting;

  if ( usingAbsoluteExposure ) {
    newSetting = config.controlValues [ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
  } else {
    newSetting = config.controlValues [ OA_CAM_CTRL_EXPOSURE ];
  }
  if ( newSetting < newMin ) { newSetting = newMin; }
  if ( newSetting > newMax ) { newSetting = newMax; }
  
  exposureSlider->setRange ( newMin, newMax );
  exposureSlider->setValue ( newSetting );
  exposureSpinbox->setRange ( newMin, newMax );
  exposureSpinbox->setValue ( newSetting );

  if ( !state.camera->hasFrameRateSupport()) {
    if ( usingAbsoluteExposure ) {
      theoreticalFPSNumerator = newSetting;
      theoreticalFPSDenominator = 1000;
      theoreticalFPS = 1000 / newSetting;
    } else {
      // we don't really have a clue here, so set it to 1?
      theoreticalFPSNumerator = 1;
      theoreticalFPSDenominator = 1;
      theoreticalFPS = 1;
    }
  }
  if ( usingAbsoluteExposure ) {
    if ( newSetting ) {
      state.mainWindow->showFPSMaxValue ( 1000 / newSetting );
    }
  } else {
    state.mainWindow->clearFPSMaxValue();
  }
}


void
ControlWidget::updateExposure ( int value )
{
  if ( usingAbsoluteExposure ) {
    if ( value ) {
      state.mainWindow->showFPSMaxValue ( 1000 / value );
    }
    config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = value;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, value );
    // convert value back to microseconds from milliseconds
    state.camera->setControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, value * 1000 );
    if ( !state.camera->hasFrameRateSupport()) {
      theoreticalFPSNumerator = value;
      theoreticalFPSDenominator = 1000;
    }
    if ( state.settingsWidget ) {
      state.settingsWidget->updateControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE,
          value );
    }
  } else {
    state.mainWindow->clearFPSMaxValue();
    state.camera->setControl ( OA_CAM_CTRL_EXPOSURE, value );
    config.controlValues[ OA_CAM_CTRL_EXPOSURE ] = value;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_EXPOSURE, value );
    if ( state.settingsWidget ) {
      state.settingsWidget->updateControl ( OA_CAM_CTRL_EXPOSURE, value );
    }
  }
}


void
ControlWidget::frameRateChanged ( void )
{
  if ( ignoreFrameRateChanges ) {
    return;
  }

  int index = framerateSlider->sliderPosition();
  if ( framerateSlider->isSliderDown()) {
    return;
  }

  /*
   * Leaving this in means the first frame setting for a camera breaks
   *
  if ( config.frameRateNumerator == frameRateNumerator[ index ] &&
      config.frameRateDenominator == frameRateDenominator[ index ] ) {
    return;
  }
  */

  config.frameRateNumerator = frameRateNumerator[ index ];
  config.frameRateDenominator = frameRateDenominator[ index ];
  theoreticalFPSNumerator = config.frameRateNumerator;
  theoreticalFPSDenominator = config.frameRateDenominator;
  theoreticalFPS = config.frameRateDenominator /
      config.frameRateNumerator;

  if ( state.settingsWidget ) {
    state.settingsWidget->updateFrameRate ( index );
  }

  if ( !state.camera->isInitialised()) {
    return;
  }

  state.camera->setFrameInterval ( config.frameRateNumerator,
      config.frameRateDenominator );
}


void
ControlWidget::updateGain ( int value )
{
  if ( !ignoreGainChanges ) {
    config.controlValues[ OA_CAM_CTRL_GAIN ] = value;
    SET_PROFILE_CONTROL( OA_CAM_CTRL_GAIN, value );
    state.camera->setControl ( OA_CAM_CTRL_GAIN, value );
    if ( state.settingsWidget ) {
      state.settingsWidget->updateControl ( OA_CAM_CTRL_GAIN, value );
    }
  }
}


void
ControlWidget::updateSelectableControl ( int control )
{
  if ( !ignoreSelectableControlChanges ) {
    int value = selectableControlSpinbox[ control ]->value();
    config.controlValues[ control ] = value;
    SET_PROFILE_CONTROL( control, value );
    state.camera->setControl ( control, value );
    if ( state.settingsWidget ) {
      state.settingsWidget->updateControl ( control, value );
    }
  }
}


void
ControlWidget::updateSelectableCheckbox ( int control )
{
  int autoControl = oaGetAutoForControl ( control );
  int value = ( selectableControlCheckbox[ control ]->isChecked()) ? 1 : 0;
  int origValue = value;

  if ( OA_CAM_CTRL_AUTO_EXPOSURE == autoControl &&
       state.camera->hasControl ( autoControl ) == OA_CTRL_TYPE_BOOLEAN ) {
    value = value ? OA_EXPOSURE_AUTO : OA_EXPOSURE_MANUAL;
  }
  config.controlValues[ autoControl ] = value;
  SET_PROFILE_CONTROL( autoControl, value );
  state.camera->setControl ( autoControl, value );
  if ( selectableControlSlider[ control ] ) {
    selectableControlSlider[ control ]->setEnabled ( !value );
    selectableControlSpinbox[ control ]->setEnabled ( !value );
  }
  if ( state.settingsWidget ) {
    state.settingsWidget->updateControl ( autoControl, origValue );
  }
}


void
ControlWidget::enableFPSControl ( int state )
{
  framerateSlider->setEnabled ( state );
  framerateMenu->setEnabled ( state );
}


int
ControlWidget::getFPSNumerator ( void )
{
  return theoreticalFPSNumerator;
}


int
ControlWidget::getFPSDenominator ( void )
{
  return theoreticalFPSDenominator;
}


void
ControlWidget::updateFromConfig ( void )
{
  int setFrameRate = -1;
  int exposureSetting = -1;
  int64_t min, max, step, def;
  int type;

  if ( state.camera->isInitialised()) {

    for ( int c = 0; c < OA_CAM_CTRL_LAST_P1; c++ ) {
      type = state.camera->hasControl ( c );
      if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {

        state.camera->controlRange ( c, &min, &max, &step, &def );

        if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE == c ) {
          type = state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE );
          if ( OA_CTRL_TYPE_INT32 == type ||
              OA_CTRL_TYPE_INT64 == type ) {
            state.camera->controlRange ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, &min,
                &max, &step, &def );
            // convert from microseconds to milliseconds
            min /= 1000;
            if ( min < 1 ) { min = 1; }
            max /= 1000;
            if ( max < 1 ) { max = 1; }
 
            if ( config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] < min ) {
              config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = min;
              SET_PROFILE_CONTROL( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, min );
            }
            if ( config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] > max ) {
              config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ] = max;
              SET_PROFILE_CONTROL( OA_CAM_CTRL_EXPOSURE_ABSOLUTE, max );
            }
          }
        } else {
          if ( config.controlValues[ c ] < min ) {
            config.controlValues[ c ] = min;
            SET_PROFILE_CONTROL( c, min );
          }
          if ( config.controlValues[ c ] > max ) {
            config.controlValues[ c ] = max;
            SET_PROFILE_CONTROL( c, max );
          }
        }
      }
    }

    type = state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE );
    if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
      exposureSetting = config.controlValues[ OA_CAM_CTRL_EXPOSURE_ABSOLUTE ];
    } else {
    type = state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE );
      if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
        exposureSetting = config.controlValues[ OA_CAM_CTRL_EXPOSURE ];
      }
    }

    if ( state.camera->hasFrameRateSupport()) {

      updateFrameRates();

      const FRAMERATE* rates = state.camera->frameRates ( config.imageSizeX,
          config.imageSizeY );

      for ( int i = 0; setFrameRate < 0 && i < OA_MAX_FRAMERATES &&
          rates[i].denominator; i++ ) {
        if ( rates[i].numerator == config.frameRateNumerator &&
            rates[i].denominator == config.frameRateDenominator ) {
          setFrameRate = i;
        }
      }
    }
  }

  for ( int c = 0; c < OA_CAM_CTRL_LAST_P1; c++ ) {
    if ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE != c && OA_CAM_CTRL_EXPOSURE != c ) {
      type = state.camera->hasControl ( c );
      if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
        selectableControlSpinbox[ c ]->setValue ( config.controlValues[ c ]);
      }
    }
  }

  if ( useExposureDropdown ) {
    int foundDropdownValue = -1;
    int numItems = expMenu->count();
    for ( int i = 0; i < numItems; i++ ) {
      if ( exposureSetting >= minSettings[i] && exposureSetting <=
          maxSettings[i] ) {
        foundDropdownValue = i;
      }
    }
    if ( foundDropdownValue == -1 ) {
      qWarning() << "can't find new exposure setting in dropdown";
    } else {
      expMenu->setCurrentIndex ( foundDropdownValue );
      exposureSlider->setRange ( minSettings[ foundDropdownValue ],
          maxSettings[ foundDropdownValue ] );
      exposureSpinbox->setRange ( minSettings[ foundDropdownValue ],
          maxSettings[ foundDropdownValue ] );
    }
  }
  exposureSpinbox->setValue ( exposureSetting );

  if ( setFrameRate >= 0 ) {
    framerateMenu->setCurrentIndex ( setFrameRate );
  }

  if ( usingAbsoluteExposure ) {
    if ( exposureSetting ) {
      state.mainWindow->showFPSMaxValue ( 1000 / exposureSetting );
    }
  } else {
    state.mainWindow->clearFPSMaxValue();
  }

}


void
ControlWidget::setSelectableControl1 ( int index )
{
  _setSelectableControl ( 0, index );
}


void
ControlWidget::setSelectableControl2 ( int index )
{
  _setSelectableControl ( 1, index );
}


void
ControlWidget::_setSelectableControl ( int selector, int index )
{
  if ( ignoreSelectableControlChanges ) {
    return;
  }
  ignoreSelectableControlChanges = 1;

  // The easy bits -- change the current control, drop the old widgets and
  // display the new ones

  int thisOne = selector;
  int theOther = 1 - selector;

  if ( selectableControlCheckbox[ config.selectableControl[ thisOne ]] ) {
    selectableControlCheckbox[ config.selectableControl[ thisOne ]]->hide();
  }
  selectableControlSlider[ config.selectableControl[ thisOne ]]->hide();
  selectableControlSpinbox[ config.selectableControl[ thisOne ]]->hide();

  config.selectableControl[ thisOne ] =
      selectableControlsAllowed[ thisOne ][ index ];

  grid->removeItem ( grid->itemAtPosition ( thisOne + 2, 1 ));
  grid->removeItem ( grid->itemAtPosition ( thisOne + 2, 2 ));
  grid->removeItem ( grid->itemAtPosition ( thisOne + 2, 3 ));
  int autoctrl;
  autoctrl = oaGetAutoForControl ( config.selectableControl[ thisOne ] );
  if ( selectableControlCheckbox[ config.selectableControl[ thisOne ]] &&
      state.camera->hasControl ( autoctrl )) {
    grid->addWidget (
        selectableControlCheckbox[ config.selectableControl[ thisOne ]],
        thisOne + 2, 1 );
    selectableControlCheckbox[ config.selectableControl[ thisOne ]]->show();
  }
  grid->addWidget (
      selectableControlSlider[ config.selectableControl[ thisOne ]],
      thisOne + 2, 2 );
  grid->addWidget (
      selectableControlSpinbox[ config.selectableControl[ thisOne ]],
      thisOne + 2, 3 );
  selectableControlSlider[ config.selectableControl[ thisOne ]]->show();
  selectableControlSpinbox[ config.selectableControl[ thisOne ]]->show();

  // Now the entire second menu needs rebuilding from the available
  // controls (minus the one we just chose).

  selectableControlMenu[ theOther ]->clear();

  int c2 = 0, type;
  for ( int c = 0; c < OA_CAM_CTRL_LAST_P1; c++ ) {
    selectableControlIndexes[ theOther ][ c ] = -1;
    selectableControlsAllowed[ theOther ][ c ] = 0;
    if ( c != OA_CAM_CTRL_GAIN && c != OA_CAM_CTRL_EXPOSURE &&
        c != OA_CAM_CTRL_EXPOSURE_ABSOLUTE ) {
      type = state.camera->hasControl ( c );
      if ( OA_CTRL_TYPE_INT32 == type || OA_CTRL_TYPE_INT64 == type ) {
        if ( c != config.selectableControl[ thisOne ]) {
          selectableControlMenu[ theOther ]->addItem (
              tr ( oaCameraControlLabel[c] ));
          selectableControlIndexes[ theOther ][ c ] = c2;
          selectableControlsAllowed[ theOther ][ c2 ] = c;
          c2++;
        }
      }
    }
  }

  selectableControlMenu[ theOther ]->setCurrentIndex (
      selectableControlIndexes[ theOther ][
      config.selectableControl[ theOther ]]);

  ignoreSelectableControlChanges = 0;
}


int
ControlWidget::getSpinboxMinimum ( int control )
{
  // FIX ME -- at some point merge the gain and exposure sliders and
  // spinboxes with the others to avoid these special cases
  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return exposureSpinbox->minimum();
      break;
    default:
      if ( !selectableControlSpinbox[ control ] ) {
        qWarning() << "ControlWidget::getSpinboxMinimum on unset control"
            << control;
        return 0;
      }
      return selectableControlSpinbox[ control ]->minimum();
      break;
  }
}


int
ControlWidget::getSpinboxMaximum ( int control )
{
  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return exposureSpinbox->maximum();
      break;
    default:
      if ( !selectableControlSpinbox[ control ] ) {
        qWarning() << "ControlWidget::getSpinboxMaximum on unset control"
            << control;
        return 0;
      }
      return selectableControlSpinbox[ control ]->maximum();
      break;
  }
}


int
ControlWidget::getSpinboxValue ( int control )
{
  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return exposureSpinbox->value();
      break;
    default:
      if ( !selectableControlSpinbox[ control ] ) {
        qWarning() << "ControlWidget::getSpinboxValue on unset control"
            << control;
        return 0;
      }
      return selectableControlSpinbox[ control ]->value();
      break;
  }
}


int
ControlWidget::getSpinboxStep ( int control )
{
  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      return exposureSpinbox->singleStep();
      break;
    default:
      if ( !selectableControlSpinbox[ control ] ) {
        qWarning() << "ControlWidget::getSpinboxStep on unset control"
            << control;
        return 0;
      }
      return selectableControlSpinbox[ control ]->singleStep();
      break;
  }
}


void
ControlWidget::updateSpinbox ( int control, int value )
{
  switch ( control ) {
    case OA_CAM_CTRL_EXPOSURE:
    case OA_CAM_CTRL_EXPOSURE_ABSOLUTE:
      exposureSpinbox->setValue ( value );
      break;
    default:
      if ( !selectableControlSpinbox[ control ] ) {
        qWarning() << "ControlWidget::updateSpinbox on unset control"
            << control;
        return;
      }
      selectableControlSpinbox[ control ]->setValue ( value );
      break;
  }
}


void
ControlWidget::updateCheckbox ( int control, int value )
{
  // FIX ME -- this function is called from the settings widget
  // the test below might be more appropriately done there
  if ( oaIsAuto ( control )) {
    int baseControl;
    // Now we could be in trouble.  If the control is auto exposure, we
    // need to know if it's auto exposure for the absolute or non-absolute
    // version
    if ( OA_CAM_CTRL_AUTO_EXPOSURE == control ) {
      if ( state.camera->hasControl ( OA_CAM_CTRL_EXPOSURE_ABSOLUTE )) {
        baseControl = OA_CAM_CTRL_EXPOSURE_ABSOLUTE;
      } else {
        baseControl = OA_CAM_CTRL_EXPOSURE;
      }
    } else {
      baseControl = oaGetControlForAuto ( control );
    }
    selectableControlCheckbox[ baseControl ]->setChecked ( value );
  } else {
    config.controlValues[ control ] = value;
    SET_PROFILE_CONTROL( control, value );
    state.camera->setControl ( control, value );
  }
}


QStringList
ControlWidget::getFrameRates()
{
  int i, n;
  QStringList rates;

  n = framerateMenu->count();
  for ( i = 0; i < n; i++ ) {
    rates << framerateMenu->itemText ( i );
  }

  return rates;
}


void
ControlWidget::updateFrameRate ( int index )
{
  framerateSlider->setValue ( index );
}
